' /////////////////////////////////////////////////////////////////////////////
{ 
 Propeller C3 Onboard FLASH memory demo serial terminal version - c3_flash_demo_sterm_010.spin

 Note: this version is identical to the "c3_flash_demo_010.spin" version that runs
 with a local PS/2 keyboard and mouse, except this version requires only a PC running
 the Parallax serial terminal connected to the C3's serial port. You can download
 the Parallax serial terminal here:

 http://www.parallax.com/Portals/0/Downloads/sw/propeller/Parallax-Serial-Terminal.exe     


 AUTHOR: Andre' LaMothe
 LAST MODIFIED: 9/3/10
 VERSION: 1.0
 LICENSE: MIT
 
 COMMENTS: This demo illustrates the use of the on board 1MB SPI serial FLASH
 memory module. The FLASH memory is manufactured by Atmel part #AT26DF081A-SSU.
 The datasheet for the chip can be found here:

 http://www.atmel.com/dyn/products/product_card.asp?part_id=3822

 The chip runs up to 70 Mhz on the SPI bus, has 256 byte pages and is compatible
 with the JEDEC spec for FLASH memory and pinouts. In other words, if you have
 written a FLASH driver for say a Microchip, or ST part then chances are it will
 take moments to port it to this chip.

 That said, FLASH chips have HUGE functionality and writing a generic driver that
 supports all the functions would be a waste. Therefore, I have provided a "starter"
 API here to get you going, so you can write, read, erase the chip and start
 experimenting. I am sure there are numerous serial FLASH drivers you can find
 for the chip and even some in SPIN! Here's a list of the API functions:

 PUB Flash_Read( flash_start_addr, buffer_ptr, num_bytes)  - Reads bytes from flash into a buffer. 
 PUB Flash_Write( flash_start_addr, buffer_ptr, num_bytes) - Writes bytes to flash from buffer. 
 PUB Flash_Erase( block ) - Erases a block of the flash. 
 PUB Flash_Erase_Chip( wait ) - Erases the entire chip.
 PUB Flash_Close          - Closes the flash (not much now, but this is a good place to de-allocate things in the future).       
 PUB Flash_Open           - Opens the flash (again not much now, but this is a good place to allocate things in the future). 
 PUB Flash_Write_Status ( status ) - Writes the STATUS register of the flash memory. 
 PUB Flash_Read_Status             - Reads the STATUS register of the flash memory.
 
 These functions more or less give you access to the FLASH memory and are based on
 a handful of the FLASH commands (read the datasheet to learn more). 

 Additionally, this demo has a number of defines/constants to help deal with the
 chip, so pay attention to those if you want to pull the source from this file to
 build an application then you will need those constants. Additionally, there is
 data structure that holds starting addresses of "sectors". The Atmel chip like
 many others has a number of quirks, one of them is that instead of having same
 sized sectors, it has different sized sectors at differnet addresses, so we need
 a list of these since the code to generate them procedurally would be 10x longer
 than a simple table. Thus, you need the data structure "sector_addresses" in the
 DAT section below.

 The demo program is a bit complex since it basically gives you a conduit to the FLASH
 memory, so you can read/write data, etc. So, if you know what an assembly language
 "monitor" is, it feels like that -- if you don't, in other words, all you do is read
 and write memory and type a lot :) However, you can easily use the API to save data
 from a mic or something else plugged into the A/D port, then play it back, or whatever.
 Point is, this is a 1 MegaByte FLASH -- a good place to store data that STAYS with the
 C3 and can't be pulled out like the SD card.   
  
 The action all occurs in the main() function, below the main() are a number of "utility"
 functions  included in each demo for reference. If you wish you can break them out into an
 external spin module. They are mostly for parsing and various random functions needed
 in each demo. The only important set of functions for the C3 are the "SPI" bus control
 functions, pay close attention to those since they control the selection of the various
 SPI devices on the bus. 
 
 REQUIRES:
 
         (1) C3 connected to PC via USB UART.
         (2) Parallax serial terminal running (and connected to serial port C3 is attached to, COM1,2,3...)

                                                           
   
}
' /////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON

'//////////////////////////////////////////////////////////////////////////////
' PROP CHIP / SYSTEM DEFINES //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' These settings are for 5 MHZ XTALS
  _clkmode = xtal1 + pll16x         ' enable external clock and pll times 16
  _xinfreq = 5_000_000              ' set frequency to 5 MHZ

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000
  
  _stack   = 128                    ' accomodate display memory and stack

'//////////////////////////////////////////////////////////////////////////////
' C3 PIN MAP DEFINES //////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' status/debug pin also controls VGA buffer enable
  STATUS_LED_BUS_MUX     = 15 ' pin that controls VGA enable as well as a status/debug LED

  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

  ' PS/2 port pin defines
  PS2_DATA               = 26 ' PS/2 port clock pin
  PS2_CLK                = 27 ' PS/2 data pin               

  ' Port A pin defines
  PORTA0                 = 0
  PORTA1                 = 1
  PORTA2                 = 2
  PORTA3                 = 3
  PORTA4                 = 4
  PORTA5                 = 5
  PORTA6                 = 6
  PORTA7                 = 7

  ' Servo port 3-pin header output pins (shared with Port A P4..P7)
  ' note 1: each I/O pins is buffered thru a 50 ohm resistor to minimize long transmission line effects from cables
  ' note 2: each "servo" port has a 3 pin header [signal, power, ground], each power is feed from a jumper adjacent to the
  ' servo port and has a 22uF cap across it as a current source, thus the "servo" ports aside from using the servo 3-pin cable
  ' pin outs, CAN drive servos!!! Care must be taken not to sink too much current, if you find that the board resets under harsh
  ' conditions while driving multiple simulataneous servos then you can add a bulk 2200uF cap for energy storage. There are two holes
  ' pre-drilled by the 2.1mm power entry port, simply solder a 2200uF cap under or above the board, just make sure to get the polarity
  ' correct: minus to the left (next to left edge of board), plus to the right (next to switch).
  SERVO_P4               = 4 
  SERVO_P5               = 5
  SERVO_P6               = 6
  SERVO_P7               = 7


  ' Port B pin defines (muxed with VGA port)
  PORTB0                 = 16  
  PORTB1                 = 17
  PORTB2                 = 18
  PORTB3                 = 19
  PORTB4                 = 20
  PORTB5                 = 21
  PORTB6                 = 22
  PORTB7                 = 23

  ' VGA port pin defines
  VGA_HSYNC              = 16
  VGA_VSYNC              = 17
  VGA_B0                 = 18
  VGA_B1                 = 19
  VGA_G0                 = 20
  VGA_G1                 = 21
  VGA_R0                 = 22
  VGA_R1                 = 23

  ' NTSC/PAL composite video pin defines
  COMPVID0               = 12
  COMPVID1               = 13
  COMPVID2               = 14

  ' Audio output pin define (PWM)
  AUDIO_MONO             = 24

  ' I2C pin defines (standard Prop pins)
  I2C_SCL                = 28
  I2C_SDA                = 29
 
  ' TX/RX pin defines (standard Prop pins)
  UART_TX                = 31
  UART_RX                = 30

   ' ASCII codes for ease of parser development
  ASCII_A       = 65
  ASCII_B       = 66
  ASCII_C       = 67
  ASCII_D       = 68
  ASCII_E       = 69
  ASCII_F       = 70
  ASCII_G       = 71
  ASCII_H       = 72
  ASCII_O       = 79  
  ASCII_P       = 80
  ASCII_Z       = 90
  ASCII_0       = 48
  ASCII_9       = 57
  ASCII_LEFT   = $C0
  ASCII_RIGHT  = $C1
  ASCII_UP     = $C2
  ASCII_DOWN   = $C3 
  ASCII_BS     = $08 ' backspace
  ASCII_DEL    = $7F ' delete   
  ASCII_LF     = $0A ' line feed 
  ASCII_CR     = $0D ' carriage return
  ASCII_ESC    = $1B ' escape
  ASCII_HEX    = $24 ' $ for hex
  ASCII_BIN    = $25 ' % for binary
  ASCII_LB     = $5B ' [ 
  ASCII_SEMI   = $3A ' ; 
  ASCII_EQUALS = $3D ' = 
  ASCII_PERIOD = $2E ' .
  ASCII_COMMA  = $2C ' ,
  ASCII_SHARP  = $23 ' #
  ASCII_NULL   = $00 ' null character
  ASCII_SPACE  = $20 ' space
  ASCII_TAB    = $09 ' horizontal tab

  ' null pointer, null character
  NULL         = 0  

'//////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY DEFINES ////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

' defines for the proper FLASH installed on C3 4MB, 8MB, 16MB
' 4Mb (512K Bytes)
AT26F004        = 4  ' numeric codes for flash version
AT25DF041A      = 4

' 8Mb (1M Bytes)
AT26DF081A      = 8
AT25DF081A      = 8

' 16 Mb (2M Bytes)
AT25DF161       = 16

' set type of FLASH ROM currently installed in C3 (8MBit) for conditional compilation of code -- not supported in SPIN, so manually done below :( 
FLASH_ROM             = AT26DF081A 
NUM_FLASH_ROM_SECTORS = 19 ' 11 for 4MB, 19 for 8MB, must be set manually due to lack of conditional compilation


' Atmel 2F004/8xxx series basic commands
WRITE_ENABLE     = $06
WRITE_DISABLE    = $04
READ_STATUS      = $05
WRITE_STATUS     = $01
READ_DATA        = $03

SEQ_PROGRAM      = $AF  
BYTE_PROGRAM     = $02 

BLOCK_ERASE4     = $20
BLOCK_ERASE32    = $52
BLOCK_ERASE64    = $D8

SECTOR_ERASE     = $20    

CHIP_ERASE       = $60 
JEDEC_ID         = $9F

PROTECT_SECTOR   = $36 
UNPROTECT_SECTOR = $39 

' custom commands we make up, not part of chip spec, just used here, start at code 0xF0
UNPROTECT_CHIP   = $F0


'//////////////////////////////////////////////////////////////////////////////
' VAR SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  
VAR

  ' generic globals to make demo writing easy
  long g_data, g_data2, g_status    
  long g_data32, g_data16, g_spi_result
  long g_key, g_temp1, g_temp2, g_temp3, g_index, g_length
  long g_tempx, g_tempy 
  ' this buffer is used as the read/write buffer for the FLASH, so you can't read or write
  ' more bytes than the size of this buffer
  byte g_sbuffer[4096] 
                      

'//////////////////////////////////////////////////////////////////////////////
' DAT SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

DAT

'  conditional compilation for FLASH memory variants not supported, so just comment out either declaration

' #if (FLASH_ROM == AT26F004)

'// we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
'// 11 sectors [64K, 64K, 64K, 64K, 64K, 64K, 64K, 32K, 8K, 8k, 16K ]
'//            |<------- (7) 64K sectors ------->|  1      2     1 

{
    ' 11 sectors in 4MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 32K sector 7
                     long $78000   ' 8K sector 8  
                     long $7A000   ' 8K sector 9
                     long $7D000   ' 8K sector 10
}

' #if (FLASH_ROM == AT26DF081A)

' we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
' 19 sectors [64K, 64K, 64K, ..., 64K, 64K, 64K, 64K, 16K, 8K, 8k, 32K ]
'            |<----------- (15) 64K sectors -------->| 1     2     1 

    ' 19 sectors in 8MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 64K sector 7 
                     long $80000   ' 64K sector 8
                     long $90000   ' 64K sector 9
                     long $A0000   ' 64K sector 10
                     long $B0000   ' 64K sector 11
                     long $C0000   ' 64K sector 12
                     long $D0000   ' 64K sector 13
                     long $E0000   ' 64K sector 14
                          
                     long $F0000   ' 16K sector 15
                     long $F4000   '  8K sector 16
                     long $F6000   '  8K sector 17
                          
                     long $F8000    ' 32K sector 18

'//////////////////////////////////////////////////////////////////////////////
'OBJS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

  ' current drivers used in this version of SPI interface
  serial     : "FullDuplexserial_drv_014.spin"       ' the full duplex serial driver
  gfx_ntsc : "C3_GFX_DRV_010.SPIN"                 ' instantiate new NTSC tile driver


'//////////////////////////////////////////////////////////////////////////////
' MAIN ENTRY POINT TO DEMO/////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB Main : status

  ' let the system initialize 
  Delay_US( 1*1_000_000 )

  ' initialize the graphics tile engine
  gfx_ntsc.start

  ' intialize serial driver (only works if nothing else is on serial port)
  serial.start(31, 30, %0000, 38_400) ' receive pin, transmit pin, baud rate

  ' /////////////////////////////////////////////////////////////////////////////
  ' INITIALIZE I/O PINS /////////////////////////////////////////////////////////
  ' /////////////////////////////////////////////////////////////////////////////

  ' blink the status LED 3 times to show board is "alive", then enable VGA (leave LOW)
  DIRA[STATUS_LED_BUS_MUX] := 1 ' set to output
  OUTA[STATUS_LED_BUS_MUX] := 0

  repeat 6
    OUTA[STATUS_LED_BUS_MUX] := !OUTA[STATUS_LED_BUS_MUX]
    repeat 25_000  

  ' turn on VGA buffer
  OUTA[STATUS_LED_BUS_MUX] := 0

  ' initialize SPI IO (eventhough, we aren't using it in this demo at all, good habit to set it up)
  SPI_Init
  
  ' begin test loop

  ' clear screen
  serial.tx( $00 )

  repeat   
    Flash_Demo

         
' end Main

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Demo

  '/////////////////////////////////////////////////////////////////////////////
  ' FLASH MEMORY DEMO //////////////////////////////////////////////////////////
  '/////////////////////////////////////////////////////////////////////////////

  ' open the flash memory up (unprotect it)
  Flash_Open
  
  repeat
    ' draw menu
    serial.tx( ASCII_CR )   
    serial.tx( ASCII_CR )   
    serial.txstring( string ("C3 FLASH Memory Demo Menu V1.0 (Serial Terminal Version)") )
    serial.tx( ASCII_CR )   
    serial.tx( ASCII_CR )    
    serial.txstring( string ("1.Read STATUS register.") )
    serial.tx( ASCII_CR )    
    serial.txstring( string ("2.Write STATUS register.") )
    serial.tx( ASCII_CR )    
    serial.txstring( string ("3.Write [start,value,numbytes].") )
    serial.tx( ASCII_CR )    
    serial.txstring( string ("4.Read [start,numbytes].") )
    serial.tx( ASCII_CR )   
    serial.txstring( string ("5.Erase chip.") )
    serial.tx( ASCII_CR )   
    serial.txstring( string ("6.Exit back to main menu.") )
    serial.tx( ASCII_CR )   
    serial.tx( ASCII_CR )    
    serial.txstring( string ("Selection?") )

    ' get string from user
    Get_String_Serial(@g_sbuffer, 8 )

    ' convert to integer
    g_key := atoi2(@g_sbuffer, 8)

    ' what is user requesting?
   case g_key
     1:
       g_temp1 := Flash_Read_Status      
       serial.tx( ASCII_CR )    
       serial.txstring( string ("Status Register = ") )
       serial.dec( g_temp1 )

     2:
       serial.tx( ASCII_CR )    
       serial.txstring( string ("Value to write?") )

       ' get string from user
       Get_String_Serial(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       ' write status bits
       Flash_Write_Status ( g_temp1 )
       
       serial.tx( ASCII_CR )   
       serial.txstring( string ("Writing ") )
       serial.dec( g_temp1 )

     3:
       serial.tx( ASCII_CR )    
       serial.txstring( string ("Start address?") )

       ' get string from user
       Get_String_Serial(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       serial.tx( ASCII_CR )    
       serial.txstring( string ("Value to write?") )

       ' get string from user
       Get_String_Serial(@g_sbuffer, 8 )

       ' convert to integer
       g_temp2 := atoi2(@g_sbuffer, 8)

       serial.tx( ASCII_CR )    
       serial.txstring( string ("Number of bytes to write?") )

       ' get string from user
       Get_String_Serial(@g_sbuffer, 8 )

       ' convert to integer
       g_temp3 := atoi2(@g_sbuffer, 8)

       ' fill buffer with value
       bytefill ( @g_sbuffer, g_temp2, g_temp3) 

       ' write buffer to flash
       Flash_Write( g_temp1, @g_sbuffer, g_temp3 )

     4:

       serial.tx( ASCII_CR )    
       serial.txstring( string ("Start address?") )

       ' get string from user
       Get_String_Serial(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       serial.tx( ASCII_CR )    
       serial.txstring( string ("Number of bytes to read?") )

       ' get string from user
       Get_String_Serial(@g_sbuffer, 8 )

       ' convert to integer
       g_temp3 := atoi2(@g_sbuffer, 8)

       ' read flash into buffer
       Flash_Read( g_temp1, @g_sbuffer, g_temp3 )

      ' print buffer to screen
       serial.tx( ASCII_CR )    
       serial.txstring( string ("Bytes read:") )
       
       repeat g_index from 0 to g_temp3-1
         ' print starting address to left every 8 bytes
         if ((g_index // 16) == 0)
           serial.tx( ASCII_CR )   
           serial.hex(g_index + g_temp1, 6 )
           serial.txstring( string (":") )                

         ' print byte
         serial.hex( g_sbuffer [ g_index ], 2 )
         serial.tx(  "," )                    

    
     5: 'erase chip
       serial.tx( ASCII_CR )   
       serial.tx( ASCII_CR )    
       serial.txstring( string ("Erasing chip...") )
       Flash_Erase_Chip( 1 )
       serial.tx( ASCII_CR )    
       serial.txstring( string ("Erasure complete!") )

     ' return to caller
     6: return


' end Flash_Test

' /////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY API (ported from my C/C++ libraries note syntax leftovers!)
' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Read_Status | _spi_data
{{
Reads the FLASH memory status register

PARMS: None.
RETURNS: status register value
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, READ_STATUS, $FF )

  ' now read data returned
  _spi_data := SPI_Write_Read( 8, $FF, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return results
  return ( _spi_data )   

' // end Flash_Read_Status

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Write_Status ( status ) | _spi_data
{{
Writes the FLASH memory status register

PARMS:

status - value to update the status register with

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_STATUS, $FF )
  _spi_data := SPI_Write_Read( 8, status, $FF )  
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return success for now
  return ( 1 )   

' // end Flash_Write_Status

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Open | _index, _spi_data, _flash_start_addr
{{

This functions "opens" the FLASH memory simply by unprotecting all sectors starting at address 0

PARMS: None.
RETURNS: None.

}}


  ' use global unprotect to unprotect entire chip at once, write "0" to bits 2,3,4,5
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
  SPI_Select_Channel( 0 )

  Flash_Write_Status ( Flash_Read_Status & %11_0000_11 )
  Delay_MS( 10 )   
  ' // return success always
  return( 1 )

{  

  ' this code is used to iterate thru all sectors manually and have more control
  ' iterate thru all sectors and unprotect them
  repeat _index from 0 to NUM_FLASH_ROM_SECTORS-1

    '// STEP 1: enable writing to chip  ///////////////////////////////////////
    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    Delay_MS( 10 )

    '// unprotect sector  ///////////////////////////////////

    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, UNPROTECT_SECTOR, $FF )

    '// get sector address
    _flash_start_addr := sector_addresses[ _index ]

    '// write sector address, there are only a handful of sectors since each is 4-64K
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    ' delay 10
    Delay_MS( 10 ) 

    ' // end for index

  ' // return success always
  return( 1 )
}

' // end Flash_Open

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Close | _index, _spi_data, _flash_start_addr      
{{
This function "closes" the FLASH by protecting all of the sectors on the flash chip.

PARMS: None.
RETURNS: None.

}}

  '// protecting the sectors seems to have a side effect, we are not considering???
  '// we need to read the data sheet more, for now, just return, and don't protect 
  '// all the sectors, this is overkill anyway for now ... 
  return(1)


' // end Flash_Close

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Erase_Chip( wait ) | _spi_data
{{
This function erases the entire chip. The erase command takes a considerable amount
of time, thus this function issues the command and either returns immediately or waits.
To determine if the erase is complete the function polls the status register bit 0
which is the RDY/BSY bit, 0=device is idle, 1=device is busy

PARMS: wait - boolean, 1=wait for erasure to complete, 0-return immediately.

RETURNS: None.

}}
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, CHIP_ERASE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' test if user requested wait for completion?
  if (wait)
    repeat
      ' erase operation still ongoing?
      if ((Flash_Read_Status & $01)==0)
        return (1)
     
  '// return success for now
  return(1)

' /////////////////////////////////////////////////////////////////////////////


PUB Flash_Erase( block ) | _index, _spi_data, _flash_start_addr
{{
This function erases a block of FLASH memory.

PARMS:

block - block address to erase.

RETURNS: None.

}}
  '// STEP 1: enable writing to chip again /////////////////////////////////////////////////////
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
  Delay_MS( 100 )
   
  '// STEP 2: erase 4K block of chip
  _flash_start_addr := FLASH_BLOCK_TO_ADDR ( block )
   
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, BLOCK_ERASE4, $FF )
   
  '// write block address, there are only a handful of blocks since each is 4-64K
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// erasure takes a long time... 
  Delay_MS( 500 )
   
  '// return success for now
  return(1)
   
' // end Flash_Erase

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Write( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function writes num_bytes bytes to the FLASH memory at flash_start_addr from the sent buffer_ptr.

It uses the byte/page write function. Although, there is a "sequential" byte write function available
on the AT26DF081A the new rev of the Atel chips (AT25xxxx series) coming out in 2011 have replaced the sequential write with
another function :(, thus we simply have to check for page boundaries and re-issue another page write
every 256 bytes.

PARMS:

flash_start_addr - starting address in the FLASH memory to write the data at.
buffer_ptr       - local byte pointer to Prop memory holding data to write to FLASH memory.
num_bytes        - number of bytes to write to FLASH from source buffer_ptr. 

}}

  '// STEP 1: enable writing to the chip again
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// STEP 2: write buffer to address in flash //////////////////
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
   
  '// initiate byte/page program command
  _spi_data := SPI_Write_Read( 8, BYTE_PROGRAM, $FF )
   
  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )

  '// write first byte of 
  _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ 0 ], $FF )

  _index := 1   

  '// write remaining bytes
  repeat until (_index => num_bytes )

    ' now test if we are on a page boundary, if so finish page, and start new page
    if ( ((flash_start_addr + _index) // 256) == 0)
      '// set CS to SPI select channel 0 (null)
      SPI_Select_Channel( 0 )

      ' replace this with polling...
      Delay_MS(10) 

      '// set CS to SPI select channel 3 (FLASH)
      SPI_Select_Channel( SPI_CHAN_FLASH )
      _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
      SPI_Select_Channel( 0 )

      SPI_Select_Channel( SPI_CHAN_FLASH )
               
      '// re-issue byte/page program command
      _spi_data := SPI_Write_Read( 8,BYTE_PROGRAM,$FF )

      '// write starting address byte by byte, MSB first...
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 16, $FF )
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 8, $FF )
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 0, $FF )

      ' end if 

    '// write the next byte of data
    _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ _index ], $FF )

    _index++
    
    ' // end for index    

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' write final bytes left in internal buffer, replace this with polling...
  Delay_MS(10) 
     
  '// disable SPI interface
   
  '// return success
  return(1)
   
' // end Flash_Write

' ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Flash_Read( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function reads num_bytes from the FLASH memory starting at memory location flash_start_addr and stores the data in
buffer_ptr.

PARMS:

flash_start_addr - starting address in FLASH memory to read data from.
buffer_ptr       - local Prop byte pointer to store the read data at.
num_bytes        - the number of bytes to read from the FLASH memory. 

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )

  '// initiate read command
  _spi_data := SPI_Write_Read( 8, READ_DATA, $FF )

  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )
  
  '// now read data 
  repeat _index from  0 to num_bytes-1
    byte [buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' // end for index

  '// disable SPI interface

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// return success 
  return(1)

' // end Flash_Read

' /////////////////////////////////////////////////////////////////////////////

'//////////////////////////////////////////////////////////////////////////////
' FLASH MACROS ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB FLASH_ADDR_TO_BLOCK( n )
  n := (n >> 12)

' /////////////////////////////////////////////////////////////////////////////
  
PUB FLASH_BLOCK_TO_ADDR( n )
  n := (n << 12)

' /////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////
' SPI ROUTINES///////// ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init


' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel


' /////////////////////////////////////////////////////////////////////////////
' UTILITY ROUTINES ////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB Get_String_Serial(string_ptr, max_length) | _length, _ch
{{
This simple function is a single line editor that allows user to enter a string from the
serial terminal and echos them back, when the user hits <ENTER> | <RETURN> the function
exits and returns the string. The function has simple editing and allows <BACKSPACE> to
delete the last character, that's it!

The function outputs to the serial terminal 

PARMS: string_ptr - pointer to string buffer to receive input line 
       max_length - maximum length of string
       echo_on    - echos all received characters back to terminal    

RETURNS: length of string, 0 if NULL.

}}

  ' current length of string buffer
  _length := 0  

  ' draw cursor
  
  repeat 

    ' draw cursor
    serial.tx( "_" )
    serial.tx( $03 )

    ' wait for key  
    _ch := serial.rx

    ' debugging
    ' term_ntsc.dec ( _ch )
     
    case _ch
       ASCII_LF, ASCII_CR: ' return    
 
        ' null terminate string and return
        byte [string_ptr][_length] := ASCII_NULL

        ' echo character
        serial.tx( _ch )
        return( _length )

       ASCII_BS, ASCII_DEL: ' backspace (edit)

         if (_length > 0)
           ' move cursor back once to overwrite last character on screen
           serial.tx ( ASCII_SPACE )
           serial.tx ( $03 )

           ' echo character
           serial.tx( _ch )
    
           ' decrement length
           _length--
 
       other:    ' all other cases
         ' insert character into string 
         byte [string_ptr][_length] := _ch

         ' echo character
         serial.tx( _ch )

         ' update length
         if (_length < max_length )
           _length++
         else
           ' move cursor back once to overwrite last character on screen
           serial.tx( $03 )          

     
' // end Get_String_Serial


' /////////////////////////////////////////////////////////////////////////////

PUB StrCpy( dest_str_ptr, source_str_ptr ) | _index
{{
mimics C strcpy, copies source -> destination including NULL
}}

' test if there is storage
if (dest_str_ptr == NULL)
  return (NULL)

_index := 0

repeat while (byte [ source_str_ptr ][_index] <> NULL)
  ' copy next byte
   byte [ dest_str_ptr ][_index] := byte [ source_str_ptr ][_index]    
   _index++

' null terminate
byte [ dest_str_ptr ][_index] := NULL

' return number of bytes copied
return ( _index )

' // end StrCpy

' /////////////////////////////////////////////////////////////////////////////

PUB StrUpper(string_ptr)
{{
upcases a string
}}

  if (string_ptr <> NULL)
    repeat while (byte[ string_ptr ] <> NULL)
      byte[ string_ptr ] :=  ToUpper( byte[ string_ptr ] )
      string_ptr++  

  ' return string
  return (string_ptr)

' // end StrUpper

' /////////////////////////////////////////////////////////////////////////////

PUB ToUpper(ch)
{{
returns the uppercase of the sent character
}}

if (ch => $61 and ch =< $7A)
  return(ch-32)
else
  return(ch)

' // end ToUpper

' /////////////////////////////////////////////////////////////////////////////

PUB IsInSet(ch, set_string)
{{
tests if sent character is in string
}}

repeat while (byte[set_string][0] <> NULL)
  if (ch == byte[set_string++][0])
    'serial.txstring(string ("found!") )
    return( ch )

'serial.txstring(string ("not found!") )
' not found
return( -1 )    

' // end IsInSet

' /////////////////////////////////////////////////////////////////////////////

PUB IsSpace(ch)
{{
tests if sent character is white space, cr, lf, space, tab
}}

if ( (ch == ASCII_SPACE) or (ch == ASCII_LF) or (ch == ASCII_CR) or (ch == ASCII_TAB))
  return (ch)
else
  return(-1)  

' // end IsSpace

' /////////////////////////////////////////////////////////////////////////////

PUB IsNull(ch)
{{
tests if sent character is null
}}

if ( (ch == NULL))
  return (1)
else
  return(-1)  

' // end IsNull

' /////////////////////////////////////////////////////////////////////////////

PUB IsDigit(ch)
{{
tests if sent character is a number 0..9, returns integer 0..9
}}

if (ch => ASCII_0 and ch =< ASCII_9)
  return (ch-ASCII_0)
else
  return(-1)  

' // end IsDigit

' /////////////////////////////////////////////////////////////////////////////

PUB IsAlpha(ch)
{{
tests if sent character is a number a...zA...Z
}}

ch := ToUpper(ch)

if ( (ch => ASCII_A) and (ch =< ASCII_Z))
  return (ch)
else
  return(-1)  

' end IsAlpha

' /////////////////////////////////////////////////////////////////////////////

PUB IsPunc(ch)
{{
tests if sent character is a punctuation symbol !@#$%^&*()--+={}[]|\;:'",<.>/?
}}

ch := ToUpper(ch)

if (  ((ch => 33) and (ch =< 47)) or  ((ch => 58) and (ch =< 64)) or ((ch => 91) and (ch =< 96)) or ((ch =>123) and (ch =< 126))  ) 
  return (ch)
else
  return(-1)  

' // end IsPunc
                                                                                
' /////////////////////////////////////////////////////////////////////////////

PUB HexToDec(n)
{{
converts hex digit to decimal
}} 
  if ( (n => "0") and (n =< "9") )
    return (n - ASCII_0)
  elseif ( (n => "A") and (n =< "F") )
    return (n - "A" + 10)
  else
    return ( 0 )  

' // end HexToDec

' /////////////////////////////////////////////////////////////////////////////

PUB itoa(value, string_ptr) | i
{{
 converts value into a string and stores in string_ptr with NULL termination
 }}

  if value < 0
    -value
    byte [ string_ptr++ ]  := "-"

  i := 1_000_000_000

  repeat 10
    if value => i
      byte [ string_ptr++ ]  := value / i + "0"
      value //= i
      result~~
    elseif result or i == 1
      byte [ string_ptr++ ]  := "0"
    i /= 10

  ' null terminate
  byte [ string_ptr ] := NULL    

' // end itoa

' /////////////////////////////////////////////////////////////////////////////

PUB atoi2(string_ptr, length) | _index, _sum, _ch, _sign
{{
this function tries to convert the string to a number, supports binary %, hex $, decimal (default)
string_ptr can be to a null terminated string, or the length can be overridden by sending the length shorter
than the null terminator, ignores ws
Eg. %001010110, $EF, 25
}}

' initialize vars
_index := 0
_sum   := 0
_sign  := 1
 
' consume white space
repeat while (IsSpace( byte[ string_ptr ][_index] ) <> -1)
  _index++
 
' is there a +/- sign?
if (byte [string_ptr][_index] == "+")
  ' consume it
  _index++
elseif (byte [string_ptr][_index] == "-")
  ' consume it
  _index++
  _sign := -1    
     
' try to determine number base
if (byte [string_ptr][_index] == ASCII_HEX)
  _index++
  repeat while ( ( IsDigit(_ch := byte [string_ptr][_index]) <> -1) or ( IsAlpha(_ch := byte [string_ptr][_index]) <> -1)  )
    _index++
    _sum := (_sum << 4) + HexToDec( ToUpper(_ch) )
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if hex number    
elseif (byte [string_ptr][_index] == ASCII_BIN)
  repeat while ( IsDigit(_ch := byte [string_ptr][++_index]) <> -1)
    _sum := (_sum << 1) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if binary number  
else
  ' must be in default base 10, assume that
  repeat while ( IsDigit(_ch := byte [string_ptr][_index++]) <> -1)
    _sum := (_sum * 10) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
 
' else, have no idea of number format!
return(0)
 
' // end atoi2

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US

'//////////////////////////////////////////////////////////////////////////////

DAT

newline_string          byte      $0D, $0A, $00  ' $0D carriage return, $0A line feed

{{
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}