' /////////////////////////////////////////////////////////////////////////////
{ 
 Propeller C3 NES Gamepad Demo -  c3_nesgamepad_demo_010.spin
 AUTHOR: Andre' LaMothe
 LAST MODIFIED: 9/3/10
 VERSION: 1.0

 LICENSE: MIT
 
 COMMENTS: This demo illustrates using a NES Gamepad with the C3. The NES gamepads
 have special connectors on them, so an adapter is needed to interface a generic
 NES controller to the C3. The Parallax NES Gamepad Adapter part #32368 shown below:

 http://www.parallax.com/StoreSearchResults/tabid/768/txtSearch/NES/List/0/SortField/4/ProductID/613/Default.aspx

 Can easily fit into one of the expansion headers on the C3. For this demo, and driver
 its assumed that the NES adapter is plugged into the C3 as shown in the ASCII diagram
 below, Port A, using IO signals P5...P0. Reading NES Gamepads is simply a matter
 of strobing the latch signal, and then clocking the button states out of the data
 lines. The NES controller have a simple parallel to serial convertor in them. You
 can read about NES interfacing on the Parallax NES adapter page.

 The only challenge with interfacing the NES adapater with the C3, is lack of +5 and GND
 signals in the exact position they need to be to match the NES adapter, however, luckily
 this isn't a problem since we can generate a "virtual" GND and +5 by setting the IO's P1 and
 P0 to HIGH and LOW respectively. This sets up a 3.3V and GND that can source/sink about 40mA.
 This is more than enough for the chips inside the NES gamepads! The API to read the gamepads
 consists of two functions (one to initialize the IO pins, and the other to read the pads):

  PUB NES_Read_Gamepad - Reads both gamepads (even if they aren't plugged in).
  PUB NES_Init_Gamepad - Initializes the IO pins for the Parallax NES Adapter.
 
 The action all occurs in the main() function, below the main() are a number of "utility"
 functions  included in each demo for reference. If you wish you can break them out into an
 external spin module. They are mostly for parsing and various random functions needed
 in each demo. The only important set of functions for the C3 are the "SPI" bus control
 functions, pay close attention to those since they control the selection of the various
 SPI devices on the bus. 
 
 REQUIRES:
           (1) PS/2 keyboard plugged in and A/V port connected to NTSC TV.
           (2) NES Gamepad adapater from Parallax Inc. (or equivalent circuit)
           (3) 1-2 NES compatible gamepads.

  - Game Pad with Parallax NES adapter part #32368, plugs into P7-P0 port as follows:


                  CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

 Note: ONLY insert ONE row of the header into the C3, otherwise the other row on the
 adapter will short into the SPI, I2C, etc.

   
}
' /////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON

'//////////////////////////////////////////////////////////////////////////////
' PROP CHIP / SYSTEM DEFINES //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' These settings are for 5 MHZ XTALS
  _clkmode = xtal1 + pll16x         ' enable external clock and pll times 16
  _xinfreq = 5_000_000              ' set frequency to 5 MHZ

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000
  
  _stack   = 128                    ' accomodate display memory and stack

'//////////////////////////////////////////////////////////////////////////////
' C3 PIN MAP DEFINES //////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' status/debug pin also controls VGA buffer enable
  STATUS_LED_BUS_MUX     = 15 ' pin that controls VGA enable as well as a status/debug LED

  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

  ' PS/2 port pin defines
  PS2_DATA               = 26 ' PS/2 port clock pin
  PS2_CLK                = 27 ' PS/2 data pin               

  ' Port A pin defines
  PORTA0                 = 0
  PORTA1                 = 1
  PORTA2                 = 2
  PORTA3                 = 3
  PORTA4                 = 4
  PORTA5                 = 5
  PORTA6                 = 6
  PORTA7                 = 7

  ' Servo port 3-pin header output pins (shared with Port A P4..P7)
  ' note 1: each I/O pins is buffered thru a 50 ohm resistor to minimize long transmission line effects from cables
  ' note 2: each "servo" port has a 3 pin header [signal, power, ground], each power is feed from a jumper adjacent to the
  ' servo port and has a 22uF cap across it as a current source, thus the "servo" ports aside from using the servo 3-pin cable
  ' pin outs, CAN drive servos!!! Care must be taken not to sink too much current, if you find that the board resets under harsh
  ' conditions while driving multiple simulataneous servos then you can add a bulk 2200uF cap for energy storage. There are two holes
  ' pre-drilled by the 2.1mm power entry port, simply solder a 2200uF cap under or above the board, just make sure to get the polarity
  ' correct: minus to the left (next to left edge of board), plus to the right (next to switch).
  SERVO_P4               = 4 
  SERVO_P5               = 5
  SERVO_P6               = 6
  SERVO_P7               = 7


  ' Port B pin defines (muxed with VGA port)
  PORTB0                 = 16  
  PORTB1                 = 17
  PORTB2                 = 18
  PORTB3                 = 19
  PORTB4                 = 20
  PORTB5                 = 21
  PORTB6                 = 22
  PORTB7                 = 23

  ' VGA port pin defines
  VGA_HSYNC              = 16
  VGA_VSYNC              = 17
  VGA_B0                 = 18
  VGA_B1                 = 19
  VGA_G0                 = 20
  VGA_G1                 = 21
  VGA_R0                 = 22
  VGA_R1                 = 23

  ' NTSC/PAL composite video pin defines
  COMPVID0               = 12
  COMPVID1               = 13
  COMPVID2               = 14

  ' Audio output pin define (PWM)
  AUDIO_MONO             = 24

  ' I2C pin defines (standard Prop pins)
  I2C_SCL                = 28
  I2C_SDA                = 29
 
  ' TX/RX pin defines (standard Prop pins)
  UART_TX                = 31
  UART_RX                = 30

  ' ASCII codes for ease of parser development
  ASCII_A      = 65
  ASCII_B      = 66
  ASCII_C      = 67
  ASCII_D      = 68
  ASCII_E      = 69
  ASCII_F      = 70
  ASCII_G      = 71
  ASCII_H      = 72
  ASCII_O      = 79  
  ASCII_P      = 80
  ASCII_Z      = 90
  ASCII_0      = 48
  ASCII_9      = 57
  ASCII_LEFT   = $C0
  ASCII_RIGHT  = $C1
  ASCII_UP     = $C2
  ASCII_DOWN   = $C3 
  ASCII_BS     = $C8 ' backspace
  ASCII_DEL    = $C9 ' delete
  ASCII_LF     = $0A ' line feed 
  ASCII_CR     = $0D ' carriage return
  ASCII_ESC    = $CB ' escape
  ASCII_HEX    = $24 ' $ for hex
  ASCII_BIN    = $25 ' % for binary
  ASCII_LB     = $5B ' [ 
  ASCII_SEMI   = $3A ' ; 
  ASCII_EQUALS = $3D ' = 
  ASCII_PERIOD = $2E ' .
  ASCII_COMMA  = $2C ' ,
  ASCII_SHARP  = $23 ' #
  ASCII_NULL   = $00 ' null character
  ASCII_SPACE  = $20 ' space
  ASCII_TAB    = $09 ' horizontal tab

  ' null pointer, null character
  NULL         = 0  

'//////////////////////////////////////////////////////////////////////////////
' NES GAMEPAD DEFINES /////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

{
     X      X     CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

}

  ' IO pins for NES interface 
  NES_CLK    =  5    
  NES_LATCH  =  4    
  NES_DATA0  =  3    
  NES_DATA1  =  2    
  NES_5V     =  1
  NES_GND    =  0

  NES_LATCH_DELAY  = $40

  ' button ids/bit masks
  ' NES bit encodings general for state bits
  NES_RIGHT  = %00000001
  NES_LEFT   = %00000010
  NES_DOWN   = %00000100
  NES_UP     = %00001000
  NES_START  = %00010000
  NES_SELECT = %00100000
  NES_B      = %01000000
  NES_A      = %10000000

  ' NES bit encodings for NES gamepad 0
  NES0_RIGHT  = %00000000_00000001
  NES0_LEFT   = %00000000_00000010
  NES0_DOWN   = %00000000_00000100
  NES0_UP     = %00000000_00001000
  NES0_START  = %00000000_00010000
  NES0_SELECT = %00000000_00100000
  NES0_B      = %00000000_01000000
  NES0_A      = %00000000_10000000

  ' NES bit encodings for NES gamepad 1
  NES1_RIGHT  = %00000001_00000000
  NES1_LEFT   = %00000010_00000000
  NES1_DOWN   = %00000100_00000000
  NES1_UP     = %00001000_00000000
  NES1_START  = %00010000_00000000
  NES1_SELECT = %00100000_00000000
  NES1_B      = %01000000_00000000
  NES1_A      = %10000000_00000000

    
'//////////////////////////////////////////////////////////////////////////////
' VARS SECTION ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////        
  
VAR

  ' generic globals to make demo writing easy
  long g_data, g_data2, g_status    
  long g_data32, g_data16, g_spi_result
  long g_key, g_temp1, g_temp2, g_temp3, g_index, g_length
  long g_tempx, g_tempy 
  byte g_sbuffer[256]


'//////////////////////////////////////////////////////////////////////////////
'OBJS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

' current drivers used in this version of SPI interface 
kbd        : "keyboard_010.spin"                   ' instantiate keyboard driver
gfx_ntsc   : "C3_GFX_DRV_010.SPIN"                 ' instantiate new NTSC tile driver


'//////////////////////////////////////////////////////////////////////////////
' MAIN ENTRY POINT TO DEMO/////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB Main : status

  ' let the system initialize 
  Delay_US( 1*1_000_000 )

  ' initialize the graphics tile engine
  gfx_ntsc.start

  ' delay a moment
  Delay_US(1*1_000_000)
  
  ' /////////////////////////////////////////////////////////////////////////////
  ' INITIALIZE I/O PINS /////////////////////////////////////////////////////////
  ' /////////////////////////////////////////////////////////////////////////////

  ' blink the status LED 3 times to show board is "alive", then enable VGA (leave LOW)
  DIRA[STATUS_LED_BUS_MUX] := 1 ' set to output
  OUTA[STATUS_LED_BUS_MUX] := 0

  repeat 6
    OUTA[STATUS_LED_BUS_MUX] := !OUTA[STATUS_LED_BUS_MUX]
    repeat 125_000 

  ' turn on VGA buffer
  OUTA[STATUS_LED_BUS_MUX] := 0

  ' initialize SPI IO (eventhough, we aren't using it in this demo at all, good habit to set it up)
  SPI_Init
  
  ' begin test loop

  ' clear screen
  gfx_ntsc.Out_Term( $00 )


  'start keyboard on chameleon pins, keyboard is always started even if its not plugged in
  ' we need it to read the user selection :)
  kbd.start(PS2_DATA, PS2_CLK)
   
  ' flush buffer first  
  kbd.clearkeys

  ' initialize IO for gamepad adapter
  NES_Init_Gamepad
   
  repeat
    NES_Gamepad_Demo


         
       
         
' end Main

' /////////////////////////////////////////////////////////////////////////////

PUB NES_Gamepad_Demo

  '/////////////////////////////////////////////////////////////////////////////
  ' READS NES GAME PADS ////////////////////////////////////////////////////////
  '/////////////////////////////////////////////////////////////////////////////
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("C3 NES Game Pad Demo V1.0") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string ("Insert NES Adapter into C3.") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string ("Insert NES Pads in Adapter.") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string ("Press any key to read Pads.") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string ("Press <ESC> to exit to menu.") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term

    ' initialize game pads
    NES_Init_Gamepad
    
    ' wait for keypress 
    repeat while (kbd.gotkey == FALSE)

    ' enter loop
    repeat 

      ' read the gamepads
      g_temp1 := NES_Read_Gamepad

      ' print results in binary
      gfx_ntsc.String_Term( string (" NES0[") )
      gfx_ntsc.Bin_Term( (g_temp1 ) & $FF, 8 )
      gfx_ntsc.String_Term( string ("],") )

      gfx_ntsc.String_Term( string ("NES1[") )      
      gfx_ntsc.Bin_Term( (g_temp1 >> 8) & $FF, 8 )
      gfx_ntsc.String_Term( string ("] ") )

      gfx_ntsc.Newline_Term
                  
      ' test for ESC
      if (kbd.gotkey == TRUE)
        if ((g_key := kbd.key) == ASCII_ESC)
          return

' end NES_Gamepad_Demo
        
' ////////////////////////////////////////////////////////////////////

PUB NES_Init_Gamepad

{{
Initializes the IO for the NES gamepad adapter

PARMS: None.
RETURNS: Nothing.


 DEFAULT SETUP (change the pin IO assignments if you move the inteface etc.)

     X      X     CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

  ' IO pins for NES interface 
  NES_CLK    =  5    
  NES_LATCH  =  4    
  NES_DATA0  =  3    
  NES_DATA1  =  2    
  NES_5V     =  1
  NES_GND    =  0



' NES Bit Encoding
'
' RIGHT  = %00000001
' LEFT   = %00000010
' DOWN   = %00000100
' UP     = %00001000
' START  = %00010000
' SELECT = %00100000
' B      = %01000000
' A      = %10000000

}}

' step 1: set I/Os
DIRA [ NES_CLK ]   := 1 ' output
DIRA [ NES_LATCH ] := 1 ' output
DIRA [ NES_DATA1 ] := 0 ' input
DIRA [ NES_DATA0 ] := 0 ' input

' power IOs, we are using the PROP to feed power and GND to the GAMEPAD, it uses a 4021 CMOS chip, so only a few mA are needed,
' but if the controller has a lot of LEDs, you may need to manually connect 3.3V to the +5V
' note: the NES controller WILL work just fine with 3.3V which is what we use here

DIRA [ NES_5V ]  := 1 ' output
DIRA [ NES_GND ] := 1 ' output

OUTA [ NES_5V ]  := 1 ' set to 3.3V HIGH
OUTA [ NES_GND ] := 0 ' set to 0V LOW, virtual ground

' end NES_Init_Gamepad

' ////////////////////////////////////////////////////////////////////

PUB NES_Read_Gamepad : nes_bits |  i

{{
 Reads both gamepads in parallel encodes 8-bits for each in format
 right game pad #1 [15..8] : left game pad #0 [7..0]

 NOTE: Assumes IO pins have previously been setup (call to NES_Init_Gamepad eg.)

 PARMS: None.
 RETURNS: 16-bit NES button states in the format msb[8:NES gamepad 1 | 8:NES gamepad 0]lsb,
 where the 8-bits of each game controller is outlined below in the "NES bit encoding" table.


 DEFAULT SETUP (change the pin IO assignments if you move the inteface etc.)

                 |--------------------------------------|
                 |  Parallax NES Adapter                |
                 |                                      |
     X      X     CLK   Latch  Data1  Data2  +5V     GND
 _________________________________________________________
 |  P7  |  P6  |  P5  |  P4  |  P3  |  P2  |  P1  |  P0  |  PORT A
 ---------------------------------------------------------    

  ' IO pins for NES interface 
  NES_CLK    =  5    
  NES_LATCH  =  4    
  NES_DATA0  =  3    
  NES_DATA1  =  2    
  NES_5V     =  1
  NES_GND    =  0

}}

' NES Bit Encoding
'
' RIGHT  = %00000001
' LEFT   = %00000010
' DOWN   = %00000100
' UP     = %00001000
' START  = %00010000
' SELECT = %00100000
' B      = %01000000
' A      = %10000000


' step 1: set clock and latch to 0
OUTA [ NES_CLK   ] := 0 
OUTA [ NES_LATCH ] := 0 
'Delay(1)

' step 2: set latch to 1
OUTA [ NES_LATCH ] := 1 
'Delay(1)

' step 3: set latch to 0
OUTA [ NES_LATCH ] := 0 

' step 4: read first bit of each game pad

' data is now ready to shift out
' first bit is ready 
nes_bits := 0

' left controller
nes_bits := INA[ NES_DATA0 ] | (INA[ NES_DATA1 ] << 8)

' step 5: read next 7 bits
repeat i from 0 to 6
 OUTA [ NES_CLK ] := 1
 'Delay(1)             
 OUTA [ NES_CLK ] := 0 
 nes_bits := (nes_bits << 1)
 nes_bits := nes_bits | INA[ NES_DATA0 ] | (INA[ NES_DATA1 ] << 8)

 'Delay(1)             
' invert bits to make positive logic
nes_bits := (!nes_bits & $FFFF)

' end NES_Read_Gamepad

' /////////////////////////////////////////////////////////////////////////////
' SPI ROUTINES///////// ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init


' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel


' /////////////////////////////////////////////////////////////////////////////
' UTILITY ROUTINES ////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB Get_String(string_ptr, max_length) | _length
{{
This simple function is a single line editor that allows user to enter keys from the keyboard
and then echos them to the screen, when the user hits <ENTER> | <RETURN> the function
exits and returns the string. The function has simple editing and allows <BACKSPACE> to
delete the last character, that's it!

The function outputs to the NTSC terminal

PARMS:

RETURNS: length of string, 0 if NULL.

}}

  ' current length of string buffer
  _length := 0  

  ' draw cursor
  
  repeat 

    ' draw cursor
    gfx_ntsc.Out_Term( "_" )
    gfx_ntsc.Out_Term( $08 )
  
    ' wait for keypress 
    repeat while (kbd.gotkey == FALSE)

    ' user entered a key process it

    ' get key from buffer
    g_key := kbd.key
     
    case g_key
       ASCII_LF, ASCII_CR: ' return    
 
        ' null terminate string and return
        byte [string_ptr][_length] := ASCII_NULL
     
        return( _length )


       ASCII_BS, ASCII_DEL, ASCII_LEFT: ' backspace (edit)

         if (_length > 0)
           ' move cursor back once to overwrite last character on screen
           gfx_ntsc.Out_Term( ASCII_SPACE )
           gfx_ntsc.Out_Term( $08 )          
           gfx_ntsc.Out_Term( $08 )
           
           ' echo character
           gfx_ntsc.Out_Term( ASCII_SPACE )
           gfx_ntsc.Out_Term( $08 )
         
           ' decrement length
           _length--
 
       other:    ' all other cases
         ' insert character into string 
         byte [string_ptr][_length] := g_key

         ' update length
         if (_length < max_length )
           _length++
         else
           ' move cursor back once to overwrite last character on screen
           gfx_ntsc.Out_Term( $08 )          

         ' echo character
         gfx_ntsc.Out_Term( g_key )
     
' // end Get_String


' /////////////////////////////////////////////////////////////////////////////

PUB StrCpy( dest_str_ptr, source_str_ptr ) | _index
{{
mimics C strcpy, copies source -> destination including NULL
}}

' test if there is storage
if (dest_str_ptr == NULL)
  return (NULL)

_index := 0

repeat while (byte [ source_str_ptr ][_index] <> NULL)
  ' copy next byte
   byte [ dest_str_ptr ][_index] := byte [ source_str_ptr ][_index]    
   _index++

' null terminate
byte [ dest_str_ptr ][_index] := NULL

' return number of bytes copied
return ( _index )

' // end StrCpy

' /////////////////////////////////////////////////////////////////////////////

PUB StrUpper(string_ptr)
{{
upcases a string
}}

  if (string_ptr <> NULL)
    repeat while (byte[ string_ptr ] <> NULL)
      byte[ string_ptr ] :=  ToUpper( byte[ string_ptr ] )
      string_ptr++  

  ' return string
  return (string_ptr)

' // end StrUpper

' /////////////////////////////////////////////////////////////////////////////

PUB ToUpper(ch)
{{
returns the uppercase of the sent character
}}

if (ch => $61 and ch =< $7A)
  return(ch-32)
else
  return(ch)

' // end ToUpper

' /////////////////////////////////////////////////////////////////////////////

PUB IsInSet(ch, set_string)
{{
tests if sent character is in string
}}

repeat while (byte[set_string][0] <> NULL)
  if (ch == byte[set_string++][0])
    'serial.txstring(string ("found!") )
    return( ch )

'serial.txstring(string ("not found!") )
' not found
return( -1 )    

' // end IsInSet

' /////////////////////////////////////////////////////////////////////////////

PUB IsSpace(ch)
{{
tests if sent character is white space, cr, lf, space, tab
}}

if ( (ch == ASCII_SPACE) or (ch == ASCII_LF) or (ch == ASCII_CR) or (ch == ASCII_TAB))
  return (ch)
else
  return(-1)  

' // end IsSpace

' /////////////////////////////////////////////////////////////////////////////

PUB IsNull(ch)
{{
tests if sent character is null
}}

if ( (ch == NULL))
  return (1)
else
  return(-1)  

' // end IsNull

' /////////////////////////////////////////////////////////////////////////////

PUB IsDigit(ch)
{{
tests if sent character is a number 0..9, returns integer 0..9
}}

if (ch => ASCII_0 and ch =< ASCII_9)
  return (ch-ASCII_0)
else
  return(-1)  

' // end IsDigit

' /////////////////////////////////////////////////////////////////////////////

PUB IsAlpha(ch)
{{
tests if sent character is a number a...zA...Z
}}

ch := ToUpper(ch)

if ( (ch => ASCII_A) and (ch =< ASCII_Z))
  return (ch)
else
  return(-1)  

' end IsAlpha

' /////////////////////////////////////////////////////////////////////////////

PUB IsPunc(ch)
{{
tests if sent character is a punctuation symbol !@#$%^&*()--+={}[]|\;:'",<.>/?
}}

ch := ToUpper(ch)

if (  ((ch => 33) and (ch =< 47)) or  ((ch => 58) and (ch =< 64)) or ((ch => 91) and (ch =< 96)) or ((ch =>123) and (ch =< 126))  ) 
  return (ch)
else
  return(-1)  

' // end IsPunc
                                                                                
' /////////////////////////////////////////////////////////////////////////////

PUB HexToDec(n)
{{
converts hex digit to decimal
}} 
  if ( (n => "0") and (n =< "9") )
    return (n - ASCII_0)
  elseif ( (n => "A") and (n =< "F") )
    return (n - "A" + 10)
  else
    return ( 0 )  

' // end HexToDec

' /////////////////////////////////////////////////////////////////////////////

PUB itoa(value, string_ptr) | i
{{
 converts value into a string and stores in string_ptr with NULL termination
 }}

  if value < 0
    -value
    byte [ string_ptr++ ]  := "-"

  i := 1_000_000_000

  repeat 10
    if value => i
      byte [ string_ptr++ ]  := value / i + "0"
      value //= i
      result~~
    elseif result or i == 1
      byte [ string_ptr++ ]  := "0"
    i /= 10

  ' null terminate
  byte [ string_ptr ] := NULL    

' // end itoa

' /////////////////////////////////////////////////////////////////////////////

PUB atoi2(string_ptr, length) | _index, _sum, _ch, _sign
{{
this function tries to convert the string to a number, supports binary %, hex $, decimal (default)
string_ptr can be to a null terminated string, or the length can be overridden by sending the length shorter
than the null terminator, ignores ws
Eg. %001010110, $EF, 25
}}

' initialize vars
_index := 0
_sum   := 0
_sign  := 1
 
' consume white space
repeat while (IsSpace( byte[ string_ptr ][_index] ) <> -1)
  _index++
 
' is there a +/- sign?
if (byte [string_ptr][_index] == "+")
  ' consume it
  _index++
elseif (byte [string_ptr][_index] == "-")
  ' consume it
  _index++
  _sign := -1    
     
' try to determine number base
if (byte [string_ptr][_index] == ASCII_HEX)
  _index++
  repeat while ( ( IsDigit(_ch := byte [string_ptr][_index]) <> -1) or ( IsAlpha(_ch := byte [string_ptr][_index]) <> -1)  )
    _index++
    _sum := (_sum << 4) + HexToDec( ToUpper(_ch) )
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if hex number    
elseif (byte [string_ptr][_index] == ASCII_BIN)
  repeat while ( IsDigit(_ch := byte [string_ptr][++_index]) <> -1)
    _sum := (_sum << 1) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if binary number  
else
  ' must be in default base 10, assume that
  repeat while ( IsDigit(_ch := byte [string_ptr][_index++]) <> -1)
    _sum := (_sum * 10) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
 
' else, have no idea of number format!
return(0)
 
' // end atoi2

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US

'//////////////////////////////////////////////////////////////////////////////

DAT

newline_string          byte      $0D, $0A, $00  ' $0D carriage return, $0A line feed

{{
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}
