{
/////////////////////////////////////////////////////////////////////////////

FILENAME: C3_SD_DEMO_010.SPIN

AUTHOR: Andre' LaMothe, Nurve Networks LLC
LAST MODIFIED: 9.3.10
LICENSE: MIT

I. DESCRIPTION

This demo calls the driver which contains 3 layers of software to faciliate the interfacing of SD cards,
specifically the Prop C3 SD Storage Card. However, the functions can be used with
any propeller chip interface simply by changing the SPI bus pins and modifying the CS "chip select"
line logic since the C3 uses a counter and decoder to select SPI CS lines, you must modify the function:

   SPI_Select_Channel( channel# )

   To simply assert the CS line for your particular hardware (if not the C3). Likewise,
   if you want to use another SD card driver for the C3, then all you need to do is find
   function (code) that asserts and de-asserts the CS and replace it with the C3 compatible SPI
   bus function SPI_Select_Channel( channel# ), or similar code.   
   

The 3 layers to communicate to the SD card are:

----------------------------------------------------------------------------------------
|     FAT16 (High Level - read DOS formatted FAT16 file system files on card)          |
----------------------------------------------------------------------------------------
|     SD Card Interface (Medium Level - initialize card, read, write sectors, etc.)    |
----------------------------------------------------------------------------------------
|     SPI Interface (Low Level - send and receive bytes directly to the SPI interface) |
---------------------------------------------------------------------------------------- 

And of course sitting on top of the FAT16 layer is the application itself.

The driver consists of function calls to access all 3 layers of the model, so you can do
things like access the SD card as if it were a disk drive, print the directory, load files
of any size, etc. Or you might want to forget using DOS FAT16 format, and just read and write
sectors directly to the SD card, lastly, you might want to modify the SPI routines and code
them in assembly language.

Everything is written in SPIN for ease of use and review to see what's going on and the functions
are heavily commented.

SPI Interface Links:

  http://www.vti.fi/midcom-serveattachmentguid-b469d17c67e60d4e3dbb25b0d099ad68/TN15_SPI_Interface_Specification.pdf

SD Card Links:

  http://www.cs.ucr.edu/~amitra/sdcard/ProdManualSDCardv1.9.pdf
  http://www.sandisk.com/Assets/File/OEM/Manuals/SD_SDIO_specsv1.pdf

FAT16 Links:

  http://www.maverick-os.dk/FileSystemFormats/FAT16_FileSystem.html
  http://www.compuphase.com/mbr_fat.htm


II. API OVERVIEW

The following is a short list of the API functions with single line explanations. Refer to the CON and VAR sections
below for insight into the data structures used by the API.

*** Initialization Functions ***

  PUB Start ( buffer_ptrs ) - Initializes the driver and writes the callers pointer storage space with local disk buffer addresses 

*** SPI Functions ***

  PUB SPI_Init(mode) - Initializes the hardware SPI interface I/O pins from propeller to card. 

  PUB SPI_Send_Recv_Byte(spi_data8) - Sends a single byte to SPI interface and receives one at the same time. 

  PUB SPI_Read_Byte - Reads the SPI buffer, writes a dummy values of $FF. 

  PUB SPI_Write_Byte(spi_data8) - Writes a byte to the SPI interface. 


*** SD Functions ***

  PUB SD_Mount - Mounts the SD card by initializing and placing it into SPI mode.

  PUB SD_Unmount - Unmounts the previously mounted SD card.

  PUB SD_Send_Command( command8, address32 ) - Sends a generic command to the SD card. 

  PUB SD_Read_Sector(sector32, sectorbuffer16) - Reads a sector from the SD card.

  PUB SD_Write_Sector(sector32, sectorbuffer16) - Writes a sector to the SD card.

  PUB SD_Wait_Write_Complete - Internal function used in the write command to make sure the flash has been updated.

  PUB SD_Print_Sector(sector32, sect_ptr, start_byte, end_byte, base, print_addr) - Diagnostic function that prints the contents of a sector to terminal.

  PUB SD_Read_WP - reads the single bit "write protect" signal on the SD card mechanical

  PUB SD_Read_CD - reads the single bit "card inserted" signal on the SD card mechanical


*** FAT16 Functions ***

Lower level functions (initialization of FAT16 system)

  PUB FAT_Read_MBR(mbr_ptr) - Reads the MBR (Master Boot Record) from the SD card, sector 0.

  PUB FAT_Load_Partition_Entry(partition, mbr_ptr) - Loads the requested 16-byte partition entry from the loaded MBR.

  PUB FAT_Print_Partition_Entry(partition, mbr_ptr) - Pretty prints the partition entry to terminal.

  PUB FAT_Load_Partition_Boot_Rec(pbr_ptr) - Loads the partition boot record referred to by loaded partition entry.

  PUB FAT_Print_Partition_Boot_Rec - Pretty prints the partition boot record (aka volume record or simply boot record).

High level functions (general file I/O)

  PUB FAT_Print_Directory - Print the root file directory to the terminal (like DOS DIR command).
  
  PUB FAT_File_Open(filename_ptr, file_handle_ptr) - Opens filename and fills in the sent file handle structure.

  PUB FAT_File_Close(file_handle_ptr) - Closes the file handle.

  PUB FAT_File_Read(file_handle_ptr, buffer_ptr, count) - Reads bytes from file referred to by the sent file handle.

  PUB FAT_File_Seek(file_handle_ptr, count, mode) - Seeks file pointer to a specific location in file for reading.

*** Utility Functions ***

  PUB To_ASCII(inchar, replace_char) - Used to map non-printable characters to printable. 

  PUB itoa(value, sptr) - Converts an integer to an NULL terminated ASCII string.

  PUB ToUpper(char) - Converts lower case ASCII to upper case. 

  PUB Strncomp(string_ptr1, string_ptr2, length) - Compares strings.

  PUB _Min(a,b) - returns the min.

  PUB _Max(a,b) - returns the max.

*** Exported Object Functions ***

  PUB tvt_bin( value, digits ) - Prints a binary number to the terminal with specific number of digits.
  
  PUB tvt_out( char ) - Outputs a single character to terminal.
       
  PUB tvt_dec( value ) - Prints a number in decimal format to terminal.
 
  PUB tvt_hex( value, digits ) - Prints a hex number to the terminal with specific number of digits.
  
  PUB tvt_setx( new_x ) - Sets the x cursor position on terminal.
 
  PUB tvt_sety( new_y ) - Sets the y cursor position on terminal.
  
  PUB tvt_getx - Gets the current x cursor position in terminal. 
 
  PUB tvt_gety - Gets the current y cursor position in terminal.

  PUB tvt_pstring( string_ptr ) - Prints a NULL terminated string to the terminal.

  PUB tvt_erase - Erases completely the character under the current cursor position.  

III. NOTES

1. Make sure to refer to the return value from each function, the functions have non-homgeneous
   return codes. For example, in some cases 0 is an error, in some cases it means success.

2. Many of the driver functions have sprinkled terminal print statements that are left in them
   incase you want to output more debug information. If you don't want any NTSC screen printing
   support simply remove all references to term.* calls in the code and don't include the terminal
   driver.

3. The functions are written for maximum clarity, not speed. I suggest re-writing everything
   in ASM, but these SPIN versions are reasonably easy to follow, so you can learn the ropes
   of SPI, SD, FAT16 drivers.

4. The FAT16 driver layer is minimal and doesn't write files, and only supports the root
   directory and no folders. But, that should suffice for 99% of applications.

5. Everything has been tested, but don't assume everything is working perfectly, if
   something doesn't work as expected test it thouroughly, but it could very well the
   drivers. They were only tested funtionally, but not in any large application, so
   they are not threshed out and probably have bugs here and there. But, nonetheless
   good for tutorial use and to get you started, so you dont' have to write your own!

7. Remember, ALWAYS format your SD cards FAT16, and make sure you do NOT use
   folders and all your file names are ALWAYS 8.3. Also, its suggested that you
   use 1-2G SD cards, these seem to work the best. 

//////////////////////////////////////////////////////////////////////////////
}


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON

  _clkmode = xtal1 + pll16x      ' enable external clock and pll times 8
  _xinfreq = 5_000_000   ' set frequency to 10 MHZ plus some error

'//////////////////////////////////////////////////////////////////////////////
' VARIABLES SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

VAR

  ' DRIVER SUPPORT VARS -------------------------------------------------------
  ' when using the driver, calling application must define these variables, they can have
  ' any name, but suggest using names below
  
  ' these variables will point to the sector buffers in the driver itself, when the initial call to the SD driver's Start is made
  ' the address of the first pointer @diskbuff_ptr is passed to the function, then the function fills in each pointer with
  ' the address of the SD drivers local memory buffers, so that the caller can access the buffers and not need to allocate
  ' more buffers itself. The mbr and pbr buffers are volatile and should only be read or re-loaded with appropriate
  ' driver calls, but diskbuff is temporary and can be used between calls as storage, but the driver "may" destroy
  ' the contents call to call, each buffer is a byte buffer, so when you want to access the actual bytes, you must
  ' cast to byte and use syntax like byte [diskbuff_ptr][index] etc.
  '
  'application calls ---------------------> SDMAX_DRV_010.Start with address of 1st pointer @diskbuff_ptr   
  '                                            |
  '         Application locals                 |  Driver Local buffers                         
  '           diskbuff_ptr <-------------- @ diskbuff [ ... ]
  '           mbrbuff_ptr  <-------------- @ mbrbuff  [ ... ] 
  '           pbrbuff_ptr  <-------------- @ pbrbuff  [ ... ] 
  '
  long diskbuff_ptr ' generic disk sector buffer, initialized by call to SD driver's Start
  long mbrbuff_ptr  ' master boot record buffer , initialized by call to SD driver's Start
  long pbrbuff_ptr  ' partition boot record     , initialized by call to SD driver's Start

  ' END DRIVER SUPPORT VARS ---------------------------------------------------

  ' demo program globals

  long filehandle[4] ' generic file handle first_cluster, file_length, curr_pos, directory_entry
  ' long first_cluster ' first cluster of file
  ' long filesize      ' size in bytes of file
  ' long curr_read_pos ' current read position, -1 means end of file, or anything past the size  
  '                      convention is that the pointer is always incremented then read
  '                      from, thus if curr_read_pos = 0 then the byte at location 0 has
  '                      already been read, the next byte read will be at 1
  ' long dir_entry     ' the directory index of the entry 
  
  byte filename[16]
  byte input_str[32]  

'//////////////////////////////////////////////////////////////////////////////
' OBJECT DECLARATION SECTION //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

   key   : "keyboard_010.spin"        ' instantiate a keyboard object       
   sd    : "c3_sd_drv_010.spin"       ' instantiate SD card driver object, use "sdmax_drv_nd_010.spin" version to minimize debugging output
  
'//////////////////////////////////////////////////////////////////////////////
' EXPORT PUBLICS  /////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' MAIN ENTRY POINT ////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////


PUB Main | temp1, temp2, index, index2, display_bytes, sector, bytes_read, select, mode, strlength 

  'start keyboard on pins 26,27 
  key.start(26,27)

  ' step 1: start the SD driver up, also starts up tv terminal for local printing (necessary)
  ' notice that the call MUST be made with the address of the receiver pointer area
  sd.Start(@diskbuff_ptr)

  sd.tvt_pstring(STRING("Prop C3 SD Storage Card Demo V1.0"))
  sd.tvt_out($0D)
  sd.tvt_out($0D)

  ' Step 2: set global debug level of SD driver (optional, defaults to DEBUG_BRIEF)
  sd.Set_Debug_Level (sd#DEBUG_BRIEF)

  sd.tvt_pstring(STRING("Initializing SPI interface..."))
  sd.tvt_out($0D)
  sd.tvt_out($0D)  

  ' step 3: initialize SPI interface in mode 0 (necessary)
  sd.SPI_Init(0)

  sd.tvt_pstring(STRING("Press any Key to Continue..."))
  key.newkey 
  
  ' enter main menu, card MUST be mounted with selection 1. before any operation
  repeat 

    ' set debug level off, so we don't get too much junk on screen
    sd.Set_Debug_Level (sd#DEBUG_OFF)

    ' print main menu out
    sd.tvt_out($0)
    sd.tvt_pstring(STRING("Prop C3 SD Max Storage Card Demo Menu v1.0"))
    sd.tvt_out($0D)
    sd.tvt_pstring(STRING("1. Mount/Re-mount SD Card.         "))
    sd.tvt_out($0D)    
    sd.tvt_pstring(STRING("2. Print Master Boot Record in Hex."))
    sd.tvt_out($0D)    
    sd.tvt_pstring(STRING("3. Print Partition Entry 0.        "))
    sd.tvt_out($0D)
    sd.tvt_pstring(STRING("4. Print Partition Boot Record.    "))
    sd.tvt_out($0D)
    sd.tvt_pstring(STRING("5. Directory of Files on SD Card.  "))
    sd.tvt_out($0D)
    sd.tvt_pstring(STRING("6. Load File and Print to Screen.  "))
    sd.tvt_out($0D)
    sd.tvt_pstring(STRING("7. Print Raw Sector to Screen.     "))
    sd.tvt_out($0D)
    'sd.tvt_pstring(STRING("8. Print WP, CD signals to Screen. "))
    'sd.tvt_out($0D)
    
    sd.tvt_out($0D)                    
    sd.tvt_pstring(STRING("Select?"))
    
    ' get user input with crude single line editor call
    strlength := getline(@input_str, 2)
    sd.tvt_out($0D)    

    ' convert selection to number
    select := atoi(@input_str)
    
    ' what operation is user requesting?
    case select

      1:
        ' mount the SD card, if one is already mounted, it is dismounted at the same
        ' time, this allows the user to hot swamp SD cards, and simple hit selection #1 and
        ' mount the new card

        ' clean the screen off as well
        sd.tvt_out($00)
        
        ' set debug level off, so we don't get too much junk on screen
        sd.Set_Debug_Level (sd#DEBUG_BRIEF)
      
        ' step 1: attempt to mount the SD card (necessary)
        sd.SD_Mount
         
        sd.tvt_pstring(STRING("Reading FAT16 Master Boot Record (MBR)..."))
        sd.tvt_out($0D)
         
        ' Step 2: read Master Boot Record (necessary) 
        sd.FAT_Read_MBR(mbrbuff_ptr)
         
        ' Step 3: load the values from the first partition into the globals (necessary)
        sd.FAT_Load_Partition_Entry(0, mbrbuff_ptr)
         
        sd.tvt_pstring(STRING("Reading 1st partition boot entry..."))
        sd.tvt_out($0D)
         
        ' read in the partition boot record selected previously with the loaded partition entry
        sd.FAT_Load_Partition_Boot_Rec(pbrbuff_ptr)

        ' at this point all the data structures are loaded, the Master Boot Record (MBR), the Partition Boot Sector (PBR)
        ' and the SD card is ready to go, now the application can use the low level SPI, SD, or FAT16 calls.
    
      2:
        ' print the boot sector out 64 bytes at a time and allow user to pause and hit key to continue
        repeat index from 0 to 511 step 64
          sd.SD_Print_Sector(0, diskbuff_ptr, index,index+63, 16,0)
          sd.tvt_pstring(STRING("More..."))
          key.newkey          
 
      3:
         ' print the 0th partition record entry in human readable form 
         sd.FAT_Print_Partition_Entry(0, mbrbuff_ptr)
      
      4:
         ' print the parsed partition boot record in human readable form
         sd.FAT_Print_Partition_Boot_Rec
      
      5: ' print the file directory, no pausing though, so files will scroll off screen
         sd.FAT_Print_Directory
      
      6: ' load and print a files to screen in hex, decimal, or ascii format
         sd.tvt_pstring(STRING("Filename in 8.3 format?"))
         strlength := getline(@filename, 13)
         sd.tvt_out($0D)

         sd.tvt_pstring(STRING("Print modes: 10=Dec, 16=Hex, 127=ASCII"))
         sd.tvt_out($0D)

         sd.tvt_pstring(STRING("Enter Print Mode (10,16,127)?"))         
         getline(@input_str, 6)
         mode := atoi ( @input_str )
         sd.tvt_out($0D)
           
         ' open file for reading, must do error checking here, notice the use of the filehandle passed
         ' to call, all relevant data is returned in the file handle datastructure
         if ( sd.FAT_File_Open( @filename, @filehandle) <> -1)
            sd.tvt_pstring(STRING("First Cluster = "))
            sd.tvt_dec(filehandle[0])
            sd.tvt_out($0D)
           
            sd.tvt_pstring(STRING("File Size     = "))
            sd.tvt_dec(filehandle[1])
            sd.tvt_out($0D)
                    
            sd.tvt_pstring(STRING("Current Pos   = "))
            sd.tvt_dec(filehandle[2])
            sd.tvt_out($0D)            

            sd.tvt_pstring(STRING("Directory Entry = "))
            sd.tvt_dec(filehandle[3])    
            sd.tvt_out($0D)

            sd.tvt_pstring(STRING("More..."))
            sd.tvt_out($0D)              
            key.newkey
            
            ' compute number of bytes to print out to screen depending on base
            if (mode == 10 or mode == 16)
              display_bytes := 96 
            else ' ascii mode, easier to display more data on screen
              display_bytes := 768            
      
            ' read in 64 bytes at a time and print to screen, so user can review them   

            repeat index from 0 to (filehandle[1]-1) step display_bytes
              ' read next block of bytes 
              bytes_read := sd.FAT_File_Read(@filehandle, diskbuff_ptr, display_bytes)  
           
              ' use the print sector function to pretty print the data without loading a sector
              sd.SD_Print_Sector(-1, diskbuff_ptr, 0, bytes_read-1 , mode, index)

              sd.tvt_pstring(STRING("More..."))
              sd.tvt_out($0D)

              ' test for ESC key to exit printing               
              if (key.newkey == $CB)
                quit          
         else
            sd.tvt_pstring(STRING("File "))
            sd.tvt_pstring( @filename)
            sd.tvt_pstring(STRING(" not found!"))
            sd.tvt_out($0D)    
       
      7: ' perform a low level sector read and print a sector to screen in hex, decimal, or ascii format
         sd.tvt_pstring(STRING("Sector # to Load and Print?"))
         sector := atoi ( getline(@filename, 9) )
         sd.tvt_out($0D)

         sd.tvt_pstring(STRING("Print modes: 10=Dec, 16=Hex, 127=ASCII"))
         sd.tvt_out($0D)

         sd.tvt_pstring(STRING("Enter Print Mode (10,16,127)?"))         
         getline(@input_str, 6)
         mode := atoi ( @input_str )
         sd.tvt_out($0D)

         ' read the sector into temp disk buffer
         sd.SD_Read_Sector(sector, diskbuff_ptr)         
            
         ' compute number of bytes to print out to screen depending on base
         if (mode == 10 or mode == 16)
           display_bytes := 96 
         else ' ascii mode, easier to display more data on screen
           display_bytes := 512            
         
         ' print "display" bytes at a time and print to screen, so user can review them   
         repeat index from 0 to 511 step display_bytes
           ' use the print sector function to pretty print the data without loading a sector
           ' since we already have the data loaded
           sd.SD_Print_Sector(-1, diskbuff_ptr, 0, sd._Min( display_bytes-1, 511 - index) , mode, index)
         
           sd.tvt_pstring(STRING("More..."))
           sd.tvt_out($0D)              
           if (key.newkey == $CB)
             quit          
      { ' not supported with C3
      8: ' print out the state of the Write Protect and Card Inserted signals on the SD card mechanical coming into the electrical interface
        sd.tvt_pstring(STRING("Press ESC to stop monitoring."))
        sd.tvt_out($0D)              
        repeat 500_000              

        repeat
          ' if user hits ESC exit printing              
          if (key.key == $CB)
            quit          

          sd.tvt_pstring(STRING("Write Protect="))
          sd.tvt_dec(sd.SD_Read_WP)
          sd.tvt_pstring(STRING(", Card Inserted="))
          sd.tvt_dec(sd.SD_Read_CD)
          sd.tvt_out($0D)            
      }
           
      other: ' default case, something witty from 2001
        sd.tvt_pstring(STRING("I'm sorry Dave, I can't do that..."))      
         
    ' let user see results and return to menu
    sd.tvt_out($0D)      
    sd.tvt_pstring(STRING("Press any Key to Continue..."))
    key.newkey 

  ' no operating system to return to, but if you did then always unmount the SD card
  ' if no SD is mounted, can't hurt!
  sd.SD_Unmount

' end Main

'//////////////////////////////////////////////////////////////////////////////
' FUNCTIONS  //////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////



' /////////////////////////////////////////////////////////////////////////////

PUB atoi(string_ptr) | sign, factor, number
' DESCRIPTION: converts the sent string to an integer and returns it, assumes the string is
' an integer in ASCII format, only flexibility function has is it will strip leading spaces
'
' INPUTS:
' OUTPUTS:
' -----------------------------------------------------------------------------

  sign   := 1
  number := 0

  ' if string is NULL return 0
  if (string_ptr == 0)
    return (0) 

  ' strip leading whitespace (spaces only)
  repeat while (byte[string_ptr] == $20)
    string_ptr++
    
  ' test for negative sign "-"
  if (byte[string_ptr]==$2D)
    sign := -1
    string_ptr++

  ' now convert the number
  repeat while ( (byte[string_ptr] => $30) and (byte[string_ptr] =< $39) ) 
    number := 10*number + (byte[string_ptr++] - $30)

  return ( number*sign )

' end atoi

'//////////////////////////////////////////////////////////////////////////////

PUB getline(string_ptr, maxlength) | cx, cy, ch, sindex
' DESCRIPTION:
'
' INPUTS:
' OUTPUTS:
' -----------------------------------------------------------------------------

  ' initialize single line editor
  sindex := 0
  cx := sd.tvt_getx
  cy := sd.tvt_gety

  repeat
    ' draw cursor
    sd.tvt_out( "_" )
    sd.tvt_setx( sd.tvt_getx - 1 )
  
    ' get next key from keyboard
    ch := key.getkey

    ' check for printable characters
    if ( (ch => $20) and (ch =< $7E) )
      ' check if we are at end of line
      if (sindex < maxlength-1)

        byte[string_ptr][sindex++] := ch
     
        ' erase the character and cursor with special call to draw a blank (added to terminal driver)
        sd.tvt_erase

        ' print the character 
        sd.tvt_out( ch )
        
    elseif (ch == $C8) ' backspace character
      if (sindex > 0)
        ' insert space
        byte[string_ptr][sindex--] := $20

        ' erase the character and cursor
        sd.tvt_erase
        sd.tvt_setx( sd.tvt_getx - 1 )
        sd.tvt_erase

    elseif (ch == $0D) ' enter character
      ' null terminate
      byte[string_ptr][sindex] := 0
      ' return length
      return(sindex)      
  
' end getline


'//////////////////////////////////////////////////////////////////////////////
' DATA SECTION  ///////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

DAT

{{
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}
