' /////////////////////////////////////////////////////////////////////////////
{ 
 Propeller C3 Servo Demo -  c3_servo_demo_010.spin
 AUTHOR: Andre' LaMothe
 LAST MODIFIED: 9/3/10
 VERSION 1.0
 LICENSE: MIT


 COMMENTS: This demo illustrates the use of the "servo" ports on the C3. 
  There are 4 servo compatible 3-pin ports on the C3, these are not designed to
  drive servos per se, that is the ports are designed to use the 3-wire "servo"
  connectors that lots of customers like. That said, you "can" drive servos with
  the ports as long as you don't use huge servos with a lot of current draw and
  you don't try to drive multiple servos at the same time. Each servo port has a
  22uF cap across the power feed which is used to store charge when the servo
  needs high currents, this keeps the regulators voltages from dropping too low
  and resetting the processor. Now, if you want to drive servos a lot harder you
  can either get an external driver board, or you can try some of the following
  techniques:
   
  To drive servos then a good 100-220uf or greater should be put across each power
  feed as well as shunt inductive kick diodes on the servo side as shown below:
   
         C3 Side             Servo Side            Servo
                     :
                     :
             50ohm   :                           _________
   P4..P7 --/\/\/-----------....----------------|         |
                     :                          |    __   |
   +Vcc   ------------------....----------------|   | |   |
             |       :            |     |       |   | |   |
            --- 22uF :           ---   ----     |   | |   |
            ---      :         C ---    /\  D   |   | |   |
             |       :            |    /__\     |  ( * )  |
             |       :            |     |       |         |
   GND    ------------------....----------------|_________|
                     :
                     :            C = 100-2200uF
                                  D = snubber diode, high current schottky or 1N4001 will do
   
  This little demo simply drives all four servo ports at the same, so you can hook
  up a servo to any port and it will work. You can connect the servo to the servo
  ports P4...P7 with the WHITE lead toward the top of the board (signal) and the BLACK
  lead toward the bottom (ground). Any standard servo will work such as Parallax Standard
  (Futaba servos). The protocol to control a servo is rather simple; you send a square wave
  at a frequency of 1/20ms. The duty cycle of the square wave controls the servo position;
  as you control the HIGH time from 1-2ms the servo will rotate thru its range (180). Here's
  a Parallax reference:
   
  http://www.parallax.com/Portals/0/Downloads/docs/prod/motors/900-00005StdServo-v2.1.pdf
   
  There are countless servo drivers on the Parallax object exchange, so there is no
  servo API here since "servos" are not C3 centric. This demo simply generates the
  servo control signal in SPIN. I suggest that if you do want to drive servos with the
  C3, you get a nice ASM driver and use it.
   
  In any event, the action all  occurs in the main() function, below the main() are a number of "utility"
  functions  included in each demo for reference. If you wish you can break them out into an
  external spin module. They are mostly for parsing and various random functions needed
  in each demo. The only important set of functions for the C3 are the "SPI" bus control
  functions, pay close attention to those since they control the selection of the various
  SPI devices on the bus.
   
  REQUIRES:
        (1) PS/2 keyboard plugged in and A/V port connected to NTSC TV. 
        (2) 1 or more Parallax Standard servos (or compatible). 
   
}
' /////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON

'//////////////////////////////////////////////////////////////////////////////
' PROP CHIP / SYSTEM DEFINES //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' These settings are for 5 MHZ XTALS
  _clkmode = xtal2 + pll16x         ' enable external clock and pll times 16
  _xinfreq = 5_000_000              ' set frequency to 5 MHZ

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000
  
  _stack   = 128                    ' accomodate display memory and stack

'//////////////////////////////////////////////////////////////////////////////
' C3 PIN MAP DEFINES //////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' status/debug pin also controls VGA buffer enable
  STATUS_LED_BUS_MUX     = 15 ' pin that controls VGA enable as well as a status/debug LED

  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

  ' PS/2 port pin defines
  PS2_DATA               = 26 ' PS/2 port clock pin
  PS2_CLK                = 27 ' PS/2 data pin               

  ' Port A pin defines
  PORTA0                 = 0
  PORTA1                 = 1
  PORTA2                 = 2
  PORTA3                 = 3
  PORTA4                 = 4
  PORTA5                 = 5
  PORTA6                 = 6
  PORTA7                 = 7

  ' Servo port 3-pin header output pins (shared with Port A P4..P7)
  ' note 1: each I/O pins is buffered thru a 50 ohm resistor to minimize long transmission line effects from cables
  ' note 2: each "servo" port has a 3 pin header [signal, power, ground], each power is feed from a jumper adjacent to the
  ' servo port and has a 22uF cap across it as a current source, thus the "servo" ports aside from using the servo 3-pin cable
  ' pin outs, CAN drive servos!!! Care must be taken not to sink too much current, if you find that the board resets under harsh
  ' conditions while driving multiple simulataneous servos then you can add a bulk 2200uF cap for energy storage. There are two holes
  ' pre-drilled by the 2.1mm power entry port, simply solder a 2200uF cap under or above the board, just make sure to get the polarity
  ' correct: minus to the left (next to left edge of board), plus to the right (next to switch).
  SERVO_P4               = 4 
  SERVO_P5               = 5
  SERVO_P6               = 6
  SERVO_P7               = 7


  ' Port B pin defines (muxed with VGA port)
  PORTB0                 = 16  
  PORTB1                 = 17
  PORTB2                 = 18
  PORTB3                 = 19
  PORTB4                 = 20
  PORTB5                 = 21
  PORTB6                 = 22
  PORTB7                 = 23

  ' VGA port pin defines
  VGA_HSYNC              = 16
  VGA_VSYNC              = 17
  VGA_B0                 = 18
  VGA_B1                 = 19
  VGA_G0                 = 20
  VGA_G1                 = 21
  VGA_R0                 = 22
  VGA_R1                 = 23

  ' NTSC/PAL composite video pin defines
  COMPVID0               = 12
  COMPVID1               = 13
  COMPVID2               = 14

  ' Audio output pin define (PWM)
  AUDIO_MONO             = 24

  ' I2C pin defines (standard Prop pins)
  I2C_SCL                = 28
  I2C_SDA                = 29
 
  ' TX/RX pin defines (standard Prop pins)
  UART_TX                = 31
  UART_RX                = 30

  ' ASCII codes for ease of parser development
  ASCII_A      = 65
  ASCII_B      = 66
  ASCII_C      = 67
  ASCII_D      = 68
  ASCII_E      = 69
  ASCII_F      = 70
  ASCII_G      = 71
  ASCII_H      = 72
  ASCII_O      = 79  
  ASCII_P      = 80
  ASCII_Z      = 90
  ASCII_0      = 48
  ASCII_9      = 57
  ASCII_LEFT   = $C0
  ASCII_RIGHT  = $C1
  ASCII_UP     = $C2
  ASCII_DOWN   = $C3 
  ASCII_BS     = $C8 ' backspace
  ASCII_DEL    = $C9 ' delete
  ASCII_LF     = $0A ' line feed 
  ASCII_CR     = $0D ' carriage return
  ASCII_ESC    = $CB ' escape
  ASCII_HEX    = $24 ' $ for hex
  ASCII_BIN    = $25 ' % for binary
  ASCII_LB     = $5B ' [ 
  ASCII_SEMI   = $3A ' ; 
  ASCII_EQUALS = $3D ' = 
  ASCII_PERIOD = $2E ' .
  ASCII_COMMA  = $2C ' ,
  ASCII_SHARP  = $23 ' #
  ASCII_NULL   = $00 ' null character
  ASCII_SPACE  = $20 ' space
  ASCII_TAB    = $09 ' horizontal tab

  ' null pointer, null character
  NULL         = 0  
    
'//////////////////////////////////////////////////////////////////////////////
' VARS SECTION ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////        
  
VAR

  ' generic globals to make demo writing easy
  long g_data, g_data2, g_status    
  long g_data32, g_data16, g_spi_result
  long g_key, g_temp1, g_temp2, g_temp3, g_index, g_length
  long g_tempx, g_tempy 
  byte g_sbuffer[256]


'//////////////////////////////////////////////////////////////////////////////
'OBJS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

  kbd        : "keyboard_010.spin"                   ' instantiate keyboard driver
  gfx_ntsc   : "C3_GFX_DRV_010.SPIN"                 ' instantiate new NTSC tile driver

'//////////////////////////////////////////////////////////////////////////////
' MAIN ENTRY POINT TO DEMO/////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB Main : status

  ' let the system initialize 
  Delay_US( 1*1_000_000 )

  ' initialize the graphics tile engine
  gfx_ntsc.start

  ' delay a moment
  Delay_US(1*1_000_000)

  'start keyboard on chameleon pins, keyboard is always started even if its not plugged in
  ' we need it to read the user selection :)
  kbd.start(PS2_DATA, PS2_CLK)

  ' flush buffer first  
  kbd.clearkeys
   
  ' /////////////////////////////////////////////////////////////////////////////
  ' INITIALIZE I/O PINS /////////////////////////////////////////////////////////
  ' /////////////////////////////////////////////////////////////////////////////

  ' blink the status LED 3 times to show board is "alive", then enable VGA (leave LOW)
  DIRA[STATUS_LED_BUS_MUX] := 1 ' set to output
  OUTA[STATUS_LED_BUS_MUX] := 0

  repeat 6
    OUTA[STATUS_LED_BUS_MUX] := !OUTA[STATUS_LED_BUS_MUX]
    repeat 125_000 

  ' turn on VGA buffer (and status LED)
  OUTA[STATUS_LED_BUS_MUX] := 0

  ' initialize SPI IO (eventhough, we aren't using it in this demo at all, good habit to set it up)
  SPI_Init

  ' clear screen
  gfx_ntsc.Out_Term( $00 )

  repeat 
    ' call servo demo function
    Servo_Demo  
             
' end Main

' /////////////////////////////////////////////////////////////////////////////

PUB Servo_Demo

  '/////////////////////////////////////////////////////////////////////////////
  ' "SERVO" PORT TEST SUITE ////////////////////////////////////////////////////
  '/////////////////////////////////////////////////////////////////////////////
{{
   There are 4 servo compatible 3-pin ports on the C3, these are not designed to
   drive servos per se, that is the ports are designed to use the 3-wire "servo"
   connectors that lots of customers like. That said, you "can" drive servos with
   the ports as long as you don't use huge servos with a lot of current draw and
   you don't try to drive multiple servos at the same time. Each servo port has a
   22uF cap across the power feed which is used to store charge when the servo
   needs high currents, this keeps the regulators voltages from dropping too low
   and resetting the processor. Now, if you want to drive servos a lot harder you
   can either get an external driver board, or you can try some of the following
   techniques:
   
   To drive servos then a good 100-220uf or greater should be put across each power
   feed as well as shunt inductive kick diodes on the servo side as shown below:
  
          C3 Side             Servo Side            Servo
                      :
                      :
              50ohm   :                           _________
    P4..P7 --/\/\/-----------....----------------|         |
                      :                          |    __   |
    +Vcc   ------------------....----------------|   | |   |
              |       :            |     |       |   | |   |
             --- 22uF :           ---   ----     |   | |   |
             ---      :         C ---    /\  D   |   | |   |
              |       :            |    /__\     |  ( * )  |
              |       :            |     |       |         |
    GND    ------------------....----------------|_________|
                      :
                      :            C = 100-2200uF
                                   D = snubber diode, high current schottky or 1N4001 will do
  
   This little demo simply drives all four servo ports at the same, so you can hook
   up a servo to any port and it will work. You can connect the servo to the servo
   ports P4...P7 with the WHITE lead toward the top of the board (signal) and the BLACK
   lead toward the bottom (ground). Any standard servo will work such as Parallax Standard
   (Futaba servos). The protocol to control a servo is rather simple; you send a square wave
   at a frequency of 1/20ms. The duty cycle of the square wave controls the servo position;
   as you control the HIGH time from 1-2ms the servo will rotate thru its range (180). Here's
   a Parallax reference:

   http://www.parallax.com/Portals/0/Downloads/docs/prod/motors/900-00005StdServo-v2.1.pdf
}}

  gfx_ntsc.Newline_Term
  gfx_ntsc.Newline_Term
  gfx_ntsc.String_Term( string ("C3 Servo Demo V1.0") )
  gfx_ntsc.Newline_Term
  gfx_ntsc.Newline_Term
  gfx_ntsc.String_Term( string ("Connect servo to any port...") )
  gfx_ntsc.Newline_Term
  gfx_ntsc.String_Term( string ("Use LEFT/RIGHT arrow keys to") )
  gfx_ntsc.Newline_Term
  gfx_ntsc.String_Term( string ("rotate servo.") )
  gfx_ntsc.Newline_Term
  gfx_ntsc.Newline_Term
  gfx_ntsc.String_Term( string ("Hit <ESC> key to exit.") )

  gfx_ntsc.Newline_Term
  gfx_ntsc.Newline_Term
  gfx_ntsc.String_Term( string ("Servo angular position: ") )

  g_tempx := gfx_ntsc.GetX
  g_tempy := gfx_ntsc.GetY

  g_temp1 := 200 ' starting position of servo, not to hyperrotate it
  g_temp2 := 0   

  ' update display
  gfx_ntsc.GotoXY( g_tempx, g_tempy )       
  gfx_ntsc.Dec_Term( (g_temp1 - 200) / 5  )      
  gfx_ntsc.String_Term( string ("  ") )
  
  ' set servo port IOs to outputs
  DIRA[ SERVO_P7..SERVO_P4 ] := %1111

  ' enter loop 
  repeat 

    OUTA[ SERVO_P4..SERVO_P7 ] := %1111
    waitcnt (cnt + (160*g_temp1))               ' the servo signal's high pulse duration

    OUTA[ SERVO_P4..SERVO_P7 ] := %0000
    waitcnt (cnt + 1_600_000 - (160*g_temp1))    ' the servo signal's low pulse duration

    ' user updating position?
    if (kbd.gotkey == TRUE)
      ' get key
      g_key := kbd.key
      
      ' right or left?
      if (g_key == ASCII_LEFT)

        if (g_temp1 < 1100)
          g_temp1 += 20

      elseif (g_key == ASCII_RIGHT)

        if (g_temp1 > 200)
          g_temp1 -= 20

      elseif (g_key == ASCII_ESC or g_key == ASCII_SPACE )
        ' set servo port IOs to back to inputs
        DIRA[ SERVO_P4..SERVO_P7 ] := %0000

        ' return to caller       
        return      

      ' update display
      gfx_ntsc.GotoXY( g_tempx, g_tempy )       
      gfx_ntsc.Dec_Term( (g_temp1 - 200) / 5  )      
      gfx_ntsc.String_Term( string ("  ") )

' end Servo_Test

' /////////////////////////////////////////////////////////////////////////////
' SPI ROUTINES///////// ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init


' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel


' /////////////////////////////////////////////////////////////////////////////
' UTILITY ROUTINES ////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////


PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US

'//////////////////////////////////////////////////////////////////////////////

DAT

newline_string          byte      $0D, $0A, $00  ' $0D carriage return, $0A line feed

{{
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}