' /////////////////////////////////////////////////////////////////////////////
{ 
 Propeller C3 Simple SRAM Demo -  c3_sram_demo_010.spin
 AUTHOR: Andre' LaMothe
 LAST MODIFIED: 9/3/10
 VERSION 1.0

 COMMENTS: This demo illustrates the C3's on board SRAMs which are implemented with
 (2) Microchip 23K256 Serial SPI SRAMs. The 23K256 supports a number of commands,
 we have  only implemented a couple here in these API functions (single byte read/write,
 sequential byte read/write, write status register). The little demo API is shown below
 that implements these:

 PUB SRAM_Init ( ... ) - initializes the SRAMs and sets them to sequential mode.
 PUB SRAM_Write( ... ) - writes bytes of data from a buffer to either SRAM.
 PUB SRAM_Fill( ... )  - fills either SRAM with a constant value. 
 PUB SRAM_Read( ... )  - reads data from either SRAM into a buffer.     
 PUB SRAM64K_MREAD( ... ) - models the pair of 32K SRAMs as a single memory of 64K and reads a byte from it.     
 PUB SRAM64K_MWRITE( ... ) - models the pair of 32K SRAMs as a single memory of 64K and writes a single byte to it.     
 
 However, there are many others you can implement, find out more by reviewing the chip
 specs and data sheets found here:

   http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
   http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

 The SRAMs only have 4 commands as shown below here:

 Instruction Name/Value   Description

  READ  0000 0011         Read data from memory array beginning at selected address
  WRITE 0000 0010         Write data to memory array beginning at selected address
  RDSR  0000 0101         Read STATUS register
  WRSR  0000 0001         Write STATUS register

 However, there are 3 modes of operation which are controlled via the status register
 write; byte mode, page mode, and sequential mode -- read the data sheet to learn more.

 The SRAMs are accessed via the SPI bus and are located on SPI channels 1 and 2.
 The SRAMs use a simple SPI command protocal (outlined in the data sheet above)
 that allow single and multiple byte reads and writes.    

 The action all occurs in the main() function, below the main() are a number of "utility"
 functions  included in each demo for reference. If you wish you can break them out into an
 external spin module. They are mostly for parsing and various random functions needed
 in each demo. The only important set of functions for the C3 are the "SPI" bus control
 functions, pay close attention to those since they control the selection of the various
 SPI devices on the bus. 
 
 REQUIRES: A voltage source to input into the A/D. A simple potentiometer can be used
 by connecting the outside pins to +5 and GND respectively on the Analog Port and
 the "wiper" to either A/D channel 0,1
   
}
' /////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON

'//////////////////////////////////////////////////////////////////////////////
' PROP CHIP / SYSTEM DEFINES //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' These settings are for 5 MHZ XTALS
  _clkmode = xtal1 + pll16x         ' enable external clock and pll times 16
  _xinfreq = 5_000_000              ' set frequency to 5 MHZ

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000
  
  _stack   = 128                    ' accomodate display memory and stack

'//////////////////////////////////////////////////////////////////////////////
' C3 PIN MAP DEFINES //////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' status/debug pin also controls VGA buffer enable
  STATUS_LED_BUS_MUX     = 15 ' pin that controls VGA enable as well as a status/debug LED

  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

  ' PS/2 port pin defines
  PS2_DATA               = 26 ' PS/2 port clock pin
  PS2_CLK                = 27 ' PS/2 data pin               

  ' Port A pin defines
  PORTA0                 = 0
  PORTA1                 = 1
  PORTA2                 = 2
  PORTA3                 = 3
  PORTA4                 = 4
  PORTA5                 = 5
  PORTA6                 = 6
  PORTA7                 = 7

  ' Servo port 3-pin header output pins (shared with Port A P4..P7)
  ' note 1: each I/O pins is buffered thru a 50 ohm resistor to minimize long transmission line effects from cables
  ' note 2: each "servo" port has a 3 pin header [signal, power, ground], each power is feed from a jumper adjacent to the
  ' servo port and has a 22uF cap across it as a current source, thus the "servo" ports aside from using the servo 3-pin cable
  ' pin outs, CAN drive servos!!! Care must be taken not to sink too much current, if you find that the board resets under harsh
  ' conditions while driving multiple simulataneous servos then you can add a bulk 2200uF cap for energy storage. There are two holes
  ' pre-drilled by the 2.1mm power entry port, simply solder a 2200uF cap under or above the board, just make sure to get the polarity
  ' correct: minus to the left (next to left edge of board), plus to the right (next to switch).
  SERVO_P4               = 4 
  SERVO_P5               = 5
  SERVO_P6               = 6
  SERVO_P7               = 7


  ' Port B pin defines (muxed with VGA port)
  PORTB0                 = 16  
  PORTB1                 = 17
  PORTB2                 = 18
  PORTB3                 = 19
  PORTB4                 = 20
  PORTB5                 = 21
  PORTB6                 = 22
  PORTB7                 = 23

  ' VGA port pin defines
  VGA_HSYNC              = 16
  VGA_VSYNC              = 17
  VGA_B0                 = 18
  VGA_B1                 = 19
  VGA_G0                 = 20
  VGA_G1                 = 21
  VGA_R0                 = 22
  VGA_R1                 = 23

  ' NTSC/PAL composite video pin defines
  COMPVID0               = 12
  COMPVID1               = 13
  COMPVID2               = 14

  ' Audio output pin define (PWM)
  AUDIO_MONO             = 24

  ' I2C pin defines (standard Prop pins)
  I2C_SCL                = 28
  I2C_SDA                = 29
 
  ' TX/RX pin defines (standard Prop pins)
  UART_TX                = 31
  UART_RX                = 30

  ' ASCII codes for ease of parser development
  ASCII_A      = 65
  ASCII_B      = 66
  ASCII_C      = 67
  ASCII_D      = 68
  ASCII_E      = 69
  ASCII_F      = 70
  ASCII_G      = 71
  ASCII_H      = 72
  ASCII_O      = 79  
  ASCII_P      = 80
  ASCII_Z      = 90
  ASCII_0      = 48
  ASCII_9      = 57
  ASCII_LEFT   = $C0
  ASCII_RIGHT  = $C1
  ASCII_UP     = $C2
  ASCII_DOWN   = $C3 
  ASCII_BS     = $C8 ' backspace
  ASCII_DEL    = $C9 ' delete
  ASCII_LF     = $0A ' line feed 
  ASCII_CR     = $0D ' carriage return
  ASCII_ESC    = $CB ' escape
  ASCII_HEX    = $24 ' $ for hex
  ASCII_BIN    = $25 ' % for binary
  ASCII_LB     = $5B ' [ 
  ASCII_SEMI   = $3A ' ; 
  ASCII_EQUALS = $3D ' = 
  ASCII_PERIOD = $2E ' .
  ASCII_COMMA  = $2C ' ,
  ASCII_SHARP  = $23 ' #
  ASCII_NULL   = $00 ' null character
  ASCII_SPACE  = $20 ' space
  ASCII_TAB    = $09 ' horizontal tab

  ' SRAM function helper constants
  SPI_MODE_BYTE       = %00_00000_1
  SPI_MODE_PAGE       = %10_00000_1
  SPI_MODE_SEQUENTIAL = %01_00000_1
  
  ' null pointer, null character
  NULL         = 0  
    
'//////////////////////////////////////////////////////////////////////////////
' VARS SECTION ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////        
  
VAR

  ' generic globals to make demo writing easy
  long g_data, g_data2, g_status    
  long g_data32, g_data16, g_spi_result
  long g_key, g_temp1, g_temp2, g_temp3, g_index, g_length
  long g_tempx, g_tempy 
  byte g_sbuffer[256]

'//////////////////////////////////////////////////////////////////////////////
'OBJS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

  ' current drivers used in this version of SPI interface 
  kbd        : "keyboard_010.spin"                   ' instantiate keyboard driver
  gfx_ntsc   : "C3_GFX_DRV_010.SPIN"                 ' instantiate new NTSC tile driver

DAT

  sram_buffer byte 1,4,9,16,25,36,49,64,81,100,121,144,169,196,225

'//////////////////////////////////////////////////////////////////////////////
' MAIN ENTRY POINT TO DEMO/////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB Main : status

  ' let the system initialize 
  Delay_US( 1*1_000_000 )

  ' initialize the graphics tile engine
  gfx_ntsc.start

  ' delay a moment
  Delay_US(1*1_000_000)
  
  ' /////////////////////////////////////////////////////////////////////////////
  ' INITIALIZE I/O PINS /////////////////////////////////////////////////////////
  ' /////////////////////////////////////////////////////////////////////////////

  ' blink the status LED 3 times to show board is "alive", then enable VGA (leave LOW)
  DIRA[STATUS_LED_BUS_MUX] := 1 ' set to output
  OUTA[STATUS_LED_BUS_MUX] := 0

  repeat 6
    OUTA[STATUS_LED_BUS_MUX] := !OUTA[STATUS_LED_BUS_MUX]
    repeat 125_000 

  ' turn on VGA buffer
  OUTA[STATUS_LED_BUS_MUX] := 0

  ' initialize SPI IO since the A/D is on the SPI
  SPI_Init

  'start keyboard on chameleon pins, keyboard is always started even if its not plugged in
  ' we need it to read the user selection :)
  kbd.start(PS2_DATA, PS2_CLK)
   
  ' flush buffer first  
  kbd.clearkeys

  ' initialize the SRAMs into sequential access mode    
  SRAM_Init( SPI_MODE_SEQUENTIAL )

  ' enter infinite test loop
  repeat
    ' clear screen
    gfx_ntsc.Out_Term( $00 )

    ' call the test
    SRAM_Test
     
' end Main

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Test

  '/////////////////////////////////////////////////////////////////////////////
  ' SRAM TEST SUITE ////////////////////////////////////////////////////////////
  '/////////////////////////////////////////////////////////////////////////////


  repeat
    ' draw menu
    gfx_ntsc.Newline_Term  
    gfx_ntsc.String_Term( string ("C3 SRAM Memory Test Menu V1.0") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("1.Read Status Reg Bank0 (B0).") )
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("2.Read Status Reg Bank1 (B1).") )
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("3.Write B0[start,val,numbytes].") )
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("4.Read B0[start,numbytes].") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string ("5.Write B1[start,val,numbytes].") )
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("6.Read B1[start,numbytes].") )
    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string ("7.Write single byte[0..64K]") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.String_Term( string ("8.Read single byte[0..64K]") )
    gfx_ntsc.Newline_Term

    gfx_ntsc.String_Term( string ("9.Exit back to main menu.") )
    gfx_ntsc.Newline_Term
    gfx_ntsc.Newline_Term    
    gfx_ntsc.String_Term( string ("Selection?") )

    ' get string from user
    Get_String(@g_sbuffer, 4 )

    ' convert to integer
    g_key := atoi2(@g_sbuffer, 4)

   ' what is user requesting?
   case g_key
     1:
       SPI_Select_Channel( SPI_CHAN_SRAM0 )
       g_temp1 := SPI_Write_Read( 16, %00000101_00000000, $FF )
       SPI_Select_Channel( 0 )
      
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Status Register Bank0 = ") )
       gfx_ntsc.Dec_Term( g_temp1 )

     2:
       SPI_Select_Channel( SPI_CHAN_SRAM1 )
       g_temp1 := SPI_Write_Read( 16, %00000101_00000000, $FF )
       SPI_Select_Channel( 0 )
      
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Status Register Bank1 = ") )
       gfx_ntsc.Dec_Term( g_temp1 )

     3:
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Start address?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Value to write?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp2 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Number of bytes to write?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp3 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term  
       gfx_ntsc.String_Term( string ("Writing Bank 0...") )

       ' write the bytes
       SRAM_Fill( 0, g_temp1, g_temp3, g_temp2 )
   
       gfx_ntsc.Newline_Term
       gfx_ntsc.Dec_Term ( g_temp3 )
       gfx_ntsc.String_Term( string (" Bytes written.") )

      
     4:

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Start address?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Number of bytes to read?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp3 := atoi2(@g_sbuffer, 8)

      ' print buffer to screen
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Bytes read:") )

       repeat g_index from g_temp1 to (g_temp1 + g_temp3 - 1)
         ' print starting address to left every 8 bytes
         if ((g_index // 8) == 0)
           gfx_ntsc.Newline_Term
           gfx_ntsc.Hex_Term (g_index, 6 )
           gfx_ntsc.String_Term( string (":") )           

         ' read a single byte at a time
         SRAM_Read(0, g_index, 1, @g_sbuffer[0]) 

         ' print byte
         gfx_ntsc.Hex_Term ( g_sbuffer[0], 2 )
         gfx_ntsc.Out_Term( "," )             

     5:
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Start address?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Value to write?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp2 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Number of bytes to write?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp3 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term  
       gfx_ntsc.String_Term( string ("Writing Bank 1...") )

       ' write the bytes
       SRAM_Fill( 1, g_temp1, g_temp3, g_temp2 )
   
       gfx_ntsc.Newline_Term
       gfx_ntsc.Dec_Term ( g_temp3 )
       gfx_ntsc.String_Term( string (" Bytes written.") )

      
     6:

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Start address?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Number of bytes to read?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp3 := atoi2(@g_sbuffer, 8)

      ' print buffer to screen
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Bytes read:") )
       
       repeat g_index from g_temp1 to (g_temp1 + g_temp3 - 1)
         ' print starting address to left every 8 bytes
         if ((g_index // 8) == 0)
           gfx_ntsc.Newline_Term
           gfx_ntsc.Hex_Term (g_index, 6 )
           gfx_ntsc.String_Term( string (":") )           

         ' read a single byte at a time
         SRAM_Read(1, g_index, 1, @g_sbuffer[0]) 

         ' print byte
         gfx_ntsc.Hex_Term ( g_sbuffer[0], 2 )
         gfx_ntsc.Out_Term( "," )             

     7:

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Write Addr [0..64K]?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Value to write?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp2 := atoi2(@g_sbuffer, 8)

       SRAM64K_MWRITE( g_temp1, g_temp2 )

     8:

       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string ("Read Addr [0..64K]?") )

       ' get string from user
       Get_String(@g_sbuffer, 8 )

       ' convert to integer
       g_temp1 := atoi2(@g_sbuffer, 8)

       gfx_ntsc.Newline_Term
       gfx_ntsc.Newline_Term    
       gfx_ntsc.String_Term( string (" SRAM[ ") )
       gfx_ntsc.Dec_Term( g_temp1 )
       gfx_ntsc.String_Term( string (" ] = ") )       

       ' read byte from SRAM
       g_temp2 := SRAM64K_MREAD( g_temp1 )
       gfx_ntsc.Dec_Term( g_temp2 )
       gfx_ntsc.Newline_Term

     ' return to caller
     9: return

  ' end SRAM_Test

' /////////////////////////////////////////////////////////////////////////////
' SRAM API ////////////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Init ( mode ) | _spi_word
{{
This function initializes both banks of the Microchip 23K256 SRAMs
serial SPI SRAMs'. The initialization simply sets the STATUS register
of each SRAM into "byte", "page" or "sequential" read/write mode to allow more
efficient access of the SRAMs. The SRAM API currently relies on "sequential"
mode, so if yoo want to use them, this function should be called with SPI_MODE_SEQUENTIAL

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  mode - %00_00000_1 = SPI_MODE_BYTE
               %10_00000_1 = SPI_MODE_PAGE
               %01_00000_1 = SPI_MODE_SEQUENTIAL

RETURNS: Nothing. 

}}

  ' first set SRAM bank 0 to sequential access mode (see page 6 of data sheet)
  SPI_Select_Channel( SPI_CHAN_SRAM0 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' now select SRAM bank 1
  SPI_Select_Channel( SPI_CHAN_SRAM1 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' and finally de-select all SPI devices
  SPI_Select_Channel( 0 )

' end SRAM_Init

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Write( bank, addr, num_bytes, src_buffer_ptr) | _index, _data, _spi_word
{{
This function writes a number of bytes to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        src_buffer_ptr - pointer to the buffer of bytes to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to write?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write sequential byte:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' and now send command to SRAM for sequential writing
    SPI_Write_Read( 24, _spi_word, $FF )
    
    ' write the bytes into the SRAM
    repeat _index from 0 to num_bytes-1

      ' get byte to write
      _data := byte[ src_buffer_ptr][ _index ]
       
      ' write the byte
      SPI_Write_Read( 8, _data, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Write

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Fill( bank, addr, num_bytes, value) | _index, _data, _spi_word
{{
This function writes a constant value to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        value          - byte value to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to fill?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write byte sequential:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' setup for sequential byte write, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to receive bytes (as many as we want to send)
    repeat _index from 0 to num_bytes-1
      SPI_Write_Read( 8, value, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Fill

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Read(bank, addr, num_bytes, dest_buffer_ptr) | _index, _data, _spi_word
{{

This function reads a number of bytes from either bank of the SRAM into a buffer.

The 23K256 supports a number of commands, we have only implemented a couple here
in these API functions (single byte read/write, multiple byte read/write).
However, there are many others you can implement, find out more by reviewing
the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  bank            - [0,1] the SRAM bank to read from, there are two banks, each holds 32K bytes
        addr            - address in SRAM to read from [0..32767]
        num_bytes       - number of bytes to read from SRAM into destination buffer
        dest_buffer_ptr - pointer to the buffer of bytes to read SRAM into

RETURNS: number of bytes read 

}}

  ' test for anything to read?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( read byte sequential:8 | address:16 )
    _spi_word := (%00000011 << 16) + (addr)

    ' setup for sequential byte read, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to be read from
    repeat _index from 0 to num_bytes-1

      ' read next byte from SRAM into receiving buffer
      byte [dest_buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )
    
    ' return bytes read
    return( num_bytes )
  else
    ' catch error
    return 0 

' end SRAM_Read

' /////////////////////////////////////////////////////////////////////////////                                                                               

PUB SRAM64K_MREAD( addr ) | _spi_word, _data
{{
This function reads a single byte from either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Read function
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
    
RETURNS: The 8-bit data from the SRAM. 

}}

    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( read byte sequential:8 | address:16 | dummy:8 )
    _spi_word := (%00000011 << 24) + (addr << 8) + 0 

    ' read the byte of data 
    _data := SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return data
    return ( _data )
    
' end SRAM64K_READ

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM64K_MWRITE( addr, data ) | _spi_word
{{
This function writes a single byte to either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Write|Fill functions
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
        data - the byte to write.    

RETURNS: Nothing. 

}}

    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( write sequential byte:8 | address:16 | data: 8)
    _spi_word := (%00000010 << 24) + ( (addr & $7FFF) << 8) + data

    ' write the byte of data 
    SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

' end SRAM64K_WRITE

' /////////////////////////////////////////////////////////////////////////////
' SPI ROUTINES///////// ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel


' /////////////////////////////////////////////////////////////////////////////
' UTILITY ROUTINES ////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB Get_String(string_ptr, max_length) | _length
{{
This simple function is a single line editor that allows user to enter keys from the keyboard
and then echos them to the screen, when the user hits <ENTER> | <RETURN> the function
exits and returns the string. The function has simple editing and allows <BACKSPACE> to
delete the last character, that's it!

The function outputs to the NTSC terminal

PARMS:

RETURNS: length of string, 0 if NULL.

}}

  ' current length of string buffer
  _length := 0  

  ' draw cursor
  
  repeat 

    ' draw cursor
    gfx_ntsc.Out_Term( "_" )
    gfx_ntsc.Out_Term( $08 )
  
    ' wait for keypress 
    repeat while (kbd.gotkey == FALSE)

    ' user entered a key process it

    ' get key from buffer
    g_key := kbd.key
     
    case g_key
       ASCII_LF, ASCII_CR: ' return    
 
        ' null terminate string and return
        byte [string_ptr][_length] := ASCII_NULL
     
        return( _length )


       ASCII_BS, ASCII_DEL, ASCII_LEFT: ' backspace (edit)

         if (_length > 0)
           ' move cursor back once to overwrite last character on screen
           gfx_ntsc.Out_Term( ASCII_SPACE )
           gfx_ntsc.Out_Term( $08 )          
           gfx_ntsc.Out_Term( $08 )
           
           ' echo character
           gfx_ntsc.Out_Term( ASCII_SPACE )
           gfx_ntsc.Out_Term( $08 )
         
           ' decrement length
           _length--
 
       other:    ' all other cases
         ' insert character into string 
         byte [string_ptr][_length] := g_key

         ' update length
         if (_length < max_length )
           _length++
         else
           ' move cursor back once to overwrite last character on screen
           gfx_ntsc.Out_Term( $08 )          

         ' echo character
         gfx_ntsc.Out_Term( g_key )
     
' // end Get_String


' /////////////////////////////////////////////////////////////////////////////

PUB StrCpy( dest_str_ptr, source_str_ptr ) | _index
{{
mimics C strcpy, copies source -> destination including NULL
}}

' test if there is storage
if (dest_str_ptr == NULL)
  return (NULL)

_index := 0

repeat while (byte [ source_str_ptr ][_index] <> NULL)
  ' copy next byte
   byte [ dest_str_ptr ][_index] := byte [ source_str_ptr ][_index]    
   _index++

' null terminate
byte [ dest_str_ptr ][_index] := NULL

' return number of bytes copied
return ( _index )

' // end StrCpy

' /////////////////////////////////////////////////////////////////////////////

PUB StrUpper(string_ptr)
{{
upcases a string
}}

  if (string_ptr <> NULL)
    repeat while (byte[ string_ptr ] <> NULL)
      byte[ string_ptr ] :=  ToUpper( byte[ string_ptr ] )
      string_ptr++  

  ' return string
  return (string_ptr)

' // end StrUpper

' /////////////////////////////////////////////////////////////////////////////

PUB ToUpper(ch)
{{
returns the uppercase of the sent character
}}

if (ch => $61 and ch =< $7A)
  return(ch-32)
else
  return(ch)

' // end ToUpper

' /////////////////////////////////////////////////////////////////////////////

PUB IsInSet(ch, set_string)
{{
tests if sent character is in string
}}

repeat while (byte[set_string][0] <> NULL)
  if (ch == byte[set_string++][0])
    'serial.txstring(string ("found!") )
    return( ch )

'serial.txstring(string ("not found!") )
' not found
return( -1 )    

' // end IsInSet

' /////////////////////////////////////////////////////////////////////////////

PUB IsSpace(ch)
{{
tests if sent character is white space, cr, lf, space, tab
}}

if ( (ch == ASCII_SPACE) or (ch == ASCII_LF) or (ch == ASCII_CR) or (ch == ASCII_TAB))
  return (ch)
else
  return(-1)  

' // end IsSpace

' /////////////////////////////////////////////////////////////////////////////

PUB IsNull(ch)
{{
tests if sent character is null
}}

if ( (ch == NULL))
  return (1)
else
  return(-1)  

' // end IsNull

' /////////////////////////////////////////////////////////////////////////////

PUB IsDigit(ch)
{{
tests if sent character is a number 0..9, returns integer 0..9
}}

if (ch => ASCII_0 and ch =< ASCII_9)
  return (ch-ASCII_0)
else
  return(-1)  

' // end IsDigit

' /////////////////////////////////////////////////////////////////////////////

PUB IsAlpha(ch)
{{
tests if sent character is a number a...zA...Z
}}

ch := ToUpper(ch)

if ( (ch => ASCII_A) and (ch =< ASCII_Z))
  return (ch)
else
  return(-1)  

' end IsAlpha

' /////////////////////////////////////////////////////////////////////////////

PUB IsPunc(ch)
{{
tests if sent character is a punctuation symbol !@#$%^&*()--+={}[]|\;:'",<.>/?
}}

ch := ToUpper(ch)

if (  ((ch => 33) and (ch =< 47)) or  ((ch => 58) and (ch =< 64)) or ((ch => 91) and (ch =< 96)) or ((ch =>123) and (ch =< 126))  ) 
  return (ch)
else
  return(-1)  

' // end IsPunc
                                                                                
' /////////////////////////////////////////////////////////////////////////////

PUB HexToDec(n)
{{
converts hex digit to decimal
}} 
  if ( (n => "0") and (n =< "9") )
    return (n - ASCII_0)
  elseif ( (n => "A") and (n =< "F") )
    return (n - "A" + 10)
  else
    return ( 0 )  

' // end HexToDec

' /////////////////////////////////////////////////////////////////////////////

PUB itoa(value, string_ptr) | i
{{
 converts value into a string and stores in string_ptr with NULL termination
 }}

  if value < 0
    -value
    byte [ string_ptr++ ]  := "-"

  i := 1_000_000_000

  repeat 10
    if value => i
      byte [ string_ptr++ ]  := value / i + "0"
      value //= i
      result~~
    elseif result or i == 1
      byte [ string_ptr++ ]  := "0"
    i /= 10

  ' null terminate
  byte [ string_ptr ] := NULL    

' // end itoa

' /////////////////////////////////////////////////////////////////////////////

PUB atoi2(string_ptr, length) | _index, _sum, _ch, _sign
{{
this function tries to convert the string to a number, supports binary %, hex $, decimal (default)
string_ptr can be to a null terminated string, or the length can be overridden by sending the length shorter
than the null terminator, ignores ws
Eg. %001010110, $EF, 25
}}

' initialize vars
_index := 0
_sum   := 0
_sign  := 1
 
' consume white space
repeat while (IsSpace( byte[ string_ptr ][_index] ) <> -1)
  _index++
 
' is there a +/- sign?
if (byte [string_ptr][_index] == "+")
  ' consume it
  _index++
elseif (byte [string_ptr][_index] == "-")
  ' consume it
  _index++
  _sign := -1    
     
' try to determine number base
if (byte [string_ptr][_index] == ASCII_HEX)
  _index++
  repeat while ( ( IsDigit(_ch := byte [string_ptr][_index]) <> -1) or ( IsAlpha(_ch := byte [string_ptr][_index]) <> -1)  )
    _index++
    _sum := (_sum << 4) + HexToDec( ToUpper(_ch) )
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if hex number    
elseif (byte [string_ptr][_index] == ASCII_BIN)
  repeat while ( IsDigit(_ch := byte [string_ptr][++_index]) <> -1)
    _sum := (_sum << 1) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if binary number  
else
  ' must be in default base 10, assume that
  repeat while ( IsDigit(_ch := byte [string_ptr][_index++]) <> -1)
    _sum := (_sum * 10) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
 
' else, have no idea of number format!
return(0)
 
' // end atoi2

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US

'//////////////////////////////////////////////////////////////////////////////

DAT

newline_string          byte      $0D, $0A, $00  ' $0D carriage return, $0A line feed