' /////////////////////////////////////////////////////////////////////////////
{
 C3_GFX_DRV_010.SPIN 
 VERSION: 1.0
 AUTHOR: Andre' LaMothe

 LAST MODIFIED: 7.20.10

 COMMENTS: Modified from Chameleon to be used in "stand alone" mode for the C3.
 
 Specs:
 - 32x24 tile map, each entry 2 bytes, format: [palette index:tile index]
 - 8x8 bitmap tiles, 4 colors each, array of 8 WORDs, each WORD represents 8 pixels, 16 bytes each tile, 4 LONGs
 - Course scrolling, horizontal and vertical, playfield width 32, 64, 128, 256
 - Fine scrolling on vertical
  
 Features:
 - Multiple tile maps 256
 - Multiple palettes 256
 - Bitmaps 256
 - All assets base pointers passed to engine each frame, thus "on the fly" changes can be made
 - Supports page flipping, double buffering by design.
 - Engine passes back vertical region and hsyn/vsync status as well as current line
}
' /////////////////////////////////////////////////////////////////////////////


'//////////////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ///////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

CON


  _clkmode = xtal1 + pll16x     ' enable external clock range 5/10MHz and pll times 8/16
  _xinfreq = 5_000_000          ' set frequency to 5/10 MHZ plus some error due to XTAL (1000-5000 usually works)
  _stack   = 256                ' accomodate display memory and stack




' GPU GFX command constants

GPU_GFX_BASE_ID         = 192                  ' starting id for GFX commands to keep them away from normal command set
GPU_GFX_NUM_COMMANDS    = 37                   ' number of GFX commands

GPU_GFX_SUBFUNC_STATUS_R = (0+GPU_GFX_BASE_ID) ' Reads the status of the GPU, Writes the GPU Sub-Function register and issues a high level command like copy, fill, etc.                                                      
GPU_GFX_SUBFUNC_STATUS_W = (1+GPU_GFX_BASE_ID) ' Writes status of the GPU, Writes the GPU Sub-Function register and issues a high level command like copy, fill, etc.

  ' sub-function constants that are executed when the GPU_GFX_SUBFUNC_STATUS_W command is issued 
  GPU_GFX_SUBFUNC_COPYMEM16 = 0 ' Copies numbytes from src -> dest in wordsize chunks
  GPU_GFX_SUBFUNC_FILLMEM16 = 1 ' Fills memory with data16, 2 bytes at a time

  GPU_GFX_SUBFUNC_COPYMEM8  = 2 ' Copies numbytes from src -> dest in byte size chunks
  GPU_GFX_SUBFUNC_FILLMEM8  = 3 ' Fills memory with low byte of data16, 1 bytes at a time 

' normal commands 

GPU_GFX_TILE_MAP_R = (2+GPU_GFX_BASE_ID) ' Reads 16-bit tile map ptr which points to the current tile map displayed.
GPU_GFX_TILE_MAP_W = (3+GPU_GFX_BASE_ID) ' Writes 16-bit tile map ptr which points to the current tile map displayed.                                                                         

GPU_GFX_DISPLAY_PTR_R = (4+GPU_GFX_BASE_ID) ' Reads 16-bit tile map ptr which points to the current display ptr into the tile map, low level print functions use this pointer.
GPU_GFX_DISPLAY_PTR_W = (5+GPU_GFX_BASE_ID) ' Writes 16-bit tile map ptr which points to the current display ptr into the tile map, low level print functions use this pointer.                                                                

GPU_GFX_TERM_PTR_R = (6+GPU_GFX_BASE_ID) ' Reads 16-bit tile map ptr which points to the current terminal ptr into the tile map, terminal level print functions use this pointer.
GPU_GFX_TERM_PTR_W = (7+GPU_GFX_BASE_ID) ' Writes 16-bit tile map ptr which points to the current terminal ptr into the tile map, terminal level print functions use this pointer.                                                                

GPU_GFX_BITMAP_R   = (8+GPU_GFX_BASE_ID) ' Reads 16-bit tile bitmaps ptr which points to the bitmaps indexed by the tilemap. 
GPU_GFX_BITMAP_W   = (9+GPU_GFX_BASE_ID) ' Reads 16-bit tile bitmaps ptr which points to the bitmaps indexed by the tilemap.                                                                         

GPU_GFX_PALETTE_R  = (10+GPU_GFX_BASE_ID) ' Reads 16-bit palette array ptr which points to the palettes in use for the tilemap.
GPU_GFX_PALETTE_W  = (11+GPU_GFX_BASE_ID) ' Writes 16-bit palette array ptr which points to the palettes in use for the tilemap.                                                                         

GPU_GFX_TOP_OVERSCAN_COL_R = (12+GPU_GFX_BASE_ID) ' Reads top overscan color drawn on the screen.
GPU_GFX_TOP_OVERSCAN_COL_W = (13+GPU_GFX_BASE_ID) ' Writes top overscan color drawn on the screen.

GPU_GFX_BOT_OVERSCAN_COL_R = (14+GPU_GFX_BASE_ID) ' Reads top overscan color drawn on the screen.
GPU_GFX_BOT_OVERSCAN_COL_W = (15+GPU_GFX_BASE_ID) ' Writes top overscan color drawn on the screen.

GPU_GFX_HSCROLL_FINE_R  = (16+GPU_GFX_BASE_ID)  ' Reads current fine horizontal scroll register (0..7) NOTE: (functionality NOT implemented yet) 
GPU_GFX_HSCROLL_FINE_W  = (17+GPU_GFX_BASE_ID)  ' Writes current fine horizontal scroll register (0..7) NOTE: (functionality NOT implemented yet)

GPU_GFX_VSCROLL_FINE_R  = (18+GPU_GFX_BASE_ID) ' Reads current fine vertical scroll register (0..7)
GPU_GFX_VSCROLL_FINE_W  = (19+GPU_GFX_BASE_ID) ' Writes current fine vertical scroll register (0..7)

GPU_GFX_SCREEN_WIDTH_R  = (20+GPU_GFX_BASE_ID)  ' Reads screen width value, 0=16 tiles, 1=32 tiles, 2=64 tiles, etc.
GPU_GFX_SCREEN_WIDTH_W  = (21+GPU_GFX_BASE_ID)  ' Writes screen width value, 0=16 tiles, 1=32 tiles, 2=64 tiles, etc.

GPU_GFX_SRC_ADDR_R      = (22+GPU_GFX_BASE_ID) ' Reads 16-bit source address for GPU operations.
GPU_GFX_SRC_ADDR_W      = (23+GPU_GFX_BASE_ID) ' Writes 16-bit source address for GPU operations.
                     
GPU_GFX_DEST_ADDR_R     = (24+GPU_GFX_BASE_ID) ' Reads 16-bit destination address for GPU operations.
GPU_GFX_DEST_ADDR_W     = (25+GPU_GFX_BASE_ID) ' Writes 16-bit destination address for GPU operations.

GPU_GFX_NUM_BYTES_R     = (26+GPU_GFX_BASE_ID) ' Reads 16-bit number representing the number of bytes for a GPU operation Sub-Function.
GPU_GFX_NUM_BYTES_W     = (27+GPU_GFX_BASE_ID) ' Writes 16-bit number representing the number of bytes for a GPU operation Sub-Function.

GPU_GFX_DATA_R          = (28+GPU_GFX_BASE_ID) ' Reads 16-bit data word uses for GPU operations or memory access operations.
GPU_GFX_DATA_W          = (29+GPU_GFX_BASE_ID) ' Writes 16-bit data word uses for GPU operations or memory access operations.

GPU_GFX_RAM_PORT8_R     = (30+GPU_GFX_BASE_ID) ' Reads 8-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                         ' After the operation, the src_addr_ptr is incremented as per the GPU configuration.

GPU_GFX_RAM_PORT8_W     = (31+GPU_GFX_BASE_ID) ' Writes 8-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                         ' After the operation, the src_add_rptr are incremented as per the GPU configuration.

GPU_GFX_RAM_PORT16_R    = (32+GPU_GFX_BASE_ID) ' Reads 16-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                         ' After the operation, the src_addr_ptr is incremented as per the GPU configuration.

GPU_GFX_RAM_PORT16_W    = (33+GPU_GFX_BASE_ID) ' Writes 16-bit data pointed to by the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                         ' After the operation, the src_add_rptr are incremented as per the GPU configuration.

GPU_GFX_RASTER_LINE_R   = (34+GPU_GFX_BASE_ID) ' Reads the current raster line being drawn from 0..191, or whatever the GPU's line resolution is set to.                             

' gpu configuration registers (currently these refer to positive increments only, but later we could make them signed and large strides if needed)
GPU_GFX_SET_AUTO_INC_R    = (35+GPU_GFX_BASE_ID) ' Reads current memory auto byte increment setting for read/write operations lower 4-bits (0..15), default 0
GPU_GFX_SET_AUTO_INC_W    = (36+GPU_GFX_BASE_ID) ' Writes the current memory auto byte increment setting for read/write operations lower 4-bits (0..15), default 0

'//////////////////////////////////////////////////////////////////////////////
' VARS SECTION ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

VAR

long cog, cogon                 ' tracks id of cog and if start was successful

' begin parameter list ////////////////////////////////////////////////////////
' tile engine data structure pointers (can be changed in real-time by app!, ASM driver is monitoring these)
long tile_map_base_ptr_parm        ' points to base of tile map memory
long tile_bitmaps_base_ptr_parm    ' points to base of tile bitmap memory
long tile_palettes_base_ptr_parm   ' points to base of palette record memory
long tile_map_control_parm         ' variable that encodes various engine controls and state


' real-time engine status variables, these are updated in real time by the
' tile engine itself, so they can be monitored outside in SPIN/ASM by game
long tile_status_bits_parm      ' vsync, hsync, etc.

' format of tile_status_bits_parm, only the Vsync status bit is updated
'
' byte 3 (unused)|byte 2 (line)|   byte 1 (tile postion)    |                     byte 0 (sync and region)      |
'|x x x x x x x x| line 8-bits | row 4 bits | column 4-bits |x x x x | region 2-bits | hsync 1-bit | vsync 1-bit|
'   b31..b24         b23..b16      b15..b12     b11..b8                    b3..b2          b1            b0
' Region 0=Top Overscan, 1=Active Video, 2=Bottom Overscan, 3=Vsync
' NOTE: In this version of the tile engine only VSYNC and REGION are valid 

' end parameter list ///////////////////////////////////////////////////////////

' new global shared parameters for GPU
long tile_src_addr_parm         ' points to source data for GPU operations
long tile_dest_addr_parm        ' points to destination data for GPU operations
long tile_numbytes_parm         ' number of bytes for various GPU operations
long tile_data16_parm           ' 16-bit data word for various operations
long tile_ramport8_parm         ' 8-bit rampport for accessing GPU side memory                        
long tile_ramport16_parm        ' 16-bit ramport for accessing GPU side memory
long tile_rasterline_parm       ' current raster line
long tile_mem_autoinc_parm      ' memory operations "auto increment" values
long tile_pitch                 ' global pitch of current tile map displayed
long tile_display_ptr           ' this points to the "direct mode" tile map memory that you want to print into
long tile_term_ptr              ' this points to the "terminal mode" tile map memory that you want to print into

long GPU_GFX_subfunc, GPU_GFX_data ' used to hold GPU sub-function commands extracted from data


' terminal vars here
long  term_col, term_row, term_num_cols, term_num_rows, term_attr, term_color, term_flag,  term_lastrow


'//////////////////////////////////////////////////////////////////////////////
'OBJS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

OBJ

'//////////////////////////////////////////////////////////////////////////////
'PUBS SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////


' COG INTERPRETER STARTS HERE...@ THE FIRST PUB

PUB Start | index_x, index_y, ch
' This is the first entry point the system will see when the PChip starts,
' execution ALWAYS starts on the first PUB in the source code for
' the top level file

' points ptrs to actual memory storage for tile engine
tile_map_base_ptr_parm      := @tile_map0
tile_bitmaps_base_ptr_parm  := @_tile_bitmaps
tile_palettes_base_ptr_parm := @_tile_palette_map

tile_map_control_parm       := $ED_ED_00_00 ' atari / c64 style colors!!!!   

' tile_map_control_parm       := $AC_AC_00_00 ' format $bc_tc_hv_ww, tc = top overscan color, bc= bottom overscan color,
                                            ' h=horizontal scroll (0..7), v=vertical scroll (0..7),
                                            ' ww = width: 0 = 32 tiles, 1 = 64 tiles, 2 = 128 tiles, 3 = 256 tiles, etc.
                                            ' green overscan with $AC
tile_status_bits_parm       := 0

' initialize the display page(s) for printing since it can go anywhere
tile_display_ptr := @tile_map0 
tile_term_ptr    := @tile_map0

' compute tile pitch at all times, incase its changed by user
tile_pitch := (1 << ((tile_map_control_parm & $000F) + 5))

' set up terminal size
term_num_cols := tile_pitch
term_num_rows := 24 ' this never changes since this is all the rows visible


' copyright and test
Print_String(@product_name_string1, 7, 18, 0, tile_display_ptr, tile_pitch)
Print_String(@product_name_string2, 7, 19, 0, tile_display_ptr, tile_pitch)
Print_String(@nurve_logo_string,    3, 23, 0, tile_display_ptr, tile_pitch)

' loop thru logo and display C3 logo
repeat index_y from 0 to 11 
  repeat index_x from 0 to 23
    ' extract next "bit" from bitmap
    ch := ( C3Bitmap[ index_y ] >> (23 - index_x)) & $01
    ch := lookupz ( ch : 32, 3*32+32+9 )
    Print_Char( ch, 4+index_x, 2+index_y, 0, tile_display_ptr, tile_pitch)

Print_String(string ("P R O P"), 12, 8, 0, tile_display_ptr, tile_pitch)

' stop the previous cog
stop

' launch a COG with ASM video driver
cogon := (cog := cognew(@HEL_GFX_Driver_Entry, @tile_map_base_ptr_parm)) > 0

' hold a sec to see logo...
repeat 500_000


  ' return back to repeat main event loop...
    
' parent COG will terminate now...if it gets to this point

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB stop

'' Stop the tile driver - frees a cog

  if cogon~
    cogstop(cog)

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

{

* Software Interface

The software interface consists of approx. 36 registers that are used to communicate with the virtual GPU. Thus all IO is performed
via these registers included memory access and uploading of data. Thus, the less data that is uploaded per frame the better.

Many registers can be read and written to, thus they might have different meaning depending on if they are being read or written.

Format of each register entry
Register Address, Read and Write, Register Name, Description

note: All final commands offset by: GPU_GFX_BASE_ID

GPU Sub-Function / Status
0 (Read) / GPU_GFX_subfunc_status
1 (write) / GPU_GFX_subfunc_status (Executes sub-function)

   (Read) - Reads the status of the GPU as defined below:
            returns 16-bit value formed from byte 1 = tile_status_bits_parm, byte 0 = tile_map_control_parm
                    bits 15 - 8               bits 7 - 0  
            eg: [ tile_status_bits_parm | tile_map_control_parm ]  
                 
   (Write) - Writes the GPU Sub-Function register and issues a high level command like copy, fill, etc.

             lower 8-bits defines sub-function
             upper 8-bits defines data for sub-function

    Sub-Function        Data                            Description                                                                        

    0 - CopyMem8       Byte size to copy 1 bytes    Copies numbytes from src -> dest in byte size chunks
    1 - FillMem8       Byte size to fill 1 bytes    Fills memory at dest with low byte of data16, 1 byte at a time 

    2 - CopyMem16      word size to copy 2 bytes    Copies numbytes from src -> dest in word size chunks
    3 - FillMem16      word size to fill 2 bytes    Fills memory at dest with data16, 2 bytes at a time 

    ..
     
Tile Map Ptr - 16-bit
2,3 (Read/Write) / tile_map_ptr - Reads or Writes the 16-bit tile map ptr which points to the current tile map displayed.

Display Ptr - 16-bit
4,5 (Read/Write) - tile_display_ptr - Reads or Writes 16-bit tile map ptr which points to the current display ptr into the tile map, low level print functions use this pointer.
                                                                
tile_term_ptr - 16-bit
6,7 (Read/Write) - tile_term_ptr - Reads or Writes 16-bit tile map ptr which points to the current terminal ptr into the tile map, terminal level print functions use this pointer.

Bitmap Ptr - 16-bit
8,9 (Read/Write) / bitmap_ptr - Reads or Writes the 16-bit tile bitmaps ptr which points to the bitmaps indexed by the tilemap. 

Palettes Ptr - 16-bit
10,11 (Read/Write) / palette_ptr - Reads or Writes the 16-bit palette array ptr which points to the palettes in use for the tilemap.
                                                                         

Overscan Colors 8-bits each 
12,13 (Read/Write) / top_overscan_col - Reads or Writes to the top overscan color drawn on the screen.
14,15 (Read/Write) / bot_overscan_col - Reads or Writes to the bottom overscan color drawn on the screen.


Horizontal and Vertical Fine Scroll Registers
16,17 (Read/Write) / hscroll_fine  -  Reads or Writes the current fine horizontal scroll register (0..7) 
18,19 (Read/Write) / vscroll_fine - Reads or Writes the current fine vertical scroll register (0..7)

Screen Width / Pitch
20,21 (Read/Write) / screen_width - Read or Writes the screen width value, 0=16 tiles, 1=32 tiles, 2=64 tiles, etc. 

Source Address for Memory Operations 16-bit
22,23 (Read/Write) / src_addr_ptr - Reads or Writes the 16-bit source address for GPU operations.


Destination Address for Memory Operations 16-bit
24,25 (Read/Write) / dest_addr_ptr - Reads or Writes the 16-bit destination address for GPU operations.


Number of Bytes to Copy, Move, Process etc. 16-bit
26,27 (Read/Write) / num_bytes - Reads or Writes the 16-bit number representing the number of bytes for a GPU operation Sub-Function.


Operational Data 16-bit
28,29 (Read/Write) / data16 - Reads or Writes to the 16-bit data word uses for GPU operations or memory access operations.


Memory Read/Write Port 16-bit
30,31 (Read/Write) / ramport8 - Reads or Writes the data from the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                                 After the operation, the src_addr_low|hi are incremented as per the GPU configuration.

Memory Read/Write Port 16-bit
32,33 (Read/Write) / ramport16 - Reads or Writes the data from the currently addressed memory location in the GPU pointed to by the src_addr_low|hi.
                                 After the operation, the src_addr_low|hi are incremented as per the GPU configuration.

Current Raster Line
34 (Read Only) / rasterline - Reads the current raster line being drawn from 0..191, or whatever the GPU's line resolution is set to.                             

GPU Configuratin Registers
35,36 (Read/Write) / set_auto_inc    - Reads or writes the current memory auto increment setting (0..15), default 0
        

}


PUB GPU_GFX_Process_Command( g_cmd, g_data16 ) : g_spi_result
  ' this function can be called with parameters to execute various commands to the register model of the GPU
  ' the SPI client running on the control COG with the virtual SPI interface will typically catch the GPU command and then
  ' pass it along here for effiecient processing.

{

  ' debug code
  Newline_Term
  String_Term( string ("GFX CMD = ( "))      
  Dec_Term( g_cmd )
  String_Term( string (", ") )
  Dec_Term( g_data16 )
  String_Term( string (" )") )
}


  ' BEGIN REGISTER INTERFACE CASE /////////////////////////////////////////////////////////////////////////////////////
  
  ' process command with proper handler
  case (g_cmd)
  
    ' GPU SUBFUNCTION/STATUS///////////////////////////////////////////////////////////////////////////////////////////

  
    GPU_GFX_SUBFUNC_STATUS_R: ' Reads the status of the GPU OR Writes the GPU Sub-Function register and issues a high level command like copy, fill, etc.                                                      

      'String_Term( string (":0"))

    GPU_GFX_SUBFUNC_STATUS_W: 
      'String_Term( string (":GFX_SUB ("))    
        
        ' extract API subfunction and data
      GPU_GFX_subfunc := g_data16 & $FF ' lower 8-bits defines sub-function
      GPU_GFX_data    := g_data16 >> 8  ' upper 8-bits defines data for sub-function

      'Dec_Term( GPU_GFX_subfunc )
      'String_Term( string (", ") )
      'Dec_Term( GPU_GFX_data )
      'String_Term( string (" )") )
        
      case (GPU_GFX_subfunc)
        
        GPU_GFX_SUBFUNC_COPYMEM16: ' 0 - Copies numbytes from src -> dest in wordsize chunks
          wordmove( tile_dest_addr_parm, tile_src_addr_parm, tile_numbytes_parm)
          'String_Term( string (":wordmove"))
          'Newline_Term
          
        GPU_GFX_SUBFUNC_FILLMEM16: ' 1 - Fills memory with data16, 2 bytes at a time
          wordfill( tile_dest_addr_parm, tile_data16_parm, tile_numbytes_parm >> 1)
          'String_Term( string (":wordfill"))
          'Newline_Term
          
        GPU_GFX_SUBFUNC_COPYMEM8:  ' 2 - Copies numbytes from src -> dest in byte size chunks
          bytemove( tile_dest_addr_parm, tile_src_addr_parm, tile_numbytes_parm >> 1) 
          'String_Term( string (":bytemove"))
          'Newline_Term
                    
        GPU_GFX_SUBFUNC_FILLMEM8:  ' 3 - Fills memory with low byte of data16, 1 bytes at a time 
          bytefill( tile_dest_addr_parm, tile_data16_parm & $FF, tile_numbytes_parm)             
          'String_Term( string (":bytefill"))
          'Newline_Term
          
        ' END CASE GPU_GFX_subfunc //////////////////////////////////////////////////////////////////////////////////////// 

    ' TILE MAP POINTER ////////////////////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_TILE_MAP_R:       ' Reads or Writes the 16-bit tile map ptr which points to the current tile map displayed.
      g_spi_result := tile_map_base_ptr_parm 
      'String_Term( string (":2"))


    GPU_GFX_TILE_MAP_W:
      tile_map_base_ptr_parm := g_data16
      'String_Term( string (":3"))

    ' TILE DISPLAY POINTER ////////////////////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_DISPLAY_PTR_R:       ' Reads or Writes the 16-bit tile display ptr which points to the location where low level printing is displayed
      g_spi_result := tile_display_ptr 
      'String_Term( string (":4"))
       
    GPU_GFX_DISPLAY_PTR_W:
      tile_display_ptr := g_data16
      'String_Term( string (":5"))

    ' TILE TERMINAL POINTER ////////////////////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_TERM_PTR_R:       ' Reads or Writes the 16-bit tile terminal ptr which points to the location where high level terminal mode printing is displayed
      g_spi_result := tile_term_ptr 
      'String_Term( string (":6"))


    GPU_GFX_TERM_PTR_W:
      tile_term_ptr := g_data16
      'String_Term( string (":7"))

    ' BITMAP POINTER //////////////////////////////////////////////////////////////////////////////////////////////////

    GPU_GFX_BITMAP_R:         ' Reads or Writes the low byte of the 16-bit tile bitmaps ptr which points to the bitmaps indexed by the tilemap. 
      g_spi_result := tile_bitmaps_base_ptr_parm
      'String_Term( string (":8"))


    GPU_GFX_BITMAP_W:         
      tile_bitmaps_base_ptr_parm := g_data16
      'String_Term( string (":9"))
      
    ' PALETTES POINTER ////////////////////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_PALETTE_R:        ' Reads or Writes the 16-bit palette array ptr which points to the palettes in use for the tilemap.
      g_spi_result := tile_palettes_base_ptr_parm
      'String_Term( string (":10"))
      
    GPU_GFX_PALETTE_W:        ' Reads or Writes the 16-bit palette array ptr which points to the palettes in use for the tilemap.
      tile_palettes_base_ptr_parm := g_data16
      'String_Term( string (":11"))
          
    ' OVERSCAN COLORS /////////////////////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_TOP_OVERSCAN_COL_R:   ' Reads or Writes to the top overscan color drawn on the screen.
      g_spi_result := (tile_map_control_parm & $00_FF_00_00) >> 16
      'String_Term( string (":12"))
      
    GPU_GFX_TOP_OVERSCAN_COL_W: 
      tile_map_control_parm := (tile_map_control_parm & $FF_00_FF_FF) | (g_data16 << 16)
      'String_Term( string (":13"))

    GPU_GFX_BOT_OVERSCAN_COL_R:   ' Reads or Writes to the bot overscan color drawn on the screen.
      g_spi_result := (tile_map_control_parm & $FF_00_00_00) >> 24
      'String_Term( string (":14"))

    GPU_GFX_BOT_OVERSCAN_COL_W: 
      tile_map_control_parm := (tile_map_control_parm & $00_FF_FF_FF) | (g_data16 << 24)
      'String_Term( string (":15"))

    ' FINE SCROLLING REGISTERS ////////////////////////////////////////////////////////////////////////////////////////

    GPU_GFX_HSCROLL_FINE_R:   ' Reads or Writes to the horizontal scroll register (not implemented in driver yet) 
      g_spi_result := (tile_map_control_parm & $00_00_F0_00) >> 12
      'String_Term( string (":16"))

    GPU_GFX_HSCROLL_FINE_W: 
      tile_map_control_parm := (tile_map_control_parm & $FF_FF_0F_FF) | (g_data16 << 8)
      'String_Term( string (":17"))

    GPU_GFX_VSCROLL_FINE_R:   ' Reads or Writes to the vertical scroll register 
      g_spi_result := (tile_map_control_parm & $00_00_0F_00) >> 8
      'String_Term( string (":18"))
      
    GPU_GFX_VSCROLL_FINE_W: 
      tile_map_control_parm := (tile_map_control_parm & $FF_FF_F0_FF) | (g_data16 << 8)
      'String_Term( string (":19"))

    ' SCREEN WIDTH / PITCH REGISTER ///////////////////////////////////////////////////////////////////////////////////

    GPU_GFX_SCREEN_WIDTH_R: ' Read or Writes the screen width value, 0=16 tiles, 1=32 tiles, 2=64 tiles, etc.
      g_spi_result := (tile_map_control_parm & $00_00_00_FF) 
      'String_Term( string (":20"))

    GPU_GFX_SCREEN_WIDTH_W: 
      tile_map_control_parm := (tile_map_control_parm & $FF_FF_FF_00) | (g_data16)

      ' re-compute tile pitch for "low level" display functions
      tile_pitch := (1 << ((tile_map_control_parm & $000F) + 5))

      ' re-adjust terminal size for "high level" terminal functions
      term_num_cols := tile_pitch
      term_num_rows := 24 ' this never changes since this is all the rows visible
      'String_Term( string (":21"))

    ' SOURCE ADDRESS FOR GPU OPERATIONS ///////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_SRC_ADDR_R:       '  Reads or Writes the low byte of the 16-bit source address for GPU operations.
      g_spi_result := tile_src_addr_parm
      'String_Term( string (":22"))

    GPU_GFX_SRC_ADDR_W:
      tile_src_addr_parm := g_data16
      'String_Term( string (":23"))

    ' DEST ADDRESS FOR GPU OPERATIONS ///////////////////////////////////////////////////////////////////////////////

    GPU_GFX_DEST_ADDR_R:       '  Reads or Writes the 16-bit destination address for GPU operations.
      g_spi_result := tile_dest_addr_parm
      'String_Term( string (":24"))

    GPU_GFX_DEST_ADDR_W:
      tile_dest_addr_parm := g_data16
      'String_Term( string (":25"))

    ' NUMBER OF BYTES FOR GPU OPERATIONS //////////////////////////////////////////////////////////////////////////////

    GPU_GFX_NUM_BYTES_R:      '  Reads or Writes the 16-bit number representing the number of bytes for a GPU operation Sub-Function.
      g_spi_result := tile_numbytes_parm
      'String_Term( string (":26"))

    
    GPU_GFX_NUM_BYTES_W:
      tile_numbytes_parm := g_data16
      'String_Term( string (":27"))

    ' 16-BIT DATA FOR GPU OPERATIONS //////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_DATA_R:           '  Reads or Writes to the 16-bit data word buffer used for GPU operations or memory access operations.
      g_spi_result := tile_data16_parm
      'String_Term( string (":28"))

 
    GPU_GFX_DATA_W:
      tile_data16_parm := g_data16
      'String_Term( string (":29"))

    ' 8-BIT DATA PORT FOR READ/WRITE OPERATIONS TO VIRTUAL MEMORY /////////////////////////////////////////////////////
    ' Note: they ALWAYS use the src register for both reading and writing
     
    GPU_GFX_RAM_PORT8_R:   ' Reads or Writes the 8-bit data from the currently addressed memory location in the GPU pointed to by the src_addr_ptr.
                           ' After the operation, the src_addr_ptr are incremented as per the GPU configuration.
      g_spi_result := BYTE[ tile_src_addr_parm]

      ' update src pointer based on auto increment value
      tile_src_addr_parm += tile_mem_autoinc_parm
      'String_Term( string (":30"))


    GPU_GFX_RAM_PORT8_W:
      BYTE[ tile_src_addr_parm] := g_data16
 
      ' update src pointer based on auto increment value
      tile_src_addr_parm += tile_mem_autoinc_parm
      'String_Term( string (":31"))


    ' 16-BIT DATA PORT FOR READ/WRITE OPERATIONS TO VIRTUAL MEMORY /////////////////////////////////////////////////////
     
    GPU_GFX_RAM_PORT16_R:  ' Reads or Writes the 16-bit data from the currently addressed memory location in the GPU pointed to by the src_addr_ptr.
                           ' After the operation, the src_addr_ptr are incremented as per the GPU configuration.
      g_spi_result := WORD[ tile_src_addr_parm]

      ' update src pointer based on auto increment value
      tile_src_addr_parm += tile_mem_autoinc_parm
      'String_Term( string (":32"))


    GPU_GFX_RAM_PORT16_W:
      WORD[ tile_src_addr_parm] := g_data16
 
      ' update src pointer based on auto increment value
      tile_src_addr_parm += tile_mem_autoinc_parm
      'String_Term( string (":33"))   

    ' CURRENT RASTER LINE /////////////////////////////////////////////////////////////////////////////////////////////
     
    GPU_GFX_RASTER_LINE_R:      '  Read (only) the current raster line being drawn from 0..191, or whatever the GPU's line resolution is set to.                             
      g_spi_result := tile_rasterline_parm 
      'String_Term( string (":34"))

    ' GPU CONFIGURATION ///////////////////////////////////////////////////////////////////////////////////////////////

    GPU_GFX_SET_AUTO_INC_R:       ' Reads or writes the current memory auto byte increment setting (0..15), default 0
      g_spi_result := tile_mem_autoinc_parm 
      'String_Term( string (":35"))
    
    GPU_GFX_SET_AUTO_INC_W:       ' Reads or writes the current memory auto byte increment setting (0..15), default 0
      tile_mem_autoinc_parm := g_data16
      'String_Term( string (":36"))
      
  return (g_spi_result)

' end PUB GPU_GFX_Process_Command(...)

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
' Text Console/Terminal Functions - These functions can be called much like the other terminal drivers for high level printing
' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB String_Term(string_ptr)

'' Print a zero-terminated string to terminal
  repeat strsize(string_ptr)
    Out_Term(byte[string_ptr++])

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Dec_Term(value) | i

'' Print a decimal number to terminal

  if value < 0
    -value
    Print_Term("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      Print_Term(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      Print_Term("0")
    i /= 10

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Hex_Term(value, digits)

'' Print a hexadecimal number to terminal

  value <<= (8 - digits) << 2
  repeat digits
    Print_Term(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Bin_Term(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    Print_Term((value <-= 1) & 1 + "0")

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Newline_Term | i

  term_col := 0
  if (++term_row == term_num_rows)
    term_row--
    'scroll lines
    wordmove(tile_term_ptr, tile_term_ptr + (tile_pitch << 1), ( (term_num_rows-1) * tile_pitch )  )

   'clear new line                             
    wordfill(tile_term_ptr + ((term_num_rows-1) * (tile_pitch << 1)), ( (term_attr << 8) | (" " - 32)), term_num_cols)


' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
       
PUB Print_Term(ch)

  ' print character at current cursor position HERE
  word[tile_term_ptr][term_col + (term_row * tile_pitch)] := (term_attr << 8) | (ch - 32)
  
  if ++term_col == term_num_cols
    Newline_Term

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Out_Term(c) | i, k

'' Output a character to terminal, this is the primary interface from the client to the driver in "terminal mode"
'' direct mode access uses the register model and controls the engine from the low level
''
''     $00 = clear screen
''     $01 = home
''     $08 = backspace
''     $09 = tab (8 spaces per)
''     $0A = set X position (X follows)
''     $0B = set Y position (Y follows)
''     $0C = set color (color follows)
''     $0D = return
''  others = printable characters

  case term_flag
    $00: case c
           $00: wordfill(tile_term_ptr, ( (term_attr << 8) | (" " - 32)), term_num_cols * term_num_rows)
                term_col := term_row := 0
           $01: term_col := term_row := 0
           $08: if term_col
                  term_col--
           $09: repeat
                  Print_Term(" ")
                while term_col & 7
           $0A..$0C: term_flag := c
                     return
           $0D: Newline_Term

           other: Print_Term(c)
           
    $0A: term_col := c // term_num_cols
    $0B: term_row := c // term_num_rows
    $0C: term_color := c & 7
  term_flag := 0


' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


PUB VSync
' waits for vsync pulse
  repeat until ((tile_status_bits_parm & $01) == $01)
  repeat until ((tile_status_bits_parm & $01) == $00)


' /////// ADDED FUNCTIONS /////////////////////////////////////////////////////

PUB SetX(new_x)
' set x cursor position
  term_col := new_x // term_num_cols 

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB SetY(new_y)
' set y cursor position
  term_row := new_y // term_num_rows

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB GotoXY( new_x, new_y )
' set x cursor position
  term_col := new_x // term_num_cols 

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

' set y cursor position
  term_row := new_y // term_num_rows

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB GetX
' retrieve x cursor position 
  return(term_col)

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB GetY
' retrieve y cursor position
  return(term_row) 

'//////////////////////////////////////////////////////////////////////////////
'LOW LEVEL DIRECT FUNCTIONS ///////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB Print_Char(ch, tx, ty, attr, tile_mem_ptr, tile_mem_pitch)
' prints a single character to the screen in "direct mode"
  word[tile_mem_ptr][tx + (ty * tile_mem_pitch)] := (attr << 8) | (ch - 32)  

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Print_String(string_ptr, tx, ty, attr, tile_mem_ptr, tile_mem_pitch) | sindex
' prints a string to the screen in "direct mode"

sindex := 0
repeat while (byte[string_ptr][sindex] <> 0)
  word[tile_mem_ptr][tx + sindex + (ty * tile_mem_pitch)] := (attr << 8) | (byte[string_ptr][sindex] - 32)  
  sindex++

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Hexdigit_To_ASCII(hex_value)
' converts a number 0..15 to ASCII 0...9, A, B, C, D, E, F
    if (hex_value > 9)
      return ( hex_value + 33 - 10 )
    else
      return( hex_value + 16 )

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Print_Num(number, base, tx, ty, attr, tile_mem_ptr, tile_mem_pitch ) | divisor, digit
' prints a number to the screen
' note this function must use locals only since its called by multiple processors, in other words it must be re-entrant

if (base == 2)
  ' print with base 2, 8 bit number
  divisor := %10000000 ' initialize bitmask
  
  repeat digit from 0 to 7
    ' print with base 2
    word[tile_mem_ptr][tx + digit + (ty * tile_mem_pitch)] :=(attr << 8) | ((number & divisor) >> (7-digit) + 16)
    divisor >>= 1

elseif (base == 10)
  divisor := 100000 ' initialize largest base needed 
  repeat digit from 0 to 5
    ' print with base 10
    word[tile_mem_ptr][tx + digit + (ty * tile_mem_pitch)] := (attr << 8) | ((number / divisor) + 16)
    number := number // divisor
    divisor /= 10
else  
  divisor := $F_0_0_0_0_0_0_0 ' initialize largest base needed 
  repeat digit from 0 to 7
    ' print with base 16
    
    word[tile_mem_ptr][tx + digit + (ty * tile_mem_pitch)] := (attr << 8) | (Hexdigit_To_ASCII ((number & divisor) >> ((7-digit)*4)))
    divisor >>= 4

' //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
' TILE DRIVER DATA
' //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DAT

' for convenience I have create 4 tile maps in data format, (2) 32x24 followed by (2) 64x24, in fact, its all just memory so you can do whatever
' you want with it

tile_maps     word
              ' you place all your 32x24 tile maps here, you can have as many as you like, in real-time simply re-point the
              ' tile_map_base_ptr_parm to any time map and within 1 frame the tile map will update

              ' 32x24 WORDS each, (0..768 WORDs,1536 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 32x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31  
tile_map0     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23  

              ' 32x24 WORDS each, (0..768 WORDs,1536 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 32x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              


              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31  
tile_map1     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23  


              ' 32x24 WORDS each, (0..768 WORDs,1536 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 32x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31  
tile_map2     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23  

              ' 32x24 WORDS each, (0..768 WORDs,1536 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 32x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              


              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31  
tile_map3     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23  

{
' uncomment these if you want to add them in for more pre-defined tile maps

              ' 64x24 WORDS each, (0..1536 WORDs,3072 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 64x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31     32      33    34     35     36     37     38     39     40     41     42     43     44     45      46    47      48    49     50     51     52     53     54     55     56     57     58     59     60     61     62     63
tile_map4     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23


              ' 64x24 WORDS each, (0..1536 WORDs,3072 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 64x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              
              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31     32      33    34     35     36     37     38     39     40     41     42     43     44     45      46    47      48    49     50     51     52     53     54     55     56     57     58     59     60     61     62     63
tile_map5     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23

              ' 64x24 WORDS each, (0..1536 WORDs,3072 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 64x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31     32      33    34     35     36     37     38     39     40     41     42     43     44     45      46    47      48    49     50     51     52     53     54     55     56     57     58     59     60     61     62     63
tile_map6     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23

              ' 64x24 WORDS each, (0..1536 WORDs,3072 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 64x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              
              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31     32      33    34     35     36     37     38     39     40     41     42     43     44     45      46    47      48    49     50     51     52     53     54     55     56     57     58     59     60     61     62     63
tile_map7     word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17             
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 23

}

' /////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////

_tile_bitmaps    WORD

' 144 bitmaps Extracted from file "c64_font_05_16b.bmp", followed by 16 blanks for user definition
' of course, user can overwrite these as well with direct memory acess, if he wishes to do so

' Extracted from file "c64_font_05_16b.bmp" at tile=(0, 0), size=(8, 8), palette index=(0)
c64_font_05__tx0_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(1, 0), size=(8, 8), palette index=(1)
c64_font_05__tx1_ty0_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(2, 0), size=(8, 8), palette index=(2)
c64_font_05__tx2_ty0_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(3, 0), size=(8, 8), palette index=(3)
c64_font_05__tx3_ty0_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_2_2_1_1_2_2_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1
              
' Extracted from file "c64_font_05_16b.bmp" at tile=(4, 0), size=(8, 8), palette index=(4)
c64_font_05__tx4_ty0_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1


' Extracted from file "c64_font_05_16b.bmp" at tile=(5, 0), size=(8, 8), palette index=(5)
c64_font_05__tx5_ty0_bitmap    WORD
              WORD %%1_2_1_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_1_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(6, 0), size=(8, 8), palette index=(6)
c64_font_05__tx6_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%2_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%2_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(7, 0), size=(8, 8), palette index=(7)
c64_font_05__tx7_ty0_bitmap    WORD
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(8, 0), size=(8, 8), palette index=(8)
c64_font_05__tx8_ty0_bitmap    WORD
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(9, 0), size=(8, 8), palette index=(9)
c64_font_05__tx9_ty0_bitmap    WORD
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(10, 0), size=(8, 8), palette index=(10)
c64_font_05__tx10_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(11, 0), size=(8, 8), palette index=(11)
c64_font_05__tx11_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(12, 0), size=(8, 8), palette index=(12)
c64_font_05__tx12_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(13, 0), size=(8, 8), palette index=(13)
c64_font_05__tx13_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(14, 0), size=(8, 8), palette index=(14)
c64_font_05__tx14_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(15, 0), size=(8, 8), palette index=(15)
c64_font_05__tx15_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(16, 0), size=(8, 8), palette index=(16)
c64_font_05__tx16_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_1_2_2_1
              WORD %%1_2_2_1_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(17, 0), size=(8, 8), palette index=(17)
c64_font_05__tx17_ty0_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(18, 0), size=(8, 8), palette index=(18)
c64_font_05__tx18_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(19, 0), size=(8, 8), palette index=(19)
c64_font_05__tx19_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(20, 0), size=(8, 8), palette index=(20)
c64_font_05__tx20_ty0_bitmap    WORD
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_2_2_2_2_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(21, 0), size=(8, 8), palette index=(21)
c64_font_05__tx21_ty0_bitmap    WORD
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(22, 0), size=(8, 8), palette index=(22)
c64_font_05__tx22_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(23, 0), size=(8, 8), palette index=(23)
c64_font_05__tx23_ty0_bitmap    WORD
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(24, 0), size=(8, 8), palette index=(24)
c64_font_05__tx24_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(25, 0), size=(8, 8), palette index=(25)
c64_font_05__tx25_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(26, 0), size=(8, 8), palette index=(26)
c64_font_05__tx26_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(27, 0), size=(8, 8), palette index=(27)
c64_font_05__tx27_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(28, 0), size=(8, 8), palette index=(28)
c64_font_05__tx28_ty0_bitmap    WORD
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(29, 0), size=(8, 8), palette index=(29)
c64_font_05__tx29_ty0_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(30, 0), size=(8, 8), palette index=(30)
c64_font_05__tx30_ty0_bitmap    WORD
              WORD %%1_1_1_1_2_2_2_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(31, 0), size=(8, 8), palette index=(31)
c64_font_05__tx31_ty0_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(0, 1), size=(8, 8), palette index=(32)
c64_font_05__tx0_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_1_2_2_1
              WORD %%1_2_2_2_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(1, 1), size=(8, 8), palette index=(33)
c64_font_05__tx1_ty1_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(2, 1), size=(8, 8), palette index=(34)
c64_font_05__tx2_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(3, 1), size=(8, 8), palette index=(35)
c64_font_05__tx3_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(4, 1), size=(8, 8), palette index=(36)
c64_font_05__tx4_ty1_bitmap    WORD
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(5, 1), size=(8, 8), palette index=(37)
c64_font_05__tx5_ty1_bitmap    WORD
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(6, 1), size=(8, 8), palette index=(38)
c64_font_05__tx6_ty1_bitmap    WORD
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(7, 1), size=(8, 8), palette index=(39)
c64_font_05__tx7_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_2_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(8, 1), size=(8, 8), palette index=(40)
c64_font_05__tx8_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(9, 1), size=(8, 8), palette index=(41)
c64_font_05__tx9_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(10, 1), size=(8, 8), palette index=(42)
c64_font_05__tx10_ty1_bitmap    WORD
              WORD %%1_2_2_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(11, 1), size=(8, 8), palette index=(43)
c64_font_05__tx11_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_1_1_2_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(12, 1), size=(8, 8), palette index=(44)
c64_font_05__tx12_ty1_bitmap    WORD
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(13, 1), size=(8, 8), palette index=(45)
c64_font_05__tx13_ty1_bitmap    WORD
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_2_1_2_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_1_2_1_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(14, 1), size=(8, 8), palette index=(46)
c64_font_05__tx14_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_2_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_2_2_2_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(15, 1), size=(8, 8), palette index=(47)
c64_font_05__tx15_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(16, 1), size=(8, 8), palette index=(48)
c64_font_05__tx16_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(17, 1), size=(8, 8), palette index=(49)
c64_font_05__tx17_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(18, 1), size=(8, 8), palette index=(50)
c64_font_05__tx18_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(19, 1), size=(8, 8), palette index=(51)
c64_font_05__tx19_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(20, 1), size=(8, 8), palette index=(52)
c64_font_05__tx20_ty1_bitmap    WORD
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(21, 1), size=(8, 8), palette index=(53)
c64_font_05__tx21_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(22, 1), size=(8, 8), palette index=(54)
c64_font_05__tx22_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(23, 1), size=(8, 8), palette index=(55)
c64_font_05__tx23_ty1_bitmap    WORD
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_1_2_1_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_2_1_2_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(24, 1), size=(8, 8), palette index=(56)
c64_font_05__tx24_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(25, 1), size=(8, 8), palette index=(57)
c64_font_05__tx25_ty1_bitmap    WORD
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(26, 1), size=(8, 8), palette index=(58)
c64_font_05__tx26_ty1_bitmap    WORD
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(27, 1), size=(8, 8), palette index=(59)
c64_font_05__tx27_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(28, 1), size=(8, 8), palette index=(60)
c64_font_05__tx28_ty1_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(29, 1), size=(8, 8), palette index=(61)
c64_font_05__tx29_ty1_bitmap    WORD
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(30, 1), size=(8, 8), palette index=(62)
c64_font_05__tx30_ty1_bitmap    WORD
              WORD %%1_1_1_1_1_2_1_1
              WORD %%1_1_1_1_2_1_2_1
              WORD %%1_1_1_2_1_1_1_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(31, 1), size=(8, 8), palette index=(63)
c64_font_05__tx31_ty1_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(0, 2), size=(8, 8), palette index=(64)
c64_font_05__tx0_ty2_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(1, 2), size=(8, 8), palette index=(65)
c64_font_05__tx1_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(2, 2), size=(8, 8), palette index=(66)
c64_font_05__tx2_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(3, 2), size=(8, 8), palette index=(67)
c64_font_05__tx3_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(4, 2), size=(8, 8), palette index=(68)
c64_font_05__tx4_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(5, 2), size=(8, 8), palette index=(69)
c64_font_05__tx5_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(6, 2), size=(8, 8), palette index=(70)
c64_font_05__tx6_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(7, 2), size=(8, 8), palette index=(71)
c64_font_05__tx7_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_2_2_2_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(8, 2), size=(8, 8), palette index=(72)
c64_font_05__tx8_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(9, 2), size=(8, 8), palette index=(73)
c64_font_05__tx9_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(10, 2), size=(8, 8), palette index=(74)
c64_font_05__tx10_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_2_2_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(11, 2), size=(8, 8), palette index=(75)
c64_font_05__tx11_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(12, 2), size=(8, 8), palette index=(76)
c64_font_05__tx12_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(13, 2), size=(8, 8), palette index=(77)
c64_font_05__tx13_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_1_2_1_2_2_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(14, 2), size=(8, 8), palette index=(78)
c64_font_05__tx14_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(15, 2), size=(8, 8), palette index=(79)
c64_font_05__tx15_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(16, 2), size=(8, 8), palette index=(80)
c64_font_05__tx16_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(17, 2), size=(8, 8), palette index=(81)
c64_font_05__tx17_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(18, 2), size=(8, 8), palette index=(82)
c64_font_05__tx18_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(19, 2), size=(8, 8), palette index=(83)
c64_font_05__tx19_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(20, 2), size=(8, 8), palette index=(84)
c64_font_05__tx20_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(21, 2), size=(8, 8), palette index=(85)
c64_font_05__tx21_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(22, 2), size=(8, 8), palette index=(86)
c64_font_05__tx22_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(23, 2), size=(8, 8), palette index=(87)
c64_font_05__tx23_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_1_1_1_2_2_1
              WORD %%2_2_1_2_1_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(24, 2), size=(8, 8), palette index=(88)
c64_font_05__tx24_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(25, 2), size=(8, 8), palette index=(89)
c64_font_05__tx25_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_2_2_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(26, 2), size=(8, 8), palette index=(90)
c64_font_05__tx26_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(27, 2), size=(8, 8), palette index=(91)
c64_font_05__tx27_ty2_bitmap    WORD
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(28, 2), size=(8, 8), palette index=(92)
c64_font_05__tx28_ty2_bitmap    WORD
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(29, 2), size=(8, 8), palette index=(93)
c64_font_05__tx29_ty2_bitmap    WORD
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(30, 2), size=(8, 8), palette index=(94)
c64_font_05__tx30_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%2_1_1_2_1_1_2_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(31, 2), size=(8, 8), palette index=(95)
c64_font_05__tx31_ty2_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_2_1_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(0, 3), size=(8, 8), palette index=(96)
c64_font_05__tx0_ty3_bitmap    WORD
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_1_1_1_1_1_1_2
              WORD %%2_1_1_1_1_1_1_2
              WORD %%2_1_1_1_1_1_1_2
              WORD %%2_1_1_1_1_1_1_2
              WORD %%2_1_1_1_1_1_1_2
              WORD %%2_1_1_1_1_1_1_2
              WORD %%2_2_2_2_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(1, 3), size=(8, 8), palette index=(97)
c64_font_05__tx1_ty3_bitmap    WORD
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(2, 3), size=(8, 8), palette index=(98)
c64_font_05__tx2_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(3, 3), size=(8, 8), palette index=(99)
c64_font_05__tx3_ty3_bitmap    WORD
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(4, 3), size=(8, 8), palette index=(100)
c64_font_05__tx4_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_1_2_2_1_1_2_2
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_2_1_1_2_2_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(5, 3), size=(8, 8), palette index=(101)
c64_font_05__tx5_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(6, 3), size=(8, 8), palette index=(102)
c64_font_05__tx6_ty3_bitmap    WORD
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_1_2_2_1_1_2_2
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_1_2_2_1_1_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(7, 3), size=(8, 8), palette index=(103)
c64_font_05__tx7_ty3_bitmap    WORD
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(8, 3), size=(8, 8), palette index=(104)
c64_font_05__tx8_ty3_bitmap    WORD
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_1_1_2_2_1_1_2
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_2_2_1_1_2_2_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_1_1_2_2_1_1_2
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_2_2_1_1_2_2_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(9, 3), size=(8, 8), palette index=(105)
c64_font_05__tx9_ty3_bitmap    WORD
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%2_1_1_2_2_1_1_2
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_2_2_1_1_2_2_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%2_1_1_2_2_1_1_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(10, 3), size=(8, 8), palette index=(106)
c64_font_05__tx10_ty3_bitmap    WORD
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(11, 3), size=(8, 8), palette index=(107)
c64_font_05__tx11_ty3_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%2_2_2_2_2_1_1_1
              WORD %%2_2_2_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(12, 3), size=(8, 8), palette index=(108)
c64_font_05__tx12_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(13, 3), size=(8, 8), palette index=(109)
c64_font_05__tx13_ty3_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%2_2_2_2_2_1_1_1
              WORD %%2_2_2_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(14, 3), size=(8, 8), palette index=(110)
c64_font_05__tx14_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_2_2_2
              WORD %%1_1_1_2_2_2_2_2
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(15, 3), size=(8, 8), palette index=(111)
c64_font_05__tx15_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(16, 3), size=(8, 8), palette index=(112)
c64_font_05__tx16_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_2_1_1_1
              WORD %%2_2_2_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(17, 3), size=(8, 8), palette index=(113)
c64_font_05__tx17_ty3_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(18, 3), size=(8, 8), palette index=(114)
c64_font_05__tx18_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(19, 3), size=(8, 8), palette index=(115)
c64_font_05__tx19_ty3_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_2_2_2
              WORD %%1_1_1_2_2_2_2_2
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(20, 3), size=(8, 8), palette index=(116)
c64_font_05__tx20_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(21, 3), size=(8, 8), palette index=(117)
c64_font_05__tx21_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(22, 3), size=(8, 8), palette index=(118)
c64_font_05__tx22_ty3_bitmap    WORD
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(23, 3), size=(8, 8), palette index=(119)
c64_font_05__tx23_ty3_bitmap    WORD
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(24, 3), size=(8, 8), palette index=(120)
c64_font_05__tx24_ty3_bitmap    WORD
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(25, 3), size=(8, 8), palette index=(121)
c64_font_05__tx25_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(26, 3), size=(8, 8), palette index=(122)
c64_font_05__tx26_ty3_bitmap    WORD
              WORD %%2_1_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%1_2_2_1_1_1_1_1
              WORD %%1_1_2_2_1_2_2_1
              WORD %%1_1_1_2_2_2_2_1
              WORD %%1_1_1_1_2_2_2_1
              WORD %%1_1_1_1_1_2_2_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(27, 3), size=(8, 8), palette index=(123)
c64_font_05__tx27_ty3_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(28, 3), size=(8, 8), palette index=(124)
c64_font_05__tx28_ty3_bitmap    WORD
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(29, 3), size=(8, 8), palette index=(125)
c64_font_05__tx29_ty3_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_2_2_2
              WORD %%1_1_1_2_2_2_2_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(30, 3), size=(8, 8), palette index=(126)
c64_font_05__tx30_ty3_bitmap    WORD
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(31, 3), size=(8, 8), palette index=(127)
c64_font_05__tx31_ty3_bitmap    WORD
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(0, 4), size=(8, 8), palette index=(128)
c64_font_05__tx0_ty4_bitmap    WORD
              WORD %%1_2_2_1_2_2_1_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(1, 4), size=(8, 8), palette index=(129)
c64_font_05__tx1_ty4_bitmap    WORD
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(2, 4), size=(8, 8), palette index=(130)
c64_font_05__tx2_ty4_bitmap    WORD
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_1_2_2_1_1_1_1
              WORD %%1_2_2_2_2_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(3, 4), size=(8, 8), palette index=(131)
c64_font_05__tx3_ty4_bitmap    WORD
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%2_2_2_2_2_2_2_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_1_1_2_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(4, 4), size=(8, 8), palette index=(132)
c64_font_05__tx4_ty4_bitmap    WORD
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_2_2_1_1
              WORD %%1_1_1_1_2_2_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(5, 4), size=(8, 8), palette index=(133)
c64_font_05__tx5_ty4_bitmap    WORD
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_1_2_2_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(6, 4), size=(8, 8), palette index=(134)
c64_font_05__tx6_ty4_bitmap    WORD
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_1_2_2_1_1_2_2
              WORD %%2_2_1_1_2_2_1_1
              WORD %%2_2_1_1_2_2_1_1
              WORD %%1_1_2_2_1_1_2_2
              WORD %%1_1_2_2_1_1_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(7, 4), size=(8, 8), palette index=(135)
c64_font_05__tx7_ty4_bitmap    WORD
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_2_2_2_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(8, 4), size=(8, 8), palette index=(136)
c64_font_05__tx8_ty4_bitmap    WORD
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_2_1_1_1_1_1
              WORD %%1_2_2_2_1_1_1_1
              WORD %%1_1_2_2_2_1_1_1
              WORD %%1_1_1_2_2_2_1_1
              WORD %%1_1_1_1_2_2_2_1
              WORD %%1_1_1_1_1_2_2_2
              WORD %%1_1_1_1_1_1_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(9, 4), size=(8, 8), palette index=(137)
c64_font_05__tx9_ty4_bitmap    WORD
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2
              WORD %%1_1_1_1_1_1_2_2

' Extracted from file "c64_font_05_16b.bmp" at tile=(10, 4), size=(8, 8), palette index=(138)
c64_font_05__tx10_ty4_bitmap    WORD
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(11, 4), size=(8, 8), palette index=(139)
c64_font_05__tx11_ty4_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%2_2_1_1_1_1_1_1
              WORD %%1_2_2_2_2_2_1_1
              WORD %%1_2_2_1_2_2_2_1
              WORD %%1_2_2_1_2_2_1_1
              WORD %%1_2_2_1_2_2_1_1
              WORD %%1_1_1_1_1_1_1_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(12, 4), size=(8, 8), palette index=(140)
c64_font_05__tx12_ty4_bitmap    WORD
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%2_2_2_2_1_1_1_1
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2
              WORD %%1_1_1_1_2_2_2_2

' these last 3 are SPACE INVADERS :)

' Extracted from file "c64_font_05_16b.bmp" at tile=(13, 4), size=(8, 8), palette index=(141)
c64_font_05__tx13_ty4_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%2_2_1_2_2_1_2_2
              WORD %%2_2_2_2_2_2_2_2
              WORD %%2_1_2_2_2_2_1_2
              WORD %%1_1_2_1_1_2_1_1
              WORD %%1_2_1_2_2_1_2_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(14, 4), size=(8, 8), palette index=(142)
c64_font_05__tx14_ty4_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_2_1_1_2_1_1
              WORD %%2_1_1_2_2_1_1_2
              WORD %%2_1_2_1_1_2_1_2
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_1_2_1_1_2_1_1
              WORD %%1_2_1_1_1_1_2_1

' Extracted from file "c64_font_05_16b.bmp" at tile=(15, 4), size=(8, 8), palette index=(143)
c64_font_05__tx15_ty4_bitmap    WORD
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_2_2_1_1_1
              WORD %%1_1_2_2_2_2_1_1
              WORD %%1_2_1_2_2_1_2_1
              WORD %%1_2_2_2_2_2_2_1
              WORD %%1_1_2_1_1_2_1_1
              WORD %%1_2_1_1_1_1_2_1
              WORD %%1_1_2_1_1_2_1_1

' Some space for blank bitmaps, so you don't have to move things around (16 of them)

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1

              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1
              WORD %%1_1_1_1_1_1_1_1



' palettes extracted from file "c64_font_04_8b.bmp" using palette color map 1
' only palette index 0,1 are used right now for character font, palette index 0 makes characters normal
' palette index 1 makes them look inverse video
_tile_palette_map    LONG

c64_font_04__palette_map    LONG
              LONG $07_07_0C_02 ' palette index 0
              LONG $07_0C_07_02 ' palette index 1

              LONG $07_0C_07_02 ' palette index 2
              LONG $07_0C_07_02 ' palette index 3
              LONG $07_0C_07_02 ' palette index 4
              LONG $07_0C_07_02 ' palette index 5
              LONG $07_0C_07_02 ' palette index 6
              LONG $07_0C_07_02 ' palette index 7
              LONG $07_0C_07_02 ' palette index 8
              LONG $07_0C_07_02 ' palette index 9
              LONG $07_0C_07_02 ' palette index 10
              LONG $07_0C_07_02 ' palette index 11
              LONG $07_0C_07_02 ' palette index 12                                                                                                                                                          
              LONG $07_0C_07_02 ' palette index 13
              LONG $07_0C_07_02 ' palette index 14
              LONG $07_0C_07_02 ' palette index 15

' /////////////////////////////////////////////////////////////////////////////

' bitmap for C3 24x12
C3Bitmap  LONG %0_1_1_1_1_1_1_1_1_1_0_0_1_1_1_1_1_1_1_1_1_1_1_0
          LONG %1_1_1_1_1_1_1_1_1_1_1_0_1_1_1_1_1_1_1_1_1_1_1_1
          LONG %1_1_1_1_1_1_1_1_1_1_1_0_1_1_1_1_1_1_1_1_1_1_1_1
          LONG %1_1_1_0_0_0_0_1_1_1_1_0_1_1_1_0_0_0_0_0_0_1_1_1
          LONG %1_1_1_0_0_0_0_1_1_1_1_0_1_1_1_0_0_0_0_0_0_1_1_1
          LONG %1_1_1_0_0_0_0_0_0_0_0_0_0_0_0_0_1_1_1_1_1_1_1_1
          LONG %1_1_1_0_0_0_0_0_0_0_0_0_0_0_0_0_1_1_1_1_1_1_1_1
          LONG %1_1_1_0_0_0_0_1_1_1_1_0_1_1_1_0_0_0_0_0_0_1_1_1
          LONG %1_1_1_0_0_0_0_1_1_1_1_0_1_1_1_0_0_0_0_0_0_1_1_1
          LONG %1_1_1_1_1_1_1_1_1_1_1_0_1_1_1_1_1_1_1_1_1_1_1_1
          LONG %1_1_1_1_1_1_1_1_1_1_1_0_1_1_1_1_1_1_1_1_1_1_1_1
          LONG %0_1_1_1_1_1_1_1_1_1_0_0_1_1_1_1_1_1_1_1_1_1_1_0

nurve_logo_string      byte "Nurve Networks LLC (",32*5+14 ,") 2011",0
product_name_string1   byte "C3 GFX Driver 1.0",0  
product_name_string2   byte "32x24 TILE ENGINE",0


 ' /////////////////////////////////////////////////////////////////////////////
' ASSEMBLY LANGUAGE HEL_GFX_DRIVER
' Engine 3 Design Notes
' Status: In progress
'
' Specs:
' - 32x24 tile map, each entry 2 bytes, format: [palette index:tile index]
' - 8x8 bitmap tiles, 4 colors each, array of 8 WORDs, each WORD represents 8 pixels, 16 bytes each tile, 4 LONGs
' - Course scrolling, horizontal and vertical, playfield width 32, 64, 128, 256
' 
' Features:
' - Multiple tile maps 256
' - Multiple palettes 256
' - Bitmaps 256
' - All assets base pointers passed to engine each frame, thus "on the fly" changes can be made
' - Supports page flipping, double buffering by design.
' - Engine passes back vertical region and hsyn/vsync status as well as current line
' /////////////////////////////////////////////////////////////////////////////


CON



 

  FNTSC         = 3_579_545      ' NTSC color clock frequency in HZ


' if one of the following sets of LNTSC / SNTSC don't work then comment them out and try the other set on your monitor

  ' these settings work on 90% of TVs and 50% of LCDs, but look a little better on TVs than the following two values
  'LNTSC         = (227*16)       ' NTSC color cycles per line (220-227) * 16
  'SNTSC         = (67*16)        ' NTSC color cycles per sync (39-44) * 16

  ' DEFAULT safe settings
  ' these settings work on 99% of all TVs and LCD monitors, but don't look as rich as previous settings in some cases, 
  LNTSC = (227*16 + 1*4)          ' NTSC color cycles per line (220-228) * 16 +- epsilon
  SNTSC = (51*16 - 0)             ' NTSC color cycles per sync (39-54) * 16 +- epsilon
  
  VNTSC         = (LNTSC-SNTSC)  ' NTSC color cycles per active video * 16
  PNTSC256      = (VNTSC/16)     ' NTSC color cycles per "compressed/expanded" on screen pixel
                                 ' allows us to put more or less pixels on the screen, but
                                 ' remember NTSC is 224 visible pixels -- period, so when we display more
                                 ' than 224 per line then we are getting chroma distortion on pixel boundaries
                                 ' a more recommended method for cleaner graphics is about 180-190 pixels horizontally
                                 ' this way you don't overdrive the chroma bandwidth which limits colors to 224
                                 ' color clocks per screen
                                 ' we are going to leave clocks per pixel the same, but CHANGE the clocks per frame to 1/2 the value
                                 ' it was for the 16 pixel tiles, this way the timing stays the same, but the waitvid will need
                                 ' 2 - 8 pixel sets 2x as fast



' Chameleon settings
  VIDEO_PINMASK    = %0111_0000     ' vcfg S = pinmask  (pin15 ->0111_0000<-pin8), only want upper nibble lower 3-bits
  VIDEO_PINGROUP   = 1              ' vcfg D = pingroup (Chameleon uses group 1, pins 8-15)
  VIDEO_SETUP      = %0_11_1_11_000 ' vcfg I = controls overall setting, we want baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
  VIDEO_CNTR_SETUP = %00001_111     ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                                    ' needn't set D,S fields since they set pin A/B I/Os, but mode 1 is internal, thus irrelvant

{

 ' HYDRA settings
  VIDEO_PINMASK    = %0000_0111     ' vcfg S = pinmask  (pin31 ->0000_0111<-pin24), only want lower 3-bits
  VIDEO_PINGROUP   = 3              ' vcfg D = pingroup (Hydra uses group 3, pins 24-31)
  VIDEO_SETUP      = %0_10_1_01_000 ' vcfg I = controls overall setting, we want baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
  VIDEO_CNTR_SETUP = %00001_111     ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                                    ' needn't set D,S fields since they set pin A/B I/Os, but mode 1 is internal, thus irrelvant

}


' format of tile_status_bits_parm, only the Vsync status bit is updated
'
' byte 3 (unused)|byte 2 (line)|   byte 1 (tile postion)    |                     byte 0 (sync and region)      |
'|x x x x x x x x| line 8-bits | row 4 bits | column 4-bits |x x x x | region 2-bits | hsync 1-bit | vsync 1-bit|
'   b31..b24         b23..b16      b15..b12     b11..b8                    b3..b2          b1            b0
' Region 0=Top Overscan, 1=Active Video, 2=Bottom Overscan, 3=Vsync
' NOTE: In this version of the tile engine only VSYNC and REGION are valid 

  ' tile engine status bits/masks
  TSB_VSYNC                     = %0000_00_01    ' Vsync 
  TSB_HSYNC                     = %0000_00_10    ' Hsync 

  ' regions indicate which parts of the vertical scan the raster is in, useful to time all rendering during the total blanking period which
  ' is any region other than "01"
  TSB_REGION_MASK               = %0000_11_00    ' region status bits bitmask

  TSB_REGION_TOP_OVERSCAN       = %0000_00_00    ' top overscan region (00)
  TSB_REGION_ACTIVE_VIDEO       = %0000_01_00    ' active video region (01)
  TSB_REGION_BOTT_OVERSCAN      = %0000_10_00    ' bottom overscan region (10)
  TSB_REGION_VSYNC              = %0000_11_00    ' vsync region (replicated in vsync bit)                                                

  
  ' register indexes
  CLKFREQ_REG = 0                ' register address of global clock frequency

  ' debuging stuff
  DEBUG_LED_PORT_MASK = $00000001 ' debug LED is on I/O P0



DAT
                org $000  ' set the code emission for COG add $000

' /////////////////////////////////////////////////////////////////////////////
' Entry point and Parameter Interface
' /////////////////////////////////////////////////////////////////////////////
{

there are 4 parameters in this engine, listed in order below
tile engine data structure pointers (can be changed in real-time by app!)
input parms

long tile_map_base_ptr_parm      ' parm 0: pointer to tile map (16x12 array, row major, 2-bytes per entry)
long tile_bitmaps_base_ptr_parm  ' parm 1: pointer to the base of the bitmap data, each 16x16, that is, 32 longs each
long tile_palettes_base_ptr_parm ' parm 2: pointer to the base of the palettes array, each palette 1 long, or 4 bytes.
long tile_map_control_parm       ' parm 3: pointer to the value that holds the overscan colors and number of "screens" or multiples of 16 that the tile map is
                                 ' format $bc_tc_hv_ww, tc = top overscan color, bc= bottom overscan color, h=horizontal scroll (0..7), v=vertical scroll (0..7),
                                 ' ww = width, set for width 16 tiles, 0 = 32 tiles, 1 = 64 tiles, 2 = 128 tiles, 3 = 256 tiles, etc.
                                 ' eg. w=0 would be 16 wide (standard), 1 would be 32 tiles, 2 would be 64 tiles, etc.
                                 ' this allows multiscreen width playfields and thus large horizontal/vertical scrolling games
                                 ' note that the final size is always a power of 2 times 16

real-time engine status variables, these are updated in real time by the
tile engine itself, so they can be monitored outside in SPIN/ASM by game
output parms

long tile_status_bits_parm      ' parm 4: pointer to status word - vsync, hsync, etc.

 format of tile_status_bits_parm, only the Vsync status bit is updated

 byte 3 (unused)|byte 2 (line)|   byte 1 (tile postion)    |                     byte 0 (sync and region)      |
|x x x x x x x x| line 8-bits | row 4 bits | column 4-bits |x x x x | region 2-bits | hsync 1-bit | vsync 1-bit|
   b31..b24         b23..b16      b15..b12     b11..b8                    b3..b2          b1            b0
 Region 0=Top Overscan, 1=Active Video, 2=Bottom Overscan, 3=Vsync
 NOTE: In this version of the tile engine only VSYNC and REGION are valid 

format of a tile map:

DAT

              ' you place all your 32x24 tile maps here, you can have as many as you like, in real-time simply re-point the
              ' tile_map_base_ptr_parm to any time map and within 1 frame the tile map will update

              ' 32x24 WORDS each, (0..768 WORDs,1536 bytes per tile map) 2-BYTE tiles (msb)[palette_index | tile_index](lsb)
              ' 32x24 tile map, each tile is 2 bytes, there are a total of 256 tiles possible, and thus 256 palettes              

              ' column     0      1      2      3      4      5      6      7      8      9     10     11     12     13     14     15      16    17      18    19     20     21     22     23     24     25     26     27     28     29     20     31  
tile_map0     word      $02_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01 ' row 0
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 1
              word      $00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 2
              word      $00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 3
              word      $00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_01,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 4
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 5
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 6
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 7
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 8
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 9
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 10
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 11
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 12
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 13
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 14
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 15
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 16
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 17
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 18
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 19
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 20
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 21
              word      $00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00 ' row 22
              word      $00_01,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_00,$00_01 ' row 23 

Example: upper left corner tile @ (0, 0) value is $02_01, this means use palette 2 and tile bitmap 1.
Note: 1. you can have as many tile maps as you wish there is no limit, they can represent rooms, etc.
      2. if you use "mappy" tool then you must make some changes tothe mapwin.ini file, they are:

      maptype="LW1H1A1-0" <--this results in 1 byte per entry, and the format is a binary single array in the format
      importskip=0        <---use dumb import, don't modify anything, just bring the tiles into the tool from your bitmap
      
      width, height,
      index 0, index 1, ..... index 31,
      index 32, index 1, ..... index 63,
      .
      .      
      index 736, index 177, ..... index 191

      as one single byte stream of course.

DAT

              ' tile bitmap memory, each tile 8x8 pixels, or 1 WORD by 8,
              ' 16-bytes each, also, note that they are mirrored right to left
              ' since the VSU streams from low to high bits, so your art must
              ' be reflected, we could remedy this in the engine, but for fun
              ' I leave it as a challenge in the art, since many engines have
              ' this same artifact
              ' for this demo, only 4 tile bitmaps defined

              ' empty tile

tile_square   word      %%3_3_3_3_3_3_3_3 ' tile 0
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_0_0_0_0_0_0_3
              word      %%3_3_3_3_3_3_3_3

Example: this particular bitmap definition is a square uses color 3 and fill with color 0,
Note: you may have up to 256 tiles


Format of palette array, each palette has 4 entries, each 1 byte, when a tile is defined in the tile map it has both an index
to the tile bitmap (0..255) as well as an index to the "palette" to use (0..255), below is an example of 4 palette entries

DAT

palette_map   ' 4 palettes for only now, palette memory (1..255) LONGs, each palette 4-BYTEs

              long $5C_5B_5A_02 ' palette 0
              long $0C_0B_0A_02 ' palette 1
              long $CC_CB_CA_02 ' paleete 2
              long $0C_0B_0A_02 ' palette 3

Note: each of the 8-bit values represents the color, luma, and chroma enable for the color as defined by the VSU hardware


}
              
HEL_GFX_Driver_Entry

              ' VCFG: setup Video Configuration register and 3-bit TV DAC pins to outputs
                        
              movs    vcfg, #VIDEO_PINMASK              ' vcfg S = pinmask  (pin31 ->0000_0111<-pin24), only want lower 3-bits
              movd    vcfg, #VIDEO_PINGROUP             ' vcfg D = pingroup (Hydra uses group 3, pins 24-31)
              movi    vcfg, #VIDEO_SETUP                ' vcfg I = controls overall setting, we want baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband

              or      dira, tvport_mask                 ' set DAC pins to output 24, 25, 26 for hydra, 12, 13, 14 for Chameleon

              ' CTRA: setup Frequency to Drive Video                        
              movi    ctra, #VIDEO_CNTR_SETUP           ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                                                        ' needn't set D,S fields since they set pin A/B I/Os, but mode 1 is internal, thus irrelvant

              ' compute the value to place in FREQ such that the final counter
              ' output is NTSC and the PLL output is 16*NTSC
              mov     r1, v_freq                        ' r1 <- TV color burst frequency in Hz, eg. 3_579_545                                             
              rdlong  r2, #CLKFREQ_REG                  ' r2 <- CLKFREQ is register 0, eg. 80_000_000
              call    #Dividefract                      ' perform r3 = 2^32 * r1 / r2

              mov     frqa, r3                          ' set frequency for counter such that bit 31 is toggling at a rate of the color burst (2x actually)
                                                        ' which means that the freq number added at a rate of CLKFREQ (usually 80.000 Mhz) results in a
                                                        ' pll output of the color burst, this is further multiplied by 16 as the final PLL output
                                                        ' thus giving the chroma hardware the clock rate of 16X color burst which is what we want :)


              mov       r0, par                         ' copy boot parameter value and read in parameters from main memory, must be on LONG boundary
              add       r0, #16
              mov       tile_status_bits_ptr, r0        ' ptr to status bits, so tile engine can pass out status of tile engine in real time


Next_Frame    ' start of new frame of 262 scanlines
              ' 26 top overscan
              ' 192 active vide
              ' 26 bottom overscan
              ' 18 vertical sync
                      
              ' read run-time parameters from main memory, user can change these values every frame
              mov       r0, par                         ' copy boot parameter value and read in parameters from main memory, must be on LONG boundary
              rdlong    tile_map_base_ptr, r0           ' base ptr to tile map itself
              add       r0, #4
              rdlong    tile_bitmaps_base_ptr, r0       ' base pointer to array of 8x8 bitmaps, each 16 bytes
              add       r0, #4
              rdlong    tile_palettes_base_ptr, r0      ' base pointer to array of palettes, each palette 4 bytes / 1 long
              add       r0, #4
              rdlong    tile_map_control, r0            ' value of overscan color and tile map width 0=32, 1=64, 2=128, etc.

              mov       tile_map_overscan_color, tile_map_control     ' extract overscan color
              shr       tile_map_overscan_color, #16

              ' bug was OR'ing colors into overscan?? Which wasn't overwritten which it should be, FIXED
              mov        v_choverscan, tile_map_overscan_color        ' COPY into system overscan colors    

              mov       tile_map_width, tile_map_control              ' extract tile map width
              and       tile_map_width, #$FF 

              mov       r0, #0                          ' clear out status bits for next frame, 0 means region top overscan, vsync FALSE
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory


' /////////////////////////////////////////////////////////////////////////////
Top_Overscan_Scanlines

              mov       r0, #TSB_REGION_TOP_OVERSCAN    ' set region to top overscan 
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              mov     r1, #26                           ' set # of scanlines

' Horizontal Scanline Loop (r1 itterations)
:Next_Scanline
        
              ' HSYNC 10.9us (Horizontal Sync) including color burst
              mov     vscl, v_shsync                    ' set the video scale register such that 16 serial pixels takes 39 color clocks (the total time of the hsync and burst)
              waitvid v_chsync, v_phsync                ' send out the pixels and the colors, the pixels are 0,1,2,3 that index into the colors, the colors are blank, sync, burst
                                                        ' we use them to create the hsync pulse itself

              ' set hsync output status bit and set region
              mov       r0, #TSB_REGION_TOP_OVERSCAN | TSB_HSYNC ' set region to top overscan and set hsync 
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              
              ' HVIS 52.6us (Visible Scanline) draw 16 huge pixels
              mov     vscl, v_shvis                     ' set up the video scale so the entire visible scan is composed of 16 huge pixels
              waitvid v_choverscan , v_phoverscan_top   ' draw 16 pixels with overscan color

              mov       r0, #TSB_REGION_TOP_OVERSCAN    ' set region to top overscan and reset hsync 
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              
              djnz    r1, #:Next_Scanline               ' are we done with the scanlines yet?

' /////////////////////////////////////////////////////////////////////////////
Active_Scanlines

              mov       r0, #TSB_REGION_ACTIVE_VIDEO    ' set region to active video
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              mov       r1, tile_map_control            ' reset scanline counter (consider scroll), in this case we count up
                                                        ' format $bc_tc_hv_ww
              shr       r1, #8                          ' shift vertical scroll value into lower bits
              and       r1, #$07                        ' mask off vertical scroll value and store
              mov       r7, r1                          ' r7 = r1+192, holds value of last line to render with scrolling
              add       r7, #192

              mov       tile_map_hscroll, tile_map_control ' now extract the horizontal scroll position
              shr       tile_map_hscroll, #12
              and       tile_map_hscroll, #$07             ' save for later

' Horizontal Scanline Loop (r1 iterations)
:Next_Scanline
        
              ' HSYNC 10.9us (Horizontal Sync) including color burst
              mov     vscl, v_shsync                    ' set the video scale register such that 16 serial pixels takes 39 color clocks (the total time of the hsync and burst)
              waitvid v_chsync, v_phsync                ' send out the pixels and the colors, the pixels are 0,1,2,3 that index into the colors, the colors are blank, sync, burst
                                                        ' we use them to create the hsync pulse itself

              ' set hsync output status bit and set region and active line
              mov       r0, r1
              shl       r0, #16                                     ' set the current line                        
              or        r0, #TSB_REGION_ACTIVE_VIDEO | TSB_HSYNC    ' set region to active video reset hsync 
              wrlong    r0, tile_status_bits_ptr                    ' write out status bits to main memory
                       
' /////////////////////////////////////////////////////////////////////////////

              ' this next section retrives the tile index, palette index for each tile and draws a single line of video composed of 1 sub
              ' scan line of each tile, the opportunity here is to realize that each tile accessed 16 times, once for each video line, thus
              ' the tile indexes and palettes themselves could be "cached" to save time, however, the added complexity is not needed yet
              ' but the next version will cache the data, so we can have more time per pixel block to do crazy stuff :)

              mov     vscl, v_spixel                    ' set up video scale register for 8 pixels at a time
              
              ' select proper sub-tile row address
              mov       r2, r1                          ' r1 holds current video line of active video (0..191)
              and       r2, #$07                        ' r2 = r1 mod 8, this is the sub-row in the tile we want to render, or the WORD index we need to offset the tile memory by

              ' access main memory can get the bitmap data
              shl       r2, #1                          ' r2 = r2*2, byte based row offset of pixels, 2 bytes per pixel row, 8 pixels, 2 bits per pixel
              add       r2, tile_bitmaps_base_ptr       ' r2 = tile_bitmaps_base_ptr + r2
                                                                                
              ' compute tile index itself for left edge of tile row, inner loop will index across as the scanline is rendered
              ' new code is needed here to support scrolling horizontally, we need to be able to tell the engine that a tile row has MORE than
              ' 8 tiles per row, so that when it access the next tile row down, it calculates properly, so this nice power of 2 math has to be
              ' revisited, however, we are going to constrain that the tiles per row must be a power of 2 multiple of 32 at least, so we don't have to use
              ' complex division yet, trying to keep each engine simple... so the playfield widths are  32, 64, 128, 256 (max), thus the
              ' only thing we need to add is a shift to the mask operation that performs the (line/8) * 32, since now the 32 could be 32, 64, 128, etc.

              ' new code with support for playfields 32, 64, 128... in size which then allows scrolling to be supported (course scrolling of course :)
              mov       tile_map_index_ptr, r1                ' r1 = line, copy it 
              and       tile_map_index_ptr, #$1F8             ' tile_map_index_ptr = [(r1 / 8) * 32..256], this is the starting tile index for a row, 0, 32, 64, ...
              shl       tile_map_index_ptr, #2                ' results in tile_map_index_ptr = [(r1 / 8) * 32]                  
              shl       tile_map_index_ptr, tile_map_width    ' this finishes off the playfield x2, x4, x8, etc. scrolling support


              shl       tile_map_index_ptr, #1                ' tile_map_index_ptr = tile_index*2, since each tile is 2 bytes, we need to convert index to byte address
              add       tile_map_index_ptr, tile_map_base_ptr ' tile_map_index_ptr = [(r1 / 8) * 32..256] + tile_map_base_ptr, this is a byte address in main memory now


              ' at this point we have everything we need for the pixels rendering aspect of the 32 tiles that will be rendered, the inner loop logic will
              ' retrieve the time map indexes, and access the 8 pixels that make up each row of each tile, BUT we need to get the palette(s)
              ' for each tile as well, each tile has its own palette, but the palette will change each group of 8-pixels across the screen since
              ' each 8-pixels represents a single line from a different tile

              ' we could cache all the palettes into the local cache, but for fun let's just read them out of main memory during the inner loop
                  
              ' render the 32 tile lines, r2 is holding proper row address, but we need to add base of actual tile we want rendered
              mov       r4, #32

:Pixel_Render_Loop

              ' read next tile index and palette index from main memory
              rdword    tile_map_word, tile_map_index_ptr
              
              ' retrieve 8-pixels of current row from proper bitmap referenced by tile
              mov       r3, tile_map_word
              and       r3, #$FF                        ' mask off upper 8-bits they hold the palette index, we aren't interested in
              shl       r3, #4                          ' r3 = tile_map_index*16 (bytes per tile bitmap)
              add       r3, r2                          ' r3 = tile_map_index*16 + tile_bitmaps_base_ptr + video_line mod 8
              rdword    r3, r3                          ' r3 = main_memory[r3], retrieve 16 bits of pixel data (8 pixels)
                                                        ' 16 clocks until hub comes around, try and be ready, move a couple instructions that aesthically
                                                        ' should be in one place between the hub reads to maximize processing/memory bandwith
                                                        
              mov       v_pixels_buffer, r3             ' r3 holds pixels now, copy to pixel out buffer

              ' retrieve palette for current tile
              mov       r5, tile_map_word
              shr       r5, #8                          ' r5 now holds the palette index and we shifted out the tile index into oblivion
              shl       r5, #2                          ' multiple by 4, since there ar 4-bytes per palette entry
                                                        ' r5 = palette_map_index*4
              add       r5, tile_palettes_base_ptr      ' r5 = palette_map_base_ptr +palette_map_index*4

              ' moved from top of loop to eat time after previous memory read!
              add       tile_map_index_ptr, #2          ' advance pointer 2 bytes to next tile map index entry, for next pass

              rdlong    v_colors_buffer, r5             ' read the palette data into the buffer
              
              'draw the pixels with the selected palette (8 pixels, waitvid will come around 2x as fast now!)
              waitvid   v_colors_buffer, v_pixels_buffer

              ' reset hsync output status bit and set region and active line
              mov       r0, r1
              shl       r0, #16                         ' set the current line                        
              or        r0, #TSB_REGION_ACTIVE_VIDEO    ' set region to active video reset hsync 
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              djnz      r4, #:Pixel_Render_Loop         ' loop until we draw 32 tiles (single pixel row of each)

' /////////////////////////////////////////////////////////////////////////////

              add       r1, #1                          
              cmp       r1, r7                  wc, wz  ' #192 normally
        if_b  jmp       #:Next_Scanline                 ' if ++r1 (current line) < 192 then loop
        
' /////////////////////////////////////////////////////////////////////////////

Bottom_Overscan_Scanlines

              mov       r0, #TSB_REGION_BOTT_OVERSCAN   ' set region to bottom overscan
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              mov       r1, #26                         ' set # of scanlines

' Horizontal Scanline Loop (r1 itterations)
:Next_Scanline


              ' HSYNC 10.9us (Horizontal Sync) including color burst
              mov       vscl, v_shsync                  ' set the video scale register such that 16 serial pixels takes 39 color clocks (the total time of the hsync and burst)
              waitvid   v_chsync, v_phsync              ' send out the pixels and the colors, the pixels are 0,1,2,3 that index into the colors, the colors are blank, sync, burst
                                                        ' we use them to create the hsync pulse itself
              ' set hsync output status bit and set region
              mov       r0, #TSB_REGION_BOTT_OVERSCAN | TSB_HSYNC ' set region to bottom overscan and set hsync 
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory


              ' HVIS 52.6us (Visible Scanline) draw 16 huge pixels that represent an entire line of video
              mov       vscl, v_shvis                   ' set up the video scale so the entire visible scan is composed of 16 huge pixels
              waitvid   v_choverscan, v_phoverscan_bot  ' draw 16 pixels with bottom overscan color

              ' reset hsync output status bit and set region
              mov       r0, #TSB_REGION_BOTT_OVERSCAN   ' set region to bottom overscan and reset hsync 
              wrlong    r0, tile_status_bits_ptr        ' write out status bits to main memory

              djnz      r1, #:Next_Scanline             ' are we done with the scanlines yet?

' /////////////////////////////////////////////////////////////////////////////
Vsync_Pulse

              ' VSYNC Pulse (Vertical Sync)
              ' 18 scanlines: 6 'high syncs', 6 'low syncs', and finally another 6 'high syncs'
              ' refer to NTSC spec, but this makes up the equalization pulses needed for a vsync

              mov       r0, #TSB_VSYNC | TSB_REGION_VSYNC ' set region to vsync as well 
              wrlong    r0, tile_status_bits_ptr          ' write out vertical sync status bit to main memory

              call      #Vsync_High
              call      #Vsync_Low
              call      #Vsync_High
'                        
              jmp       #Next_Frame                     ' that's it, do it a googleplex times...
                        
'//////////////////////////////////////////////////////////////////////////////
' SUB-ROUTINES VARIABLE DECLARATIONS
'//////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////
' vsync_high: Generate 'HIGH' vsync signal for 6 horizontal lines.
Vsync_High              
                               
              mov       r1, #6
                        
              ' HSYNC 10.9us (Horizontal Sync)
:Vsync_Loop   mov       vscl, v_shsync
              waitvid   v_chsync, v_pvsync_high_1

              ' HVIS 52.6us (Visible Scanline)
              mov       vscl, v_shvis
              waitvid   v_chsync, v_pvsync_high_2
              djnz      r1, #:Vsync_Loop

Vsync_High_Ret
              ret

' /////////////////////////////////////////////////////////////////////////////
' vsync_low: Generate 'LOW' vsync signal for 6 horizontal lines.
Vsync_Low
                               
              mov       r1, #6
                        
              ' HSYNC 10.9us (Horizontal Sync)
:Vsync_Loop   mov       vscl, v_shsync
              waitvid   v_chsync, v_pvsync_low_1

              ' HVIS 52.6us (Visible Scanline)
              mov       vscl, v_shvis
              waitvid   v_chsync, v_pvsync_low_2
              djnz      r1, #:Vsync_Loop

Vsync_Low_Ret
              ret

' /////////////////////////////////////////////////////////////////////////////
' Calculates 2^32 * r1/r2, result stored in r3, r1 must be less that r2, that is, r1 < r2
' the results of the division are a binary weighted 32-bit fractional number where each bit
' is equal to the following weights:
' MSB (31)    30    29 ..... 0
'      1/2   1/4   1/8      1/2^32
Dividefract                                     
              mov       r0,#32+1                        ' 32 iterations, are we done yet?
:Loop         cmpsub    r1,r2           wc              ' does divisor divide into dividend?
              rcl       r3,#1                           ' rotate carry into result
              shl       r1,#1                           ' shift dividend over
              djnz      r0,#:Loop                       ' done with division yet?

Dividefract_Ret
              ret                                       ' return to caller with result in r3

'//////////////////////////////////////////////////////////////////////////////
' VARIABLE DECLARATIONS
'//////////////////////////////////////////////////////////////////////////////

' general purpose registers
                        
r0            long      $0                             
r1            long      $0
r2            long      $0
r3            long      $0
r4            long      $0
r5            long      $0
r6            long      $0
r7            long      $0
                                           
' tv output DAC port bit mask
'tvport_mask   long      %0000_0111 << 24        ' Hydra DAC is on bits 24, 25, 26
tvport_mask   long      %0111_0000 << 8        ' Chameleon DAC is on bits 12, 13, 14


' output buffers to hold colors and pixels, start them off with "test" data
                        '3  2  1  0   <- color indexes
v_pixels_buffer long    %%1111222233330000
v_colors_buffer long    $5C_CC_0C_03            ' 3-RED | 2-GREEN | 1-BLUE | 0-BLACK

' pixel VSCL value for 256 visible pixels per line (clocks per pixel 8 bits | clocks per frame 12 bits )
' notice we divide PNTSC256 by 2, this in essence means draw only 8 pixels per waitvid, this way we can
' double the number of tiles, and only 8 of the 16 pixels in the "pixels" array sent to waitvid are rendered
v_spixel      long      ((PNTSC256 >> 4) << 12) + (PNTSC256 >> 1)


' hsync VSCL value (clocks per pixel 8 bits | clocks per frame 12 bits )
v_shsync      long      ((SNTSC >> 4) << 12) + SNTSC

' hsync colors (4, 8-bit values, each represent a color in the format chroma shift, chroma modulatation enable, luma | C3 C2 C1 C0 | M | L2 L1 L0 |
                        '3  2  1  0   <- color indexes
v_chsync      long      $00_00_02_8A ' SYNC (3) / SYNC (2) / BLACKER THAN BLACK (1) / COLOR BURST (0)

' hsync pixels
                        ' BP  |BURST|BW|    SYNC      |FP| <- Key BP = Back Porch, Burst = Color Burst, BW = Breezway, FP = Front Porch
'v_phsync      long      %%1_1_0_0_0_0_1_2_2_2_2_2_2_2_1_1
'v_phsync      long      %%1_1_1_1_0_0_0_0_1_2_2_2_2_1_1_1 ' perfect on all sets, looks like this is the winner!
v_phsync      long      %%1_1_1_1_0_0_0_1_2_2_2_2_2_1_1_1 ' perfect on all sets, looks like this is the winner!

' active video values
v_shvis       long      ((VNTSC >> 4) << 12) + VNTSC

' the colors used, 4 of them always
                        'red, color 3 | dark blue, color 2 | blue, color 1 | light blue, color 0
v_chvis       long      $5A_0A_0B_0C            ' each 2-bit pixel below references one of these 4 colors, (msb) 3,2,1,0 (lsb)

' the pixel pattern                             
v_phvis       long      %%3210_0123_3333_3333   ' 16-pixels, read low to high is rendered left to right, 2 bits per pixel
                                                ' the numbers 0,1,2,3 indicate the "colors" to use for the pixels, the colors
                                                ' are defined by a single byte each with represents the chroma shift, modulation,
                                                ' and luma
' the colors used, 4 of them always
                        'grey, color 3 | dark grey, color 2 | bottom overscan user controlled, color 1 | top overscan user controlled, color 0
v_choverscan  long      $07_04_00_00            ' each 2-bit pixel below references one of these 4 colors, (msb) 3,2,1,0 (lsb)

' the pixel pattern
v_phoverscan_top  long      %%0000_0000_0000_0000   ' 16-pixels, read low to high is rendered left to right, 2 bits per pixel
                                                    ' the numbers 0,1,2,3 indicate the "colors" to use for the pixels, the colors
                                                    ' are defined by a single byte each with represents the chroma shift, modulation,
                                                    ' and luma, always uses palette color 0

v_phoverscan_bot  long      %%1111_1111_1111_1111

' vsync pulses 6x High, 6x Low, 6x High
' the vertical sync pulse according to the NTSC spec should be composed of a series
' of pulses called the pre-equalization, serration pulses (the VSYNC pulse itself), and the post-equalization pulses
' there are 6 pulses of each, and they more or less inverted HSYNC, followed by 6 HSYNC pulses, followed by 6 more inverted HSYNC pulses.
' this keeps the horizontal timing circutry locked as well as allows the 60 Hz VSYNC filter to catch the "vsync" event.
' the values 1,2 index into "colors" that represent sync and blacker than black.
' so the definitions below help with generated the "high" and "low" dominate HSYNC timed pulses which are combined
' to generated the actual VSYNC pulse, refer to NTSC documentation for more details.

{                                
v_pvsync_high_1         long    %%1_1_1_1_1_1_1_1_1_1_1_2_2_2_1_1  
v_pvsync_high_2         long    %%1_1_1_1_1_1_1_1_1_1_1_1_1_1_1_1
                                
v_pvsync_low_1          long    %%2_2_2_2_2_2_2_2_2_2_2_2_2_2_1_1
v_pvsync_low_2          long    %%1_2_2_2_2_2_2_2_2_2_2_2_2_2_2_2
}

v_pvsync_high_1         long    %%1_1_1_1_1_1_1_1_1_1_2_2_2_2_1_1  
v_pvsync_high_2         long    %%1_1_1_1_1_1_1_1_1_1_1_1_1_1_1_1
                                
v_pvsync_low_1          long    %%2_2_2_2_2_2_2_2_2_2_2_2_2_1_1_1
v_pvsync_low_2          long    %%1_2_2_2_2_2_2_2_2_2_2_2_2_2_2_2

  
v_freq                  long    FNTSC

' tile engine locals
tile_row                long    $0
video_line              long    $0
tile_map_ptr            long    $0
tile_cache_ptr          long    $0
palette_cache_ptr       long    $0
tile_map_index          long    $0
tile_map_index_ptr      long    $0
tile_map_word           long    $0
tile_palette_index      long    $0

' tile engine passed parameters
tile_map_base_ptr       long    $0
tile_bitmaps_base_ptr   long    $0
tile_palettes_base_ptr  long    $0
tile_map_control        long    $0

' derived parms from passed parameters
tile_map_overscan_color long    $0      
tile_map_width          long    $0

tile_map_vscroll        long    $0
tile_map_hscroll        long    $0

tile_status_bits_ptr    long    $0

' local COG cache memories (future expansion)
tile_cache              word    $0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0
palette_cache           long    $0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0,$0

' END ASM /////////////////////////////////////////////////////////////////////